/* ======================================================================
 *
 *  CCCCC          valtest.c (aka "inputs" -- the Makefile target)
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: March 14, 2000
 * CC   CC         Last Modified: November 17, 2020 (minor edit 05/05/21)
 *  CCCCC
 *
 * Code file for a sample VR application with multiple valuators.
 *  - initially created to test the cyberglove interface.
 *  - started as a copy of travel.c
 *  - now serves as a useful application when running as an input server
 *    (hence a new Makefile target was added called "inputs" that uses this file.)
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#include <stdio.h>	/* needed for printf & all other standard I/O */
#include <signal.h>	/* needed for signal() */
#include <stdlib.h>	/* needed for exit() */

#include "freevr.h"


/* These functions will be used as arguments (for callbacks) */
void	init_gfx();
void	draw_server_world();
void	exit_application();
void	update_world();
char	*list_all_inputs();
char	*list_self_controls();

#undef	SHOW_CONFIG_LOCATION	/* Define this to have inputs & controls reveal their source */

/*********************************************************************/
int main(int argc, char* argv[])
{
	char	buffer[4096];		/* string buffer for self-control list */

	signal(SIGINT, SIG_IGN);

	vrConfigure(&argc, argv, NULL);
	vrStart();

	signal(SIGINT, exit_application);

	/* TODO: It would be nice to put these above vrStart(), since */
	/*   it will block until all input devices are open, and this */
	/*   way we could be rendering something while waiting.       */
	/* TODO: Actually, we almost *need* to have the init_gfx stuff*/
	/*   set before vrStart, so the window opening function will  */
	/*   have something to call (other than vrDoNothing()).       */
	/*   In fact, I have no idea how init_gfx() gets called at    */
	/*   all given the current setup.                             */
	vrFunctionSetCallback(VRFUNC_ALL_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_ALL_DISPLAY, vrCallbackCreate(draw_server_world, 0));
	vrFunctionSetCallback(VRFUNC_HANDLE_USR2, vrCallbackCreate(vrDbgInfo, 0));

	if (vrContext->input->num_input_devices > 0) {
		vrDbgPrintfN(SELDOM_DBGLVL+1, "input viewer: input device[0] version = '%s'\n",
			vrContext->input->input_devices[0]->version);
	} else {
		vrDbgPrintfN(SELDOM_DBGLVL+1, "input viewer: No input devices defined.\n");
	}

	vrSystemSetName("inputs -- FreeVR inputs display application");
	vrSystemSetAuthors("Bill Sherman");
	vrPrintf(list_all_inputs(buffer, vrContext->input, sizeof(buffer), verbose));
	vrPrintf(list_self_controls(buffer, vrContext->input, sizeof(buffer), verbose));
	vrSystemSetExtraInfo(buffer);	/* also add the self-controls info to the Help message */

	vrDbgPrintf("input viewer: looping\n");
	while(vrFrame()) {
		update_world();
	}

	exit_application();

	exit(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application()
{
	vrDbgPrintfN(ALWAYS_DBGLVL, "input viewer: quitting.\n");
	vrExit();
}


/***********************/
/* update_world(): ... */
/***********************/
void update_world()
{
	int		input;

	/* Nothing happens, there are no moving objects in the world. */

	/* Set DebugLevel to 200 to print this info to screen every frame */
	if (vrDbgDo(RARE_DBGLVL)) {
		vrMsgPrintf("input viewer: controller --");

		vrMsgPrintf(" buttons:");
		for (input = 0; input < vrContext->input->num_2ways; input++)
			vrMsgPrintf(" %d", vrContext->input->switch2[input]->value);

		vrMsgPrintf(" valuators:");
		for (input = 0; input < vrContext->input->num_valuators; input++)
			vrMsgPrintf(" %6.3f", vrContext->input->valuator[input]->value);

		vrMsgPrintf("\n");
	}

	/* Allow other processes to get some work done */
	vrSleep(10000);
}


/*****************************************************************************/
/* list_all_inputs(): A function that creates a string containing a list of  */
/*   all the configured inputs within the active FreeVR system.  This string */
/*   can then be printed or added to the help string.                        */
/*                                                                           */
/* The arguments are based on the vrSprintInputUI() function in vr_input.c.  */
/*   As with that, the "style" argument is currently ignored.                */
/*                                                                           */
/* TODO: Consider adding a robust version of this function as part of the    */
/*   overall vr_input.c code base.                                           */
/* TODO: Make use of the vrPrintStyle argument to control how much is output.*/
/*****************************************************************************/
char *list_all_inputs(char *str, vrInputInfo *inputinfo, int str_size, vrPrintStyle style)
{
	int	bytes_remaining = str_size;
	int	bytes_used = 0;
	int	bytes;			/* return result from snprintf() calls -- usually # of bytes written */

	int	count;
	int	devcount;
	vrInput	*input;

	/* Warning: snprintf() returns number of bytes that *would* have been written, not actual! */

	bytes = snprintf(str, bytes_remaining, "Num Buttons: %d\n", inputinfo->num_2ways);
	bytes_used += bytes; bytes_remaining -= bytes;
	if ((bytes < 1) || (bytes_remaining < 1)) { str[str_size] = '\0'; return str; }	/* for safety */

	for (count = 0; count < inputinfo->num_2ways; count++) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, "  Button %d: '%s' -- '%s' -- From Device: '%s'\n",
			count,
			inputinfo->switch2[count]->my_object->name,
			inputinfo->switch2[count]->my_object->desc_str,
			inputinfo->switch2[count]->my_device->name);
		bytes_used += bytes; bytes_remaining -= bytes;
#ifdef SHOW_CONFIG_LOCATION
		bytes = snprintf(&str[bytes_used], bytes_remaining, ANSI_TEXTFAINT "               Created in '%s' on %d\n" ANSI_TEXTNORM,
			inputinfo->switch2[count]->my_object->file_lastmod,
			inputinfo->switch2[count]->my_object->line_lastmod);
		bytes_used += bytes; bytes_remaining -= bytes;
#endif
		if ((bytes < 1) || (bytes_remaining < 1)) { str[str_size] = '\0'; return str; }	/* for safety */
	}

	bytes = snprintf(&str[bytes_used], bytes_remaining, "Num Valuators: %d\n", inputinfo->num_valuators);
	bytes_used += bytes; bytes_remaining -= bytes;
	if ((bytes < 1) || (bytes_remaining < 1)) { str[str_size] = '\0'; return str; }	/* for safety */
	for (count = 0; count < inputinfo->num_valuators; count++) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, "  Valuator %d: '%s' -- '%s' -- From Device: '%s'\n",
			count,
			inputinfo->valuator[count]->my_object->name,
			inputinfo->valuator[count]->my_object->desc_str,
			inputinfo->valuator[count]->my_device->name);
		bytes_used += bytes; bytes_remaining -= bytes;
#ifdef SHOW_CONFIG_LOCATION
		bytes = snprintf(&str[bytes_used], bytes_remaining, ANSI_TEXTFAINT "               Created in '%s' on %d\n" ANSI_TEXTNORM,
			inputinfo->valuator[count]->my_object->file_lastmod,
			inputinfo->valuator[count]->my_object->line_lastmod);
		bytes_used += bytes; bytes_remaining -= bytes;
#endif
		if ((bytes < 1) || (bytes_remaining < 1)) { str[str_size] = '\0'; return str; }	/* for safety */
	}

	bytes = snprintf(&str[bytes_used], bytes_remaining, "Num 6-sensors: %d\n", inputinfo->num_6sensors);
	bytes_used += bytes; bytes_remaining -= bytes;
	if ((bytes < 1) || (bytes_remaining < 1)) { str[str_size] = '\0'; return str; }	/* for safety */
	for (count = 0; count < inputinfo->num_6sensors; count++) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, "  6-sensor %d: '%s' -- '%s' -- From Device: '%s'\n",
			count,
			inputinfo->sensor6[count]->my_object->name,
			inputinfo->sensor6[count]->my_object->desc_str,
			inputinfo->sensor6[count]->my_device->name);
		bytes_used += bytes; bytes_remaining -= bytes;
#ifdef SHOW_CONFIG_LOCATION
		bytes = snprintf(&str[bytes_used], bytes_remaining, ANSI_TEXTFAINT "               Created in '%s' on %d\n" ANSI_TEXTNORM,
			inputinfo->sensor6[count]->my_object->file_lastmod,
			inputinfo->sensor6[count]->my_object->line_lastmod);
		bytes_used += bytes; bytes_remaining -= bytes;
#endif
		if ((bytes < 1) || (bytes_remaining < 1)) { str[str_size] = '\0'; return str; }	/* for safety */
	}

	/* TODO: Nswitches  -- num_Nways/switchN */

	/* TODO: Nsensors -- num_Nsensors/sensorN */

	/* TODO: Controls -- num_controls/control */

	return str;	/* NOTE: This is redundant to the first argument, but can be used */
}


/*****************************************************************************/
/* list_self_controls(): A function that creates a string containing a list  */
/*   of all the self-controls within the active FreeVR system.  This string  */
/*   can then be printed or added to the help string.                        */
/*                                                                           */
/* The arguments are based on the vrSprintInputUI() function in vr_input.c.  */
/*   As with that, the "style" argument is currently ignored.                */
/*                                                                           */
/* NOTE: Unlike list_all_inputs(), this function has to make the list of     */
/*   controls by looping through all the input devices because the           */
/*   overarching vrInputs structure doesn't have that list.                  */
/*                                                                           */
/* TODO: Consider adding a robust version of this function as part of the    */
/*   overall vr_input.c code base.                                           */
/* TODO: Make use of the vrPrintStyle argument to control how much is output.*/
/*****************************************************************************/
char *list_self_controls(char *str, vrInputInfo *inputinfo, int str_size, vrPrintStyle style)
{
	int	bytes_remaining = str_size;
	int	bytes_used = 0;
	int	bytes;			/* return result from snprintf() calls -- usually # of bytes written */

	int	count;
	int	devcount;
	vrInput	*input;

	/* Warning: snprintf() returns number of bytes that *would* have been written, not actual! */

	bytes = snprintf(&str[bytes_used], bytes_remaining, "Self-controls:\n");
	bytes_used += bytes; bytes_remaining -= bytes;
	if ((bytes < 1) || (bytes_remaining < 1)) {
		/* For safety make sure there's a trailing NIL */
		str[str_size] = '\0';

		/* For clarity, add some elipses & a newline */
		str[str_size-1] = '\n';
		str[str_size-2] = str[str_size-3] = str[str_size-4] = '.';
		return str; /* NOTE: This is redundant to the first argument, but can be used */
	}

	for (devcount = 0; devcount < inputinfo->num_input_devices; devcount++) {
		/* NOTE: We could list the inputs for each input device here, */
		/*   but as the default (and only choice) mapping includes    */
		/*   all inputs, such a list would be redundant.              */
		input = inputinfo->input_devices[devcount]->inputs; /* not sure how to get the count -- other than a NULL terminal on the list */

		/* List the self-controls for each input device. */
		input = inputinfo->input_devices[devcount]->self_controls;
		for (count = 0; count < inputinfo->input_devices[devcount]->num_scontrols; count++) {
#if 0 /* 1 is somewhat more verbose */
			bytes = snprintf(&str[bytes_used], bytes_remaining, "  Device[%d] (%s): self-control %d: '%s' -- '%s' [Defined at %s:%d]\n", devcount, inputinfo->input_devices[devcount]->name, count, input->name, input->desc_str, input->file_lastmod, input->line_lastmod);
#elif 1 /* 1  here (and 0 above) is medium verbose */
			bytes = snprintf(&str[bytes_used], bytes_remaining, "  Device[%d] (%s): self-control %d: '%s' -- '%s'\n", devcount, inputinfo->input_devices[devcount]->name, count, input->name, input->desc_str);
#else
			bytes = snprintf(&str[bytes_used], bytes_remaining, "  %s self-control: '%s' from '%s'\n", inputinfo->input_devices[devcount]->name, input->name, input->desc_str);
#endif
			bytes_used += bytes; bytes_remaining -= bytes;
#ifdef SHOW_CONFIG_LOCATION
			bytes = snprintf(&str[bytes_used], bytes_remaining, ANSI_TEXTFAINT "               Created in '%s' on %d\n" ANSI_TEXTNORM,
				input->file_lastmod,
				input->line_lastmod);
			bytes_used += bytes; bytes_remaining -= bytes;
#endif

			/* wrap up the string to make sure it's Kosher */
			if ((bytes < 1) || (bytes_remaining < 1)) {
				/* for safety make sure there's a trailing NIL */
				str[str_size] = '\0';

				/* for clarity, add some elipses & a newline */
				str[str_size-1] = '\n';
				str[str_size-2] = str[str_size-3] = str[str_size-4] = '.';
				return str;	/* NOTE: the return value isn't actually used. */
			}

			input = input->next;
		}
	}

	return str;	/* NOTE: This is redundant to the first argument, but can be used */
}

