/* ======================================================================
 *
 *  CCCCC          vr_input.vruidd.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: October 16, 2009
 * CC   CC         Last Modified: October 29, 2021
 *  CCCCC
 *
 * Code file for FreeVR inputs from the VruiDD input device.
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************
USAGE:
	The configuration of the inputs is similar to all input devices:

	Inputs are specified with the "input" option:
		input "<logical-name>" = "<freevr-type>(<device-name>:<VRPN-type>[<input-num>])";

		input "<logical-name>" = "2switch(button[<number>])";
		input "<logical-name>" = "valuator(valuator[<number>])";
		input "<logical-name>" = "6sensor({tracker|6sensor}[<number>])";
	  XX	input "<logical-name>" = "Nswitch(switch[<number>])"; -- The VruiDD has no concept of N-switch
	  XX	input "<logical-name>" = "Nsensor(glove[<number>])"; -- The VruiDD has no concept of N-sensor

	Controls are specified with the "control" option:
		control "<control option>" = "2switch(...)";
		control "<control option>" = "valuator(...)"; -- NOTE: no valuator oriented controls yet available for this device

	Here are the available control options for FreeVR:
		"system_pause_toggle" -- toggle the system pause flag
		"print_help" -- print info on how to use the input device
		"print_context" -- print the overall FreeVR context data structure (for debugging)
		"print_config" -- print the overall FreeVR config data structure (for debugging)
		"print_input" -- print the overall FreeVR input data structure (for debugging)
		"print_devinfo" -- print the internal VruiDD data structure (for debugging)

	Here are the FreeVR configuration argument options for the VruiDD:
		"host" - machine running the Vrui VRDeviceDaemon.
			("localhost" is the default)
		"port" - socket port of server serving the tracking data.
			(8555 is the default)
		"protocol" - the version of the Vrui Protocol in use
		"valScale" - scaling factor used to tune the valuator input range
		"trackScale"/"scale" - set the scaling factor for the location
			of the 6-sensor inputs.  [Set to convert inches to feet
			by default.]


	When using with Vive Lighthouse tracking:
		% .../RunViveTrackerNoHMD.sh


VRUIDD PROTOCOL NOTES:
	The VruiDD coordinate system can be configured to be oriented in any
	direction, and use any unit of measure.  However, the defacto standard
	for Vrui coodinates is to use a Z-Up orientation with unit inches as
	the translational measure.  NOTE: FreeVR uses a Y-Up orientation!

	The protocol is "documented in code" in the Vrui source, which can
	make it difficult to discern what's going on.  Also, Oliver Kreylos
	has begun changing the protocol between Vrui 1.x and Vrui 2.x versions
	of the Vrui library.

	Protocol 0 -- Vrui-1.x:
		APP: connect to socket
		APP: send 0x00 0x00 (CONNECT_REQUEST)
		DD: send 0x01 0x00 (CONNECT_REPLY) + 12 bytes w/ num trackers, buttons & valuators
		APP: send 0x03 0x00 (ACTIVATE_REQUEST)
		APP: send 0x07 0x00 (STARTSTREAM_REQUEST)
		DD: send continuous packets of 0x06 0x00 (PACKET_REPLY) + N bytes w/ input data
		APP: send 0x08 0x00 (STOPSTREAM_REQUEST)
		DD: send 0x06 0x00 (PACKET_REPLY) + N bytes w/ input data
		APP: send 0x04 0x00 (DEACTIVATE_REQUEST)
		APP: send 0x02 0x00 (DISCONNECT_REQUEST)
		DD: send 0x09 0x00 (STOPSTREAM_REPLY)
		APP: disconnect from socket

		(where N = 1 * <num buttons> + 4 * <num valuators> + 4*13 * <num trackers>)

		The tracker values are somehow defined in "Vrui/VRDeviceState.h",
		but I can't quite figure out how 7 values are contained in the
		"PositionOrientation" field.

	Protocol 1 -- Vrui-2.x:
		APP: connect to socket
		APP: send 0x00 0x00 (CONNECT_REQUEST)
		APP: send 0x01 0x00 0x00 0x00 (ie. the numeric code of this protocol -- "1")
		DD: send 0x01 0x00 (CONNECT_REPLY) + 4 bytes protocol value + 12 bytes w/ num trackers, buttons & valuators
		APP: send 0x03 0x00 (ACTIVATE_REQUEST)
		APP: send 0x07 0x00 (STARTSTREAM_REQUEST)
		DD: send continuous packets of 0x06 0x00 (PACKET_REPLY) + N bytes w/ input data
		APP: send 0x08 0x00 (STOPSTREAM_REQUEST)
		DD: send 0x06 0x00 (PACKET_REPLY) + N bytes w/ input data
		APP: send 0x04 0x00 (DEACTIVATE_REQUEST)
		APP: send 0x02 0x00 (DISCONNECT_REQUEST)
		DD: send 0x09 0x00 (STOPSTREAM_REPLY)
		APP: disconnect from socket

	So in the end it turns out the only difference between Protocol-0 and
	Protocol-1 is that there is a two-way exchange of 4 bytes to give a
	protocol number!


	Protocol 2 -- Vrui-3.x:

	Vrui version 3.0-001 introduced a new protocol for its VRDeviceDaemon:
	protocol-2!  It is mostly the same as protocol-1, except that it has
	a segment in the "header" that enumerates some virtual devices and
	specifies a mapping of the physical inputs to the virtual devices
	(which can be thought of as "props").  Once the mapping is read, the
	rest of protocol-2 is identical to protocol-1.

	So the only difference is the amount of data that is returned as
	part of the CONNECT_REPLY (0x01 0x00) packet:
		DD: send 0x01 0x00 (CONNECT_REPLY) + 4 bytes protocol value + 12 bytes w/ num trackers, buttons & valuators
			+ 4 bytes w/ number of virtual devices
	this is followed by a sequence of bytes with a length that can only
	be determined by parsing the data:
			- stringSequence -- name of virtual device
			- 4-byte integer -- trackType (0=None, 1=3D, 3=Ray, 7=6D)
			- 3 * 4-byte floats -- rayDirection X,Y,Z
			- 4-byte number -- it's always been zero, so I have no idea what it represents
			- 4-byte integer -- trackerIndex (-1 when trackType==None)
			- 4-byte integer -- numButtons
			- numButtons * stringSequence -- name of buttons
			- numButtons * 4-byte integer -- index of buttons
			- 4-byte integer -- numValuators
			- numValuators * stringSequence -- name of valuators
			- numValuators * 4-byte integer -- index of valuators

		where "stringSequence" is a 4-byte integer followed by ASCII bytes w/o a trailing '\0'

HISTORY:
	16-20 October 2009 (Bill Sherman)
		Wrote initial version using the example code from
		  vr_input.skeleton.c.

	28-29 October 2009 (Bill Sherman)
		Corrected a couple of issues with the 6-sensor configuration
		  options (using the "r2e" value, plus getting the quaternion
		  orientation correct).  Added the "transscale" configuration
		  option.

	18 August 2010 (Bill Sherman)
		Changed the allocation for the inputs to use the '0' version
		  of vrShmemAlloc0() so each will start with a known (0) state.
		  Also a slight improvement to _VruiDDPrintStruct().  Plus
		  fixed a bug whereby the last button wasn't being updated.
		[01/22/2021: renamed to _VruiDDPrintPrivateInfo()]

	19 July 2011 (Bill Sherman)
		Fixed a simple, but massive bug where sizeof() was called on
		  a math expression rather than a variable.

	30 March - 2 April 2012 (Bill Sherman)
		Added the ability to handle the "new" Vrui VRDeviceDaemon
		  protocol -- which in the end was just an exchange of protocol
		  numbers.

	14 September 2013 (Bill Sherman)
		I am making use of the newly created SELFCTRL_DBGLVL for all
		  the self-control outputs.  I then added some reasonable output
		  to the "print_help" callback to print out what all the device's
		  inputs are doing.

	15 September 2013 (Bill Sherman)
		Changed the "0x%p" format to the improved "%#p" format.

	16-17 September 2013 (Bill Sherman)
		I reverse engineered the new Vrui VRDeviceDaemon protocol-2.
		  As part of this I just read the new data in the header, print
		  it, and then ignore it -- which will still work both for the
		  standalone "vruiddtest" program and for FreeVR input parsing.
		  What it doesn't yet do is allow me to configure the FreeVR
		  input to take advantage of the virtual devices.

		NOTE: it's possible that the new protocol mimics in some way
		  what Oliver was doing with the record and playback input
		  scheme.

		I added command line argument parsing for "vruiddtest"
		  (-p, -prot, -nodata -- and host requires no switch).

	26 November 2020 (Bill Sherman)
		I did some cleanup work to match this file better with the
		  "vr_input.skeleton.c" file (along with "vr_input.vrpn.c").

	22 January 2021 (Bill Sherman)
		Minor rename of _VruiDDPrintStruct() to _VruiDDPrintPrivateInfo(),
		  and "print_struct" control command to "print_devinfo".

	20 April 2021 (Bill Sherman)
		Duplicating the changes to vr_input.vrpn.c performed earlier
		  this month, primarily focussed on the "vruiddtest" "-screen"
		  option.  In particular:

		Adopted new ANSI #defines from vr_debug.h (even the code that's
		  #if'd out because it's just example code from other files.)

		Added the top/bottom margins for debugging output.  Note that to
		  place debugging output into the proper location on the screen,
		  surround the output-producing code with:
			vrPrintf(ANSI_CURSORLOC_RETURN);  // before debug output
			vrPrintf(ANSI_CURSORLOC_SAVE);    // after debug output

		I added the button and valuator outputs.

		I added the Euler and Quaternion outputs.  And I set a
		  reasonable max value for translations, so now we see the X/Y/Z
		  actually move away from the zero line.

		I added a bottom line marker (not yet done in the VRPN code).  I
		  also added a top, 1/4 & 3/4 down lines to help judge rotations.

		[corresponds to Freevr Version 0.6g]

	04 May 2021 (Bill Sherman)
		Did a clean-up of some of the old code.  There's still a little
		  code for testing left, but clearing it a little at a time.

	5 May 2021 (Bill Sherman)
		Minor clean-up to remove sprintf() and make sizeof() safer.

		[corresponds to Freevr Version 0.7b]

	27 October 2021 (Bill Sherman)
		Added the 6-sensor initial position ("ipos") code to place
		  them at a reasonable default location prior to making a
		  network connection.
		Also did some general clean-up and change to more closely
		  match other position tracker input devices.

		[corresponds to Freevr Version 0.7c]

	29 October 2021 (Bill Sherman)
		Copied from vr_input.vrpn.c, the new out-of-bounds check (at
		  the top) for the _RenderValuesScreen() for the X,Y,Z values.

		[corresponds to Freevr Version 0.7c]

	6 June 2022 (Bill Sherman)
		I renamed the "transScale" that applied to the overall tracking
		  to now be "trackScale".
		[corresponds to Freevr Version 0.7d]

TODO:
	- Add a "-csv" rendering style option to "vruiddtest"

	- Have "vruiddtest" return a success/failure value when terminating
		(either because a limited-date option was given, or the
		application was terminated).  This should be based on whether
		a successful connection to a server was established.

	- Determine why "vruiddtest" isn't advancing to protocol-1 when it
		fails to connect with protocol-0 -- the man page says it should.

	- Determine why "vruiddtest" only partially quits when getting the
		two-button chord quit signal.

	- Get battery information of Vive controllers if possible

	- DONE: Make a more reasonable screen-rendering routine.

	- integrate the protocol-2 virtual devices into the FreeVR input
		system.

		- as part of this, I'll probably want to add some socket
			stream data parsing routines that 1) checks that there
			is enough data in the buffer, read the data out
			of the buffer and do the conversion to int/float/string,
			and then shift and shrink the buffer.

		- also as part of this, have the virtual devices reported as
			part of _VruiDDPrintStruct(), and therefore don't
			just print the values as they are parsed.
		        [01/22/2021: renamed to _VruiDDPrintPrivateInfo()]

		- and now Oliver's protocol is at like #5 or something, so need
			to investigate that.

	- The "vruiddtest" program always starts with these messages:
		_VruiDDInitializeDevice(): Got a bad read result.
		_VruiDDInitializeDevice(): Resource temporarily unavailable

		fix this!

	- should cleanly address the issue of data endianness over the socket.
		(Currently, if daemon and client run on same architecture, it
		should work fine.  [04/03/12: Actually, I'm not sure that's true.
		The parsing of numbers in _VruiDDInitializeDevice() assumes the
		bytes are in little-endian order.  Reading the rest of the data
		should be fine on same-architectured machines, but this part
		not.]

	- If unable to connect to the daemon right away, would be nice to
		keep checking and connect when possible.

	- DONE: Add command-line-arguments for "vruiddtest":
		* host (no switch needed)
		* port
		* protocol
		* nodata -- just read the header (print info) and quit

	- DONE: write the man page for "vruiddtest"

**************************************************************************/
#if !defined(FREEVR) && !defined(TEST_APP) && !defined(CAVE)
/* Choose how this code should be compiled (ie. for CAVE, standalone, etc). */
#  define	FREEVR
#  undef	TEST_APP
#  undef	CAVE
#endif

#undef COMM_DEBUG	/* define this to add a lot of output for debugging communications w/ the device */


/* Header includes needed for the both the FreeVR library, and test utility */
#include <stdio.h>		/* needed for printf & all other standard I/O */
#include <string.h>		/* needed for strcmp(), strdup() & strchr() */
#include <stdint.h>		/* needed for uint16_t & uint32_t types */
#include <fcntl.h>		/* for fcntl() to set socket parameters */

#include "vr_socket.h"		/* for vrSocketCall() & vrSocketClose() */
#include "vr_utils.h"		/* for vrSleep() & vrAtoI() */
#include "vr_debug.h"


#if defined (FREEVR)
/* Header includes needed for the FreeVR library, but not the test utility */
#  include "vr_math.h"		/* for VR_{X,Y,Z,W,AZIM,ELEV,ROLL} */
#  include "vr_input.h"
#  include "vr_input.opts.h"
#  include "vr_parse.h"
#  include "vr_shmem.h"
#endif


#if defined(TEST_APP)
/* Header includes needed just for the "vruiddtest" utility */
/* NOTE: this includes some FreeVR *source code* */
#  include <signal.h>		/* for signal() */
#  include <stdlib.h>		/* needed for malloc(), getenv() & atoi() */
#  include <sys/time.h>		/* needed for gettimeofday() */
#  include "vr_socket.c"	/* for vrSocketCall() & vrSocketClose() code */
#  include "vr_enums.h"
#  include "vr_utils.c"		/* needed for vrSleep() */
#  define MATH_ONLY		/* exclude the FreeVR material in vr_math.c  */
#  include "vr_math.h"		/* for VR_{X,Y,Z,W,AZIM,ELEV,ROLL} */
#  include "vr_math.c"		/* needed for Quaternion to Euler conversion */
#  define vrShmemAlloc0 malloc
#endif


#if defined(CAVE) /* { */
/* Header includes needed just for interfacing with the CaveLib */
/* NOTE: It's been over a decade since this has been tested     */
#  include <math.h>
#  include "cave.h"
#  include "cave.private.h"
#  include "cave.tracker.h"
static CAVE_SENSOR_ST initial_cave_sensor = {
		0,5,0,
		0,0,0,
		0,	/*timestamp*/
		FALSE,	/*calibrated*/
		CAVE_TRACKER_FRAME
	};
#endif /* } CAVE */


	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/

#define	BUFSIZE		1024

#define	DEFAULT_HOST	"localhost"
#define	DEFAULT_PORT	8555


/** VruiDD command sequences **/
enum MessageId {
		CONNECT_REQUEST,        /* Request to connect to server (0x00 0x00) */
		CONNECT_REPLY,          /* Positive connect reply with server layout (0x01 0x00) */
		DISCONNECT_REQUEST,     /* Polite request to disconnect from server (0x02 0x00) */
		ACTIVATE_REQUEST,       /* Request to activate server (prepare for sending packets) (0x03 0x00) */
		DEACTIVATE_REQUEST,     /* Request to deactivate server (no more packet requests) (0x04 0x00) */
		PACKET_REQUEST,         /* Requests a single packet with current device state (0x05 0x00) */
		PACKET_REPLY,           /* Sends a device state packet (0x06 0x00) */
		STARTSTREAM_REQUEST,    /* Requests entering stream mode (server sends packets automatically) (0x07 0x00) */
		STOPSTREAM_REQUEST,     /* Requests leaving stream mode (0x08 0x00) */
		STOPSTREAM_REPLY        /* Server's reply after last stream packet has been sent (0x09 0x00) */
	};

/** VruiDD bit-masks & indices **/
/* these are often used to decode specific bits that map to individual buttons */
#define VruiDD_BUTTON_x1x	0x01
#define VruiDD_BUTTON_x2x	0x02
#define VruiDD_BUTTON_x3x	0x04

#define VruiDD_BUTTONINDEX_x1x	0x00
#define VruiDD_BUTTONINDEX_x2x	0x01
#define VruiDD_BUTTONINDEX_x3x	0x02

/* VruiDD sensitivity values */
#define TRANS_SENSITIVITY	0.001
#define ROT_SENSITIVITY		0.02
#define VALUATOR_SENSITIVITY	1.0


/****************************************************************/
/*** auxiliary structure of the current data from the device. ***/
typedef struct {
		/* these are for interfacing with the daemon */
		int		fd_socket;		/* communication file descriptor */
		char		*dd_host;		/* name of daemon's host */
		int		dd_port;		/* port on host listening for application connections */
		int		open;			/* flag with VruiDD successfully open */

		/* these are for internal data parsing */
		unsigned char	buf[BUFSIZE];		/* the incoming data stream */
		int		eobuf_pos;		/* the last read byte in the buffer */
#if 0
		unsigned char	packtype;
		int		packlen;
#endif
		char		version[256];		/* self-reported version of the device */
		char		op_params[256];		/* operating parameters of the device (according to it) */
		int		protocol;		/* An integer indicating the Vrui VRDeviceDaemon protocol */

		/* information about the current values */
		int		packet_size;		/* the number of bytes in each packet reported by the Daemon */
		int		num_dd_buttons;		/* number of buttons reported by the Vrui Device Daemon */
		int		*incoming_buttons;	/* button data coming from the device daemon -- allocated in device initialization */
		int		*incoming_buttons_prev;	/* previous values for button data from the device daemon -- allocated in device initialization */
		int		num_dd_valuators;	/* number of valuator reported by the Vrui Device Daemon */
		double		*incoming_valuators;	/* valuator data coming from the device daemon -- allocated in device initialization */
		double		*incoming_valuators_prev;/* previous values for valuator data from the device daemon -- allocated in device initialization */
		int		num_dd_trackers;	/* number of position trackers reported by the Vrui Device Daemon */
		float		*incoming_trackers[13];	/* each incoming position-tracker has 13 floats */

		int		num_dd_virtdevs;	/* number of virtual devices sent by the Vrui Device Daemon */

#if 0 /* possible other items to handle (from vr_input.skeleton.c) */
		int		timer;			/* incoming timer info  TODO: or is it a checksum?*/
		int		button_change;		/* boolean indicator if button values have changed (may be an array if each change is independent) */
		int		valuator_change;	/* boolean indicator of change in valuator values (ditto) */
		int		receiver_change;	/* boolean indicator of change in 6sensor values (ditto) */
#endif

		/* information about how to filter the data for use with the VR library (FreeVR or CAVE) */
		float		scale_valuator;		/* scaling factor for valuators */
		float		scale_trans;		/* multiplier to scale from the VruiDD location units (inches) */

		/* information about the inputs and mappings */
#ifdef CAVE
		/* CAVE specific fields here */
		/* ... */

#elif defined(FREEVR)
		/* FREEVR specific fields here */

		/* 2switch/button mappings */
		int		num_buttons;		/* it is often wise to store the number of button inputs in addition to knowing the maximum possible */
		vr2switch	**button_inputs;	/* the maximum number of buttons is devinfo->num_2ways + devinfo->num_scontrols (all the self-controls are not necessarily buttons) */
		int		*button_map_button;	/* An array of numbers indicating the particular button */
		int		num_switches;
		vrNswitch	**switch_inputs;

		/* Valuator/analog mappings */
		int		num_valuators;		/* it is often wise to store the number of valuator inputs in addition to knowing the maximum possible */
		vrValuator	**valuator_inputs;	/* the maximum number of valuators is devinfo->num_valuators + devinfo->num_scontrols (all the self-controls are not necessarily valuators) */
		int		*valuator_map_valuator;	/* An array of numbers indicating the particular valuator */
		float		*valuator_map_sign;	/* An array of numbers indicating the sign of each valuator input */

		/* 6-sensor/tracker mappings */
		int		num_6sensors;		/* it is often wise to store the number of 6-sensor inputs in addition to knowing the maximum possible */
		vr6sensor	**sensor6_inputs;
		int		*tracker_map_tracker;	/* An array of numbers indicating the particular tracker */
		int		num_Nsensors;
		vrNsensor	**sensorN_inputs;

#endif /* end library-specific fields */

	} _VruiDDPrivateInfo;



	/*********************************************/
	/*** General NON public interface routines ***/
	/*********************************************/

/*************************************************************************/
/* _VruiDDInitializeStruct(): sets initial values in the device specific */
/*   internal structure to prepare for making the initial connection.    */
/*   For some devices, memory will be allocated for arrays of values.    */
/*                                                                       */
/* NOTE: Some (few) versions of this routine take a second "typename"    */
/*   argument that can be used to specify a particular device among many */
/*   that share (more or less) the same protocol.  The typename is then  */
/*   used to determine what specific features are available with this    */
/*   particular type of device.                                          */
static void _VruiDDInitializeStruct(_VruiDDPrivateInfo *aux, char *typename)
{
	aux->version[0] = '\0';
	aux->op_params[0] = '\0';

	aux->buf[0] = '\0';
	aux->eobuf_pos = 0;

	aux->protocol = 0;

#if 0 /* possible other items to handle (from vr_input.skeleton.c) */
	aux->timer = 0;
	aux->button_change = 0;
	aux->valuator_change = 0;
	aux->receiver_change = 0;
#endif

	aux->scale_valuator = VALUATOR_SENSITIVITY;	/* set the default valuator scaling factor */
	aux->scale_trans = 1.0/12.0;			/* convert from inches to feet by default */

	/* everything else is zero'd by default */
}


/**************************************************************************/
/* _VruiDDPrintPrivateInfo(): outputs the contents of the device specific */
/*   internal structure to help with configuration and system debugging.  */
/*   Output verbosity can be controlled by the "style" argument.          */
static void _VruiDDPrintPrivateInfo(FILE *file, _VruiDDPrivateInfo *aux, vrPrintStyle style)
{
	int		count;		/* loop counter */
	int		ptv_count;

	vrFprintf(file, "VruiDD input-device internal structure (%p):\n", aux);
	vrFprintf(file, "\r\tversion -- '%s'\n", aux->version);
	vrFprintf(file, "\r\toperating parameters -- '%s'\n", aux->op_params);
	vrFprintf(file, "\r\tfd_socket = %d\n\thost = '%s'\n\tport = %d\n\topen = %d\n",
		aux->fd_socket,
		aux->dd_host,
		aux->dd_port,
		aux->open);

	/* print the filter values */
	vrFprintf(file, "\r\tscale_valuator = %f\n", aux->scale_valuator);
	vrFprintf(file, "\r\tscale_trans = %f\n", aux->scale_trans);

	/* print the raw values */
	vrFprintf(file, "Buttons (%d): ", aux->num_dd_buttons);
	for (count = 0; count < aux->num_dd_buttons; count++) {
		vrFprintf(file, "%c", (aux->incoming_buttons[count] == 0 ? '.' : 'X'));
	}
	vrFprintf(file, "\n");

	vrFprintf(file, "Valuators (%d): ", aux->num_dd_valuators);
	for (count = 0; count < aux->num_dd_valuators; count++) {
		vrFprintf(file, "%6.3f ", aux->incoming_valuators[count]);
	}
	vrFprintf(file, "\n");

	vrFprintf(file, "Position-Trackers (%d): ", aux->num_dd_trackers);
	for (count = 0; count < aux->num_dd_trackers; count++) {
		vrFprintf(file, "\n\tPT %d: ", count);
		for (ptv_count = 0; ptv_count < (count == 0 ? 13:3); ptv_count++) {	/* TODO: s/b 13, but doing 3 for now */
			vrFprintf(file, "%6.3f ", aux->incoming_trackers[ptv_count][count]);
		}
	}
	vrFprintf(file, "\n");

#ifdef FREEVR /* { */
	/************************************/
	/** print the FreeVR input objects **/

	/* button inputs */
	vrFprintf(file, "\r\tbutton inputs:\n");
	for (count = 0; count < aux->num_buttons; count++) {
		vrFprintf(file, "\r\t\tbutton_input[%d] = %#p%s\n", count, aux->button_inputs[count], ((aux->button_inputs[count]->input_type == VRINPUT_CONTROL) ? " (*control*)" : ""));
	}

	/* valuator inputs */
	vrFprintf(file, "\r\tvaluator inputs:\n");
	for (count = 0; count < aux->num_valuators; count++) {
		vrFprintf(file, "\r\t\tvaluator_input[%d] = %#p%s\n", count, aux->valuator_inputs[count], ((aux->valuator_inputs[count]->input_type == VRINPUT_CONTROL) ? " (*control*)" : ""));
	}

	/* 6sensor inputs */
	/* TODO: print position-tracker values */
#if 0 /* NOT YET IMPLEMENTED */
	vrFprintf(file, "\r\t6sensor inputs (active = %d):\n", aux->active_sim6sensor);
	for (count = 0; count < aux->num_6sensors; count++) {
		vrFprintf(file, "\r\t\t6sensor_inputs[%d] = %#p\n", count, aux->sensor6_inputs[count]);
	}
#endif

#  if 0 /* TODO: not sure if we want/need this */
	vrFprintf(file, "\r\tcontrol inputs:\n");
	for (count = 0; count < MAX_CONTROLS; count++) {
		vrFprintf(file, "\r\t\tcontrol_inputs[%d] = %#p\n", count, aux->control_inputs[count]);
	}
#  endif
#endif /* } FREEVR */

	/* TODO: print some info about the current values */
}


/****************************************************************************/
/* _VruiDDPrintHelp(): outputs the mappings from the physical device to the */
/*    virtual inputs and controls to help the user understand the interface.*/
static void _VruiDDPrintHelp(FILE *file, _VruiDDPrivateInfo *aux)
{
	int		count;			/* loop counter */

#if 0 || defined(TEST_APP)
	vrFprintf(file, BOLD_TEXT "Sorry, VruiDD - print_help control not yet implemented.\n" NORM_TEXT);
#else
	vrFprintf(file, BOLD_TEXT "VruiDD - inputs:" NORM_TEXT "\n");
	for (count = 0; count < aux->num_buttons; count++) {
		if (aux->button_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->button_inputs[count]->my_object->desc_str,
				(aux->button_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->button_inputs[count]->my_object->name);
		}
	}
	for (count = 0; count < aux->num_valuators; count++) {
		if (aux->valuator_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->valuator_inputs[count]->my_object->desc_str,
				(aux->valuator_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->valuator_inputs[count]->my_object->name);
		}
	}
	for (count = 0; count < aux->num_6sensors; count++) {
		if (aux->sensor6_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->sensor6_inputs[count]->my_object->desc_str,
				(aux->sensor6_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->sensor6_inputs[count]->my_object->name);
		}
	}
#endif
}


/***************************************************************************/
/* _VruiDDReadInput(): function reads data from the socket, and then looks */
/*   for and parses packets from the daemon.  The data is then placed into */
/*   the generic portion of the "_VruiDDPrivateInfo" structure.            */
/* NOTE: _VruiDDReadInput() is placed above the __VruiDDInitializeDevice() */
/*   function to allow the latter to call the former.                      */
/* Returns number of packets read this time (-1 indicates reception of the */
/*   STOPSTREAM_REPLY message)                                             */
static int _VruiDDReadInput(_VruiDDPrivateInfo *aux)
{
	uint16_t	message;		/* VRUI VRDeviceDaemon messages are 2 byte sequences */
	int		read_result;		/* return value of the call to "read()" */
	int		packets_decoded = 0;	/* return the number of packets that are decoded */
	int		decode_pos = 0;		/* position in the buffer currently being decoded */
	uint8_t		bool_value;		/* place holder for any button data being processed */
	float		float_value;		/* place holder for any valuator or sub-tracker data being processed */
	int		count;			/* loop counter */
	int		ptv_count;		/* loop counter for position-tracker values */

	/*** read available data into available buffer space ***/
	do {
		read_result = (ssize_t)read(aux->fd_socket, &aux->buf[aux->eobuf_pos], sizeof(aux->buf) - aux->eobuf_pos);
		if (read_result > 0) {
			aux->eobuf_pos += read_result;
#ifdef COMM_DEBUG
			vrDbgPrintf("_VruiDDReadInput(): %d bytes currently in buffer.\n", aux->eobuf_pos);
#endif
		}
	} while (read_result > 0);
#ifdef COMM_DEBUG
if (aux->eobuf_pos > 0)
vrPrintf("yoyo about to decode: eobuf_pos = %d, packet_size = %d\n", aux->eobuf_pos, aux->packet_size);
#endif

	/*******************************************/
	/*** decode data contained in the buffer ***/
	while (aux->eobuf_pos >= 2) {
		message = (aux->buf[0] + (aux->buf[1] << 8));

		if (message == PACKET_REPLY) {
			/* decode only if we have enough data for at least one packet */
			if (aux->eobuf_pos >= aux->packet_size) {
				decode_pos = 2;		/* start decoding right after the Daemon message packet */

				/* loop over all the position-trackers */
				for (count = 0; count < aux->num_dd_trackers; count++) {
					for (ptv_count = 0; ptv_count < 13; ptv_count++) {
						((char *)(&float_value))[0] = aux->buf[decode_pos+0];		/* NOTE: dangerous assumption of 4-byte floats here */
						((char *)(&float_value))[1] = aux->buf[decode_pos+1];
						((char *)(&float_value))[2] = aux->buf[decode_pos+2];
						((char *)(&float_value))[3] = aux->buf[decode_pos+3];

						aux->incoming_trackers[ptv_count][count] = float_value;
						decode_pos += 4 * sizeof(char);	/* this is explicitly what we're parsing */
					}
				}

				/* loop over all the buttons */
				for (count = 0; count < aux->num_dd_buttons; count++) {
					bool_value = aux->buf[decode_pos+0];

					aux->incoming_buttons[count] = bool_value;
					decode_pos += sizeof(bool_value);
				}

				/* loop over all the valuators */
				for (count = 0; count < aux->num_dd_valuators; count++) {
					((char *)(&float_value))[0] = aux->buf[decode_pos+0];			/* NOTE: dangerous assumption of 4-byte floats here */
					((char *)(&float_value))[1] = aux->buf[decode_pos+1];
					((char *)(&float_value))[2] = aux->buf[decode_pos+2];
					((char *)(&float_value))[3] = aux->buf[decode_pos+3];

					aux->incoming_valuators[count] = float_value;
					decode_pos += 4 * sizeof(char);	/* this is explicitly what we're parsing */
				}

				memmove(aux->buf, &aux->buf[aux->packet_size], (aux->eobuf_pos-aux->packet_size));	/* shift the data by one packet */
				aux->eobuf_pos -= aux->packet_size;
				packets_decoded++;
			} else {
				break;;		/* skip out of the while loop (or we get stuck) */
			}
		} else if (message == STOPSTREAM_REPLY) {
			memmove(aux->buf, &aux->buf[2], (aux->eobuf_pos-2));					/* shift the data by two bytes */
			aux->eobuf_pos -= 2;
			vrDbgPrintf("_VruiDDReadInput(): received a STOPSTREAM_REPLY message from the daemon -- stopping.\n");
			return (-1);
		} else {
			vrErrPrintf("_VruiDDReadInput(): " RED_TEXT "Expected a 'PACKET_REPLY' message, got %02x, skipping two bytes.\n" NORM_TEXT, message);
			memmove(aux->buf, &aux->buf[2], (aux->eobuf_pos-2));	/* shift the data by two bytes */
			aux->eobuf_pos -= 2;
#if defined(COMM_DEBUG) || 1
			vrDbgPrintf("_VruiDDReadInput(): %d bytes currently in buffer.\n", aux->eobuf_pos);
{ int count;
for (count = 0; count < aux->eobuf_pos; count++)
	vrPrintf("%02x ", aux->buf[count]);
vrPrintf("\n");
}
#endif
		}
	}

#ifdef COMM_DEBUG
if (packets_decoded > 0)
vrPrintf("_VruiDDReadInput(): %d packets decoded this call.\n", packets_decoded);
#endif
	return (packets_decoded);
}


/**********************************************************/
/* _VruiDDInitializeDevice() is called in the OPEN phase  */
/*   of input interface -- after the types of inputs have */
/*   been determined (during the CREATE phase).           */
static int _VruiDDInitializeDevice(_VruiDDPrivateInfo *aux)
{
	uint16_t	message;		/* VRUI VRDeviceDaemon messages are 2 byte sequences */
	uint32_t	value;			/* VRUI 4-byte value */
	int		read_result;		/* return value of the call to "read()" */
	int		bytes_read = 0;		/* number of bytes read from socket (the Vrui Daemon) thus far */
	int		opening_bytes = 14;	/* the number of bytes to expect in the CONNECT_REPLY response */
	int		response_protocol;	/* the value the VRDeviceDaemon reports as the used protocol */
	int		count;			/* loop counter */
static	int		bad_reads = 0;		/* bad read counter -- used to try a jump-start after 5 bad reads */

	if (aux == NULL) {
		vrErrPrintf("_VruiDDInitializeDevice(): " RED_TEXT "Warning, no auxiliary data for VruiDD.\n" NORM_TEXT);
		return -1;
	}

	if (fcntl(aux->fd_socket, F_SETFL, O_RDWR | O_NONBLOCK | O_NOCTTY) < 0) {
                vrErrPrintf("_VruiDDInitializeDevice(): An error occurred while trying to set the file controls.\n");
        }


	/***********************************************/
	/* send the opening hail message to the daemon */
	message = CONNECT_REQUEST;
	write(aux->fd_socket, &message, sizeof(message));	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	if (aux->protocol > 0) {
		value = aux->protocol;
		write(aux->fd_socket, &value, sizeof(value));	/* TODO: warning -- endianness */ /* this is protocol version info */
	}
	fsync(aux->fd_socket);

	/*****************************************************************/
	/* now await the data with information about the incoming data   */
	/* The first packet of information should be 14 bytes in length: */
	/*   - 2 bytes with 0x01 0x00 (CONNECT_REPLY)                    */
	/*   - 4 bytes with the protocol value response (prot > 0)       */
	/*   - 4 bytes with number of position trackers (aka 6-sensors)  */
	/*   - 4 bytes with number of buttons                            */
	/*   - 4 bytes with number of valuators                          */
	switch (aux->protocol) {
	case 0:
		/* The first packet of information should be 14 bytes in length: */
		/*   - 2 bytes with 0x01 0x00 (CONNECT_REPLY)                    */
		/*   - 4 bytes with number of position trackers (aka 6-sensors)  */
		/*   - 4 bytes with number of buttons                            */
		/*   - 4 bytes with number of valuators                          */
		opening_bytes = 14;
		break;
	case 1:
		/* The first packet of information should be 18 bytes in length: */
		/*   - 2 bytes with 0x01 0x00 (CONNECT_REPLY)                    */
		/*   - 4 bytes with the protocol value response                  */
		/*   - 4 bytes with number of position trackers (aka 6-sensors)  */
		/*   - 4 bytes with number of buttons                            */
		/*   - 4 bytes with number of valuators                          */
		opening_bytes = 18;
		break;
	case 2:
		/* The first packet of information should be 18 bytes in length: */
		/*   - 2 bytes with 0x01 0x00 (CONNECT_REPLY)                    */
		/*   - 4 bytes with the protocol value response                  */
		/*   - 4 bytes with number of position trackers (aka 6-sensors)  */
		/*   - 4 bytes with number of buttons                            */
		/*   - 4 bytes with number of valuators                          */
		/*   - 4 bytes with number of virtual devices                    */
		opening_bytes = 22;
		break;
	}

	/* wait until the entire opening bytes sequence is received */
	while (bytes_read < opening_bytes) {
		read_result = (ssize_t)read(aux->fd_socket, &aux->buf[aux->eobuf_pos], sizeof(aux->buf) - aux->eobuf_pos);
		if (read_result > 0) {
			bytes_read += read_result;
			aux->eobuf_pos += read_result;
			bad_reads = 0;
		} else if (read_result < 0) {
			/* I don't think this should happen with a working VruiDD */
			bad_reads++;
			vrDbgPrintf("_VruiDDInitializeDevice(): Got a bad read result (%d bytes read).\n", bytes_read);
			perror("_VruiDDInitializeDevice()");
			vrSleep(100000);

			if (bad_reads > 5) {
				vrDbgPrintf("_VruiDDInitializeDevice(): 5 bad reads -- attempting to send protocol-1 message.\n");
				bad_reads = 0;
				aux->protocol = 1;
				opening_bytes = 18;
				value = aux->protocol;
				write(aux->fd_socket, &value, sizeof(value));	/* TODO: warning -- endianness */ /* this is protocol version info */
				fsync(aux->fd_socket);
			}
		}

#ifdef COMM_DEBUG
		vrDbgPrintf("_VruiDDInitializeDevice(): %d bytes read thus far.\n", bytes_read);
#endif
	}

	/* verify expected opening response received (assuming LITTLE_ENDIAN for now) */
	message = (aux->buf[0] + (aux->buf[1] << 8));
	if (message == CONNECT_REPLY) {
		vrDbgPrintf("_VruiDDInitializeDevice(): Got a CONNECT_REPLY packet, will now decode.\n");

		switch (aux->protocol) {
		default:
		case 0:
			/* protocol == 0 */
			response_protocol = 0;

			/* number of 6-sensors */
			value = (aux->buf[2] + (aux->buf[3] << 8) + (aux->buf[4] << 16) + (aux->buf[5] << 24));
			aux->num_dd_trackers = value;

			/* number of buttons */
			value = (aux->buf[6] + (aux->buf[7] << 8) + (aux->buf[8] << 16) + (aux->buf[9] << 24));
			aux->num_dd_buttons = value;

			/* number of valuators */
			value = (aux->buf[10] + (aux->buf[11] << 8) + (aux->buf[12] << 16) + (aux->buf[13] << 24));
			aux->num_dd_valuators = value;

			break;

		case 1:
			/* protocol response */
			value = (aux->buf[2] + (aux->buf[3] << 8) + (aux->buf[4] << 16) + (aux->buf[5] << 24));
			response_protocol = value;
			vrPrintf("_VruiDDInitializeDevice(): Got a response protocol value of %d (requested %d).\n", response_protocol, aux->protocol);

			/* number of 6-sensors */
			value = (aux->buf[6] + (aux->buf[7] << 8) + (aux->buf[8] << 16) + (aux->buf[9] << 24));
			aux->num_dd_trackers = value;

			/* number of buttons */
			value = (aux->buf[10] + (aux->buf[11] << 8) + (aux->buf[12] << 16) + (aux->buf[13] << 24));
			aux->num_dd_buttons = value;

			/* number of valuators */
			value = (aux->buf[14] + (aux->buf[15] << 8) + (aux->buf[16] << 16) + (aux->buf[17] << 24));
			aux->num_dd_valuators = value;

			break;
		case 2:
			/* protocol response */
			value = (aux->buf[2] + (aux->buf[3] << 8) + (aux->buf[4] << 16) + (aux->buf[5] << 24));
			response_protocol = value;
			vrPrintf("_VruiDDInitializeDevice(): Got a response protocol value of %d (requested %d).\n", response_protocol, aux->protocol);

			/* number of 6-sensors */
			value = (aux->buf[6] + (aux->buf[7] << 8) + (aux->buf[8] << 16) + (aux->buf[9] << 24));
			aux->num_dd_trackers = value;

			/* number of buttons */
			value = (aux->buf[10] + (aux->buf[11] << 8) + (aux->buf[12] << 16) + (aux->buf[13] << 24));
			aux->num_dd_buttons = value;

			/* number of valuators */
			value = (aux->buf[14] + (aux->buf[15] << 8) + (aux->buf[16] << 16) + (aux->buf[17] << 24));
			aux->num_dd_valuators = value;

			/* number of virtual devices */
			value = (aux->buf[18] + (aux->buf[19] << 8) + (aux->buf[20] << 16) + (aux->buf[21] << 24));
			aux->num_dd_virtdevs = value;

			break;
		}

		vrDbgPrintf("_VruiDDInitializeDevice(): Daemon reporting %d trackers, %d buttons and %d valuators.\n", aux->num_dd_trackers, aux->num_dd_buttons, aux->num_dd_valuators);
		if (aux->protocol >= 2)
			vrDbgPrintf("_VruiDDInitializeDevice(): Daemon reporting %d virtual devices.\n", aux->num_dd_virtdevs);

		/* now shift the data in the buffer */
		memmove(aux->buf, &aux->buf[opening_bytes], (aux->eobuf_pos-opening_bytes));	/* shift the data by the opening_bytes */
		aux->eobuf_pos -= opening_bytes;

		/* For protocol=2, read the virtual devices information */
		if (aux->protocol == 2) {
			for (count = 0; count < aux->num_dd_virtdevs; count++) {
				char	virtdev_name[256];
				int	num_buttons;
				int	num_valuators;
				int	strcount;
				int	count2;
				float	float_value;	/* NOTE: dangerous assumption of 4-byte float */
				/* first 4-byte integer has a string length for the name of the virtual device */
#if 0 /* in theory, the data ought to already be in the buffer */
				do {
					read_result = (ssize_t)read(aux->fd_socket, &aux->buf[aux->eobuf_pos], sizeof(aux->buf) - aux->eobuf_pos);
					if (read_result > 0) {
						aux->eobuf_pos += read_result;
					}
				} while (read_result > 0);
#endif

				/* get the name of this virtual device */
				value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
				for (strcount = 0; strcount < value; strcount++)
					virtdev_name[strcount] = aux->buf[4+strcount];
				virtdev_name[strcount] = '\0';
				vrPrintf("virtual device %d is %s'%s'%s (%d)\n", count, BOLD_TEXT, virtdev_name, NORM_TEXT, value);
				memmove(aux->buf, &aux->buf[4+value], (aux->eobuf_pos-(4+value)));	/* shift the data by the string data */
				aux->eobuf_pos -= 4+value;

				/* first 4-byte integer is the trackerType code */
				value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
				vrPrintf("\tGot TrackerType code of %d (7=6D)\n", value);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				/* next 3 numbers are 4-byte floats with the rayDirection */
				((char *)(&float_value))[0] = aux->buf[0];		/* NOTE: dangerous assumption of 4-byte floats here */
				((char *)(&float_value))[1] = aux->buf[1];
				((char *)(&float_value))[2] = aux->buf[2];
				((char *)(&float_value))[3] = aux->buf[3];
				vrPrintf("\trayDirection = (%f", float_value);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				((char *)(&float_value))[0] = aux->buf[0];		/* NOTE: dangerous assumption of 4-byte floats here */
				((char *)(&float_value))[1] = aux->buf[1];
				((char *)(&float_value))[2] = aux->buf[2];
				((char *)(&float_value))[3] = aux->buf[3];
				vrPrintf(" %f", float_value);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				((char *)(&float_value))[0] = aux->buf[0];		/* NOTE: dangerous assumption of 4-byte floats here */
				((char *)(&float_value))[1] = aux->buf[1];
				((char *)(&float_value))[2] = aux->buf[2];
				((char *)(&float_value))[3] = aux->buf[3];
				vrPrintf(" %f)\n", float_value);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				/* next 4-byte sequence is unknown (always been zero thus far) */
				value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
				((char *)(&float_value))[0] = aux->buf[0];		/* NOTE: dangerous assumption of 4-byte floats here */
				((char *)(&float_value))[1] = aux->buf[1];
				((char *)(&float_value))[2] = aux->buf[2];
				((char *)(&float_value))[3] = aux->buf[3];
				vrPrintf("\tUnknown Value is %d or %f (always been zero thus far)\n", value, float_value);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				/* next 4-byte integer is the trackerIndex (will be -1 if there isn't one) */
				value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
				vrPrintf("\tGot trackerIndex of %d.\n", value);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				/* next 4-byte integer is the numButtons */
				num_buttons = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
				vrPrintf("\tGot numButtons value of %d\n", num_buttons);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				/* Now loop over the button names */
				for (count2 = 0; count2 < num_buttons; count2++) {
					value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
					for (strcount = 0; strcount < value; strcount++)
						virtdev_name[strcount] = aux->buf[4+strcount];
					virtdev_name[strcount] = '\0';
					vrPrintf("\t\tbutton %d is '%s' (%d)\n", count2, virtdev_name, value);
					memmove(aux->buf, &aux->buf[4+value], (aux->eobuf_pos-(4+value)));	/* shift the data by the string data */
					aux->eobuf_pos -= 4+value;
				}

				/* Now loop over the button indices */
				if (num_buttons > 0)
					vrPrintf("\tButton indices:\n");
				for (count2 = 0; count2 < num_buttons; count2++) {
					value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
					vrPrintf("\t\tButton %d is indexed to %d\n", count2, value);
					memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
					aux->eobuf_pos -= 4;
				}

				/* next 4-byte integer is the numValuators */
				num_valuators = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
				vrPrintf("\tGot numValuators value of %d\n", num_valuators);
				memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
				aux->eobuf_pos -= 4;

				/* Now loop over the valuator names */
				for (count2 = 0; count2 < num_valuators; count2++) {
					value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
					for (strcount = 0; strcount < value; strcount++)
						virtdev_name[strcount] = aux->buf[4+strcount];
					virtdev_name[strcount] = '\0';
					vrPrintf("\t\tvaluator %d is '%s' (%d)\n", count2, virtdev_name, value);
					memmove(aux->buf, &aux->buf[4+value], (aux->eobuf_pos-(4+value)));	/* shift the data by the string data */
					aux->eobuf_pos -= 4+value;
				}

				/* Now loop over the valuator indices */
				if (num_valuators > 0)
					vrPrintf("\tValuator indices:\n");
				for (count2 = 0; count2 < num_valuators; count2++) {
					value = (aux->buf[0] + (aux->buf[1] << 8) + (aux->buf[2] << 16) + (aux->buf[3] << 24));
					vrPrintf("\t\tValuator %d is indexed to %d\n", count2, value);
					memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos-4));	/* shift the data by the 4 bytes */
					aux->eobuf_pos -= 4;
				}
			}
		}

		/* set the packet size */
		aux->packet_size = sizeof(uint16_t) +				/* message code */
				(sizeof(float)*13 * aux->num_dd_trackers) +	/* A Vrui TrackerState struct is thirteen floats */
				(sizeof(uint8_t) * aux->num_dd_buttons) +	/* A Vrui button is one uint8_t (aka "bool" in C++) */
				(sizeof(float) * aux->num_dd_valuators);	/* A Vrui valuator is one float */
		vrDbgPrintf("_VruiDDInitializeDevice(): Daemon packet size is %d bytes.\n", aux->packet_size);

		/* now allocated the memory in the generic portion of the structure */
		aux->incoming_buttons = (void *)vrShmemAlloc0(sizeof(aux->incoming_buttons[0]) * aux->num_dd_buttons);
		aux->incoming_buttons_prev = (void *)vrShmemAlloc0(sizeof(aux->incoming_buttons_prev[0]) * aux->num_dd_buttons);
		aux->incoming_valuators = (void *)vrShmemAlloc0(sizeof(aux->incoming_valuators[0]) * aux->num_dd_valuators);
		aux->incoming_valuators_prev = (void *)vrShmemAlloc0(sizeof(aux->incoming_valuators_prev[0]) * aux->num_dd_valuators);
		for (count = 0; count < 13; count++) {
			aux->incoming_trackers[count] = (void *)vrShmemAlloc0(sizeof(aux->incoming_trackers[count][0]) * aux->num_dd_trackers);	/* allocate for each of the 13 values */
#if defined(TEST_APP) /* In the test-app vrShmemAlloc0() is mapped to just malloc(), so the values need to be zeroed. */
			memset(aux->incoming_trackers[count], 0, sizeof(aux->incoming_trackers[count][0]) * aux->num_dd_trackers);
#endif
		}

		/* assign the initial "previous" values (set to force a difference in first acquisition) */
		for (count = 0; count < aux->num_dd_buttons; count++) {
			aux->incoming_buttons_prev[count] = -1;
#if defined(TEST_APP) /* In the test-app vrShmemAlloc0() is mapped to just malloc(), so the values need to be zeroed. */
			aux->incoming_buttons[count] = 0;
#endif
		}
		for (count = 0; count < aux->num_dd_valuators; count++) {
			aux->incoming_valuators_prev[count] = -9999.0;
#if defined(TEST_APP) /* In the test-app vrShmemAlloc0() is mapped to just malloc(), so the values need to be zeroed. */
			aux->incoming_valuators[count] = 0.0;
#endif
		}
	} else {
		/* unexpected response */
		vrErrPrintf("_VruiDDInitializeDevice(): " RED_TEXT "Warning, got a packet of type %02x, expecting a %02x\n" NORM_TEXT, message, CONNECT_REPLY);
		return -1;
	}

	snprintf(aux->version, sizeof(aux->version), "VruiDD protocol:%d", response_protocol);

	/* ACTIVATE_REQUEST -- Request to activate server (prepare for sending packets) (0x03 0x00) */
	message = ACTIVATE_REQUEST;
	write(aux->fd_socket, &message, sizeof(message));	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	fsync(aux->fd_socket);

	/* STARTSTREAM_REQUEST -- Requests entering stream mode (server sends packets automatically) (0x07 0x00) */
	message = STARTSTREAM_REQUEST;
	write(aux->fd_socket, &message, sizeof(message));	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	fsync(aux->fd_socket);

	return 0;
}


/**********************************************************/
/* _VruiDDCloseDevice() is called in the CLOSE phase      */
/*   of input interface -- just before the ports are      */
/*   closed, etc.                                         */
static int _VruiDDCloseDevice(_VruiDDPrivateInfo *aux)
{
	uint16_t		message;	/* VRUI VRDeviceDaemon messages are 2 byte sequences */

	/* Don't close a device that's not open */
	if (aux->open == 0)
		return 0;

	message = STOPSTREAM_REQUEST;
	write(aux->fd_socket, &message, sizeof(message));	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	fsync(aux->fd_socket);

	message = DEACTIVATE_REQUEST;
	write(aux->fd_socket, &message, sizeof(message));	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	fsync(aux->fd_socket);

	message = DISCONNECT_REQUEST;
	write(aux->fd_socket, &message, sizeof(message));	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	fsync(aux->fd_socket);


	/* TODO: send a Vrui DISCONNECT_REQUEST message (perhaps also the STOPSTREAM_REQUEST & DEACTIVATE_REQUEST messages too) */
	return -1;
}


	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


#if defined(FREEVR) /* { */
/*****************************************************************/
/***  Functions for FreeVR access of VruiDD devices for input  ***/
/*****************************************************************/


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


/************************************************************************/
/* _VruiDDParseArgs(): this function is used to read input device (or   */
/*    stream) configuration options from the FreeVR configuration file, */
/*    command-line or environment variable string.  The meaning and     */
/*    syntax for each option is described at the top of this source     */
/*    file (and will often be repeated in the man page documentation    */
/*    at the end.                                                       */
static void _VruiDDParseArgs(_VruiDDPrivateInfo *aux, char *args)
{
#if 0 /* possible other items to handle (from vr_input.skeleton.c) */
static	char	*XX_choices[] = { "up", "away", NULL };
static	int	XX_values[] = { 1, 0 };		/* integer values that map to the text strings in XX_choices */
	int 	null_value = -1;		/* for reading one of the null region values */
#endif
	float	scale_value = -1.0;		/* for reading one of the scaling factor values */

	/* In the rare case of no arguments, just return */
	if (args == NULL)
		return;

	/******************************************/
	/** Argument format: "host" "=" hostname **/
	/******************************************/
	vrArgParseString(args, "host", &(aux->dd_host));

	/****************************************/
	/** Argument format: "port" "=" number **/
	/****************************************/
	vrArgParseInteger(args, "port", &(aux->dd_port));

	/****************************************/
	/** Argument format: "protocol" "=" number **/
	/****************************************/
	vrArgParseInteger(args, "protocol", &(aux->protocol));

#if 1 /* not sure whether we'll want this */
	/********************************************/
	/** Argument format: "valScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "valscale", &(aux->scale_valuator))) {
		aux->scale_valuator = VALUATOR_SENSITIVITY;
	}
#endif

	/**********************************************/
	/** Argument format: "trackScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "trackscale", &scale_value)) {
		aux->scale_trans = scale_value;
	}

	/*****************************************/
	/** Argument format: "scale" "=" number **/ /* (redundant with above) */
	/*****************************************/
	if (vrArgParseFloat(args, "scale", &scale_value)) {
		aux->scale_trans = scale_value;
	}

	/** TODO: other arguments to parse go here **/
}


/***************************************************************************/
/* The _VruiDDGetData(): function calls the _VruiDDReadInput() function to */
/*   get the latest data into the generic portion of the "_VruiDDGetData"  */
/*   structure.  Then this routine puts that data into the FreeVR data     */
/*   structures.  I.e. this is the FreeVR-specific portion of the data     */
/*   parsing.                                                              */
static void _VruiDDGetData(vrInputDevice *devinfo)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;
	int			num_inputs;		/* the number of incoming input values read (used for debugging only) */
	int			vruidd_map;		/* local storage of the mapping code from FreeVR input number to VruiDD input number */
	int			count;			/* counter for looping over inputs of a specific type */
	float			valuator_value;		/* holder for compositing a value together from components (e.g. sign & value) */
	float			scale_trans = aux->scale_trans;

	/*******************/
	/* gather the data */
	num_inputs = _VruiDDReadInput(aux);		/* Make sure we have the latest data stored in the "_FastrakPrivateInfo" structure */

	/****************************************/
	/* Handle button (2-switch) information */
	for (count = 0; count < aux->num_buttons; count++) {
		vruidd_map = aux->button_map_button[count];
		if (vruidd_map >= aux->num_dd_buttons) {	/* NOTE: '>=' because vruidd_map is a zero-based number */
			vrErrPrintf(RED_TEXT "_VruiDDGetData(): request for button (%d) not provided by the Vrui Device Daemon (max %d).\n", vruidd_map, aux->num_dd_buttons-1 /* correct for zero-base */);
			/* TODO: I should look into creating a dummy input, and using that after the first warning is printed */
		} else if ((aux->incoming_buttons[vruidd_map]) != (aux->incoming_buttons_prev[vruidd_map])) {
			/* handle button inputs as buttons or self-controls */
			if (aux->button_inputs[count] != NULL) {
				switch (aux->button_inputs[count]->input_type) {
				case VRINPUT_BINARY:
					vrAssign2switchValue((vr2switch *)(aux->button_inputs[count]), (aux->incoming_buttons[vruidd_map] != 0));
#if 0
vrPrintf("Assignment: VruiDD button %d maps to FreeVR button %d, max = %d, val = %d, prev = %d\n", vruidd_map, count, aux->num_dd_buttons, aux->incoming_buttons[vruidd_map], aux->incoming_buttons_prev[vruidd_map]);
#endif
					break;
				case VRINPUT_CONTROL:
					vrCallbackInvokeDynamic(((vrControl *)(aux->button_inputs[count]))->callback, 1, (aux->incoming_buttons[vruidd_map] != 0));
					break;
				default:
					vrErrPrintf(RED_TEXT "_VruiDDGetData(): Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
					break;
				}
			}
		}
	}

	/* now save all the current values as the "previous" values */
	for (count = 0; count < aux->num_dd_buttons; count++) {
		aux->incoming_buttons_prev[count] = aux->incoming_buttons[count];
	}

	/*******************************/
	/* Handle valuator information */
	for (count = 0; count < aux->num_valuators; count++) {
		vruidd_map = aux->valuator_map_valuator[count];
		if (vruidd_map >= aux->num_dd_valuators) {	/* NOTE: '>=' because vruidd_map is a zero-based number */
			vrErrPrintf(RED_TEXT "_VruiDDGetData(): request for valuator (%d) not provided by the Vrui Device Daemon (max %d).\n", vruidd_map, aux->num_dd_valuators-1 /* correct for zero-base */);
			/* TODO: I should look into creating a dummy input, and using that after the first warning is printed */
		} else if ((aux->incoming_valuators[vruidd_map]) != (aux->incoming_valuators_prev[vruidd_map])) {
			/* handle valuator inputs as valuators or self-controls */
			if (aux->valuator_inputs[count] != NULL) {
				valuator_value = aux->incoming_valuators[count] * aux->scale_valuator * aux->valuator_map_sign[count];
				switch (aux->valuator_inputs[count]->input_type) {
				case VRINPUT_VALUATOR:
					vrAssignValuatorValue((vrValuator *)(aux->valuator_inputs[count]), valuator_value);
#if 0
vrPrintf("Assignment: VruiDD valuator %d maps to FreeVR valuator %d, max = %d, val = %f, prev = %f\n", vruidd_map, count, aux->num_dd_valuators, aux->incoming_valuators[vruidd_map], aux->incoming_valuators_prev[vruidd_map]);
#endif
					break;
				case VRINPUT_CONTROL:
					vrCallbackInvokeDynamic(((vrControl *)(aux->valuator_inputs[count]))->callback, 1, &valuator_value);
					break;
				default:
					vrErrPrintf(RED_TEXT "_VruiDDGetData(): Unable to handle valuator inputs that aren't Floating or Control inputs\n" NORM_TEXT);
					break;
				}
			}
		}
	}

	/* now save all the current values as the "previous" values */
	for (count = 0; count < aux->num_dd_valuators; count++) {
		aux->incoming_valuators_prev[count] = aux->incoming_valuators[count];
	}

	/****************************************/
	/* Handle 6sensor (tracker) information */
	for (count = 0; count < aux->num_6sensors; count++) {
		vr6sensor	*current_6sensor;
		vrMatrix	new_position;
		vrVector	new_location;
		vrQuat		new_orientation;

		current_6sensor = aux->sensor6_inputs[count];
		vruidd_map = aux->tracker_map_tracker[count];

		new_location.v[VR_X] = aux->incoming_trackers[VR_X][vruidd_map] * scale_trans;
		new_location.v[VR_Y] = aux->incoming_trackers[VR_Y][vruidd_map] * scale_trans;
		new_location.v[VR_Z] = aux->incoming_trackers[VR_Z][vruidd_map] * scale_trans;

		new_orientation.v[VR_X] = aux->incoming_trackers[VR_QX][vruidd_map];
		new_orientation.v[VR_Y] = aux->incoming_trackers[VR_QY][vruidd_map];
		new_orientation.v[VR_Z] = aux->incoming_trackers[VR_QZ][vruidd_map];
		new_orientation.v[VR_W] = aux->incoming_trackers[VR_QW][vruidd_map];

		vrMatrixSetIdentity(&new_position);
		vrMatrixSetTransFromVector(&new_position, &new_location);
		vrMatrixSetRotationFromQuat(&new_position, &new_orientation);

		/* since 6-sensors generally change rapidly, there is little point */
		/*   and often moderate expense in comparing for changes from the  */
		/*   current value to the new value.                               */
		/* although perhaps watching the timestamp might be a useful way   */
		/*   to do this.                                                   */
		/* TODO: perhaps verify that this is indeed a VRINPUT_6SENSOR */
		vrAssign6sensorValue(current_6sensor, &new_position, 0 /*, TODO: timestamp from daemon */);
	}
}


	/****************************************************************/
	/*    Function(s) for parsing VruiDD "input" declarations.      */
	/*                                                              */
	/*  These _VruiDD<type>Input() functions are called during the  */
	/*  CREATE phase of the input interface.                        */

/* ----------------------------------------------------------------------- */
/* The arguments for each of these "input" declaration parsers follow the  */
/*   same form, no matter what type of input or device:                    */
/*                                                                         */
/* devinfo -- a pointer to the overall structure that defines this device. */
/*            It is used to get access to the auxiliary data of the device.*/
/*                                                                         */
/* input -- a pointer to the vrGenericInput structure that is a specific   */
/*          FreeVR input, such as a single vr2switch, that is part of the  */
/*          specified input device.                                        */
/*          NOTE: the "input_type" can be determined by accessing the      */
/*          "input->input_type" field.  This can be used to have a single  */
/*          mapping function handle requests for multiple types of inputs. */
/*          See _XwindowsKeyboardInput() for an example of this.           */
/*                                                                         */
/* The remaining argument contains three strings that are from the portion */
/*   of the input config statement contained within the parentheses.  This */
/*   is pre-parsed as the "DTI" -- {<sub-Device>:}<Type>[<Instance>].  Any */
/*   of these subfields can be an empty string.  Usually:                  */
/*                                                                         */
/* device -- a string that may be used to indicate a sub-device of the     */
/*           provided hardware.  For example, an X11 input device may      */
/*           have "keyboard", "mouse" and "dial+buttons" subdevices.       */
/*           In most cases the "device" value is the empty string, and     */
/*           probably could always be included as part of the "type" value.*/
/*                                                                         */
/* type -- a string that indicates the class of input interface on the     */
/*         particular device/subdevice.  For example, an X11 input device  */
/*         may have "key", "mousebutton" and "pointerlocation" types.      */
/*         Or, a Magellan may have "button" and "6dof" types.              */
/*                                                                         */
/*         NOTE that the "type" value is also used in determining which of */
/*         these functions should be called (by matching the first element */
/*         of the vrInputFunctions list below), so it has already played a */
/*         roll in how this input is handled.  Therefore it's value is     */
/*         often ignored in the parsing function itself.  However, it is   */
/*         possible for more than one <type> string to point to the same   */
/*         device parsing function.  Thus in those cases it may be         */
/*         important to make use of this information within the parsing    */
/*         function.                                                       */
/*                                                                         */
/*         NOTE also that this string is different than the "input_type"   */
/*         data mentioned above as part of the "input" in that this is a   */
/*         string name that might describe the physical form of the input, */
/*         versus the "input_type" which describes a logical FreeVR input  */
/*         type such as a BINARY, VALUATOR or 6SENSOR input.               */
/*                                                                         */
/* instance -- a string usually used to indicate a specific instance of    */
/*             the type of data.  For example button "2" on a Magellan,    */
/*             unit "1" of a Flock of Birds, or key "Esc" of an X11        */
/*             keyboard.  Other data may also be included for altering how */
/*             this input is handled (See the VRPN device for an example). */
/*                                                                         */
/*    NOTE: since the "device" subfield is often optional, and the "type"  */
/*    subfield is used to determine which function to call, many of these  */
/*    functions will only make use of the "instance" subfield.             */
/*                                                                         */
/* The return values specify whether the initial match is confirmed, or    */
/*   rejected.  For some devices (such as the Xwindows-keyboard), there    */
/*   may be an overall match for the keyboard, but then the specific key   */
/*   instance may not match a known key for the keyboard.  In such cases   */
/*   a VRINPUT_MATCH_UNABLE is returned, indicating that the search failed,*/
/*   but that the search for a matching device should cease.  If the match */
/*   is successful, then a VRINPUT_MATCH_ABLE is returned.  If the match   */
/*   altogether fails, then a VRINPUT_NOMATCH is returned, and the search  */
/*   continues.                                                            */
/*                                                                         */
/* NOTE: these functions are called before making the physical connection  */
/*   to the device, so it is not possible to query features of the device  */
/*   within these functions.                                               */
/*                                                                         */
/* NOTE: the vrInputDTI parsing (done in vr_parse.c) is for the common     */
/*   form of how inputs are specified in the FreeVR configuration (rc)     */
/*   files.  Thus the following input specification:                       */
/*       input "2switch[left]" = "2switch(Joystick0:button[left])";        */
/*   would parse the devices input as:                                     */
/*       device (d) = "Joystick0"    ("sub-device" is probably more exact) */
/*       type (t) = "button"                                               */
/*       instance (i) = "left"                                             */
/* ----------------------------------------------------------------------- */


/***************************************************************************/
/* _VruiDDButtonInput(): this function takes a "device,type,instance"      */
/*    (DTI) configuration triplet for a "button" class of input, parses    */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "button(<sub-device>:<type>[<instance>, <args>])"   */
/*    NOTE: for the <VruiDDn> device, there may not be sub-device options. */
/* See the "List of possible inputs" in _VrpnCreateFunction() below        */
/*    for the codes that map to this function.                             */
/*                                                                         */
/* Here is a sample definition for a VruiDD button:                        */
/*     input "2switch[left]" = "2switch(Joystick0:button[0])";             */
/**************************************************************************/
static vrInputMatch _VruiDDButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;
	int			button_num;		/* the current count of FreeVR button inputs */
	int			button_choice;		/* the chosen number for this button */

	/* The VruiDD method of handling input arrays may be a combination of the way */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/*******************/
	/* select a button */
	button_num = aux->num_buttons;
	aux->num_buttons++;
	button_choice = vrAtoI(dti->instance);

	/*****************************/
	/* check the selected button */
	if (button_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_VruiDDButtonInput(): Warning, button['%s'] did not match any known button\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->button_inputs[button_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VruiDDButtonInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->button_inputs[button_num] = (vr2switch *)input;
	aux->button_map_button[button_num] = button_choice;

	vrDbgPrintfN(INPUT_DBGLVL, "_VruiDDButtonInput(): assigned button event of value %d (VruiDD button %d) to input pointer = %#p, from input named '%s'.\n",
		button_num, button_choice, aux->button_inputs[button_num], input->my_object->name);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/***************************************************************************/
/* _VruiDDValuatorInput(): this function takes a "device,type,instance"    */
/*    (DTI) configuration triplet for a "valuator" class of input, parses  */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "valuator(<sub-device>:<type>[<instance>, <args>])" */
/*    NOTE: for the <VruiDD> device, there may not be sub-device options.  */
/*                                                                         */
static vrInputMatch _VruiDDValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;
	int			valuator_num;		/* the current count of FreeVR valuator inputs */
	int			valuator_choice;	/* the chosen number for this valuator */

	/* The VruiDD method of handling input arrays may be a combination of the way */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/********************************/
	/* select a valuator (and sign) */
	valuator_num = aux->num_valuators;
	aux->num_valuators++;
	if (dti->instance[0] == '-') {
		aux->valuator_map_sign[valuator_num] = -1.0;
		valuator_choice = vrAtoI(&(dti->instance[1]));	/* skip the negation sign */
	} else {
		aux->valuator_map_sign[valuator_num] =  1.0;
		valuator_choice = vrAtoI(dti->instance);
	}

	/*******************************/
	/* check the selected valuator */
	if (valuator_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_VruiDDValuatorInput(): Warning, valuator['%s'] did not match any known valuator\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->valuator_inputs[valuator_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VruiDDValuatorInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	aux->valuator_map_valuator[valuator_num] = valuator_choice;

	vrDbgPrintfN(INPUT_DBGLVL, "_VruiDDValuatorInput(): assigned valuator event of value %d (VruiDD valuator %d) to input pointer = %#p, from input named '%s'.\n",
		valuator_num, valuator_choice, aux->valuator_inputs[valuator_num], input->my_object->name);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/***************************************************************************/
/* _VruiDD6sensorInput(): this function takes a "device,type,instance"     */
/*    (DTI) configuration triplet for a "6sensor" class of input, parses   */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "6sensor(<sub-device>:<type>[<instance>, <args>])"  */
/*    NOTE: for the <VruiDD> device, there may not be sub-device options.  */
/*                                                                         */
/* Here are sample definitions for a VruiDD position sensor:               */
/*    input "head" = "6sensor(tracker[0, r2e])";                           */
static vrInputMatch _VruiDD6sensorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;
	vrMatrix		mat_store;		/* temporary matrix storage */
	vr6sensor		*sensor6_input;		/* pointer to this new 6-sensor input */
	int			tracker_num;		/* the current count of FreeVR tracker inputs */
	int			tracker_choice;		/* the chosen number for this tracker */

	/* The VruiDD method of handling input arrays may be a combination of the way */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/********************/
	/* select a 6sensor */
	tracker_num = aux->num_6sensors;
	aux->num_6sensors++;
	tracker_choice = vrAtoI(dti->instance);

	/******************************/
	/* check the selected 6sensor */
	if (tracker_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_VruiDD6sensorInput(): Warning, tracker['%s'] did not match any known tracker\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->sensor6_inputs[tracker_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VruiDD6sensorInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->sensor6_inputs[tracker_num] = (vr6sensor *)input;
	aux->tracker_map_tracker[tracker_num] = tracker_choice;
	sensor6_input = aux->sensor6_inputs[tracker_num];

	/* vrAssign6sensorR2Exform() parses the rest of the instance setting */
	vrAssign6sensorR2Exform(sensor6_input, strchr(dti->instance, ','));

	/* Set the initial values */
	vrMatrixSetIdentity(&mat_store);
	vrMatrixPostMult(&mat_store, sensor6_input->ipos_xform);
	vrAssign6sensorValue(sensor6_input, &mat_store, -1 /* , vrCurrentWallTime() */);
	vrAssign6sensorActiveValue(sensor6_input, 0);
	vrAssign6sensorErrorValue(sensor6_input, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "_VruiDD6sensorInput(): assigned tracker event of value %d (VruiDD tracker %d) to input pointer = %#p, from input named '%s'.\n",
		tracker_num, tracker_choice, aux->sensor6_inputs[tracker_num], input->my_object->name);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}



	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the device's features.      */
	/*                                                          */
	/*  These _VruiDD<operation>Callback() functions are        */
	/*  called during the CREATE phase of the input interface.  */

/*************************************************************************/
/* This collection of callbacks (for controlling the device features) is */
/*   designed to allow the FreeVR system configuration to map physical   */
/*   inputs of this device to control the operation of the VR system     */
/*   as opposed to the norm, where inputs are used to affect the         */
/*   simulated virtual world.                                            */
/*                                                                       */
/* Each command callback takes exactly two arguments:                    */
/*   * devinfo -- this is information about this particular instance of  */
/*                  connecting with an input device or system.  This     */
/*                  provides:                                            */
/*                  1) the ability to alter the settings of the device   */
/*                     itself using the "aux" data field (for example    */
/*                     the SpaceTec devices can do a "reset" with        */
/*                     _SpacetecSensorResetCallback(), or the X11 input  */
/*                     can loop through all the simulated sensors with   */
/*                     _XwindowsSensorNextCallback();                    */
/*               or 2) these callbacks can be used to affect the overall */
/*                     VR system, such as toggling the "paused" flag.    */
/*   * value -- an integer flag that can affect the callback.            */
/*       NOTE: this assumes that the callback is from a button input, or */
/*       from a valuator passing a threshold.  For buttons, the value    */
/*       will be 1 (one) when the button is depressed.                   */
/*************************************************************************/


#if 0 /* delete me */
/***************************************************************************/
/* _VruiDDSystemPauseToggleCallback(): this callback toggles the "paused"  */
/*   field of the overall FreeVR "context" structure.  When the system is  */
/*   "paused", rendering still functions, but the inputs are frozen.  This */
/*   can be used for debugging, or for still photography of a CAVE.        */
static void _VruiDDSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VruiDD Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/*****************************************************************************/
/* _VruiDDPrintContextStructCallback(): this callback generates output of    */
/*   the overall system "context" data.  It is generally used when debugging */
/*   the library or the configuration.                                       */
static void _VruiDDPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintContext(stdout, devinfo->context, verbose);		/* NOTE: this effectively requires the inclusion of "vr_system.h" */
}

/************************************************************************/
/* _VruiDDPrintConfigStructCallback(): this callback outputs the entire */
/*    VR system configuration to the text console.                      */
static void _VruiDDPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintConfig(stdout, devinfo->context->config, verbose);	/* NOTE: this effectively requires the inclusion of "vr_config.h" */
}

/************************************************************************/
/* _VruiDDPrintInputStructCallback(): this callback outputs the current */
/*    state of the entire input structure of the running VR system.     */
static void _VruiDDPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintInput(stdout, devinfo->context->input, verbose);
}
#endif

/***************************************************************************/
/* _VruiDDPrintPrivateInfoCallback(): this callback outputs the contents   */
/*    of the device specific internal structure to help with configuration */
/*    and system debugging.                                                */
static void _VruiDDPrintPrivateInfoCallback(vrInputDevice *devinfo, int value)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VruiDDPrintPrivateInfo(stdout, aux, verbose);
}

/***********************************************************************/
/* _VruiDDPrintHelpCallback(): this callback outputs the full system's */
/*     help information string.                                        */
static void _VruiDDPrintHelpCallback(vrInputDevice *devinfo, int value)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VruiDDPrintHelp(stdout, aux);
}


	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */

/**********************************************************************/
/* This collection of callbacks (for interfacing with the device) are */
/*   the standard ways in which the FreeVR system creates an instance */
/*   of access to a particular device.  It consists of the six (6)    */
/*   basic operational functions:                                     */
/*     1) Creating the device instance                                */
/*     2) Open a communications channel with the device               */
/*     3) Close the communications channel with the device            */
/*     4) Reset the device                                            */
/*     5) Poll input data from the device                             */
/* and 6) Print information about the state of the device             */
/**********************************************************************/


/***********************************************************************/
/* _VruiDDCreateFunction(): this callback maps the configuration input */
/*    options and control options to the above callback routines that  */
/*    will then execute the code to handle each type of input/control. */
/*                                                                     */
/* There are N primary tasks of this function:                         */
/*     1) Allocate all the memory for the device instance data.        */
/*     2) Parse the options for operating this device instance.        */
/*     3) Allocate the memory for all the inputs and self-controls.    */
/*     4) Actually map the callbacks for all inputs and self-controls. */
/* and 5) Set the flag that the device is configured and ready to go.  */
static void _VruiDDCreateFunction(vrInputDevice *devinfo)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_VruiDDInputs[] = {
				{ "button", VRINPUT_2WAY, _VruiDDButtonInput },
				{ "valuator", VRINPUT_VALUATOR, _VruiDDValuatorInput },
				{ "6sensor", VRINPUT_6SENSOR, _VruiDD6sensorInput },
				{ "tracker", VRINPUT_6SENSOR, _VruiDD6sensorInput },	/* a synonym for "6sensor" */
				{ NULL, VRINPUT_UNKNOWN, NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_VruiDDControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", vrInputSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", vrInputPrintContextStructCallback },
				{ "print_config", vrInputPrintConfigStructCallback },
				{ "print_input", vrInputPrintInputStructCallback },
				{ "print_devinfo", _VruiDDPrintPrivateInfoCallback },
				{ "print_help", _VruiDDPrintHelpCallback },
		/** TODO: other callback control functions go here **/
				/* other controls */

				/* end of the list */
				{ NULL, NULL } };

	_VruiDDPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxiliary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(*aux));
	aux = (_VruiDDPrivateInfo *)devinfo->aux_data;
	_VruiDDInitializeStruct(aux, devinfo->type);

	/******************/
	/* handle options */
	aux->dd_host = vrShmemStrDup(DEFAULT_HOST);	/* initialize to the default host value */
	aux->dd_port = DEFAULT_PORT;			/* initialize to the default port value */
	_VruiDDParseArgs(aux, devinfo->args);

	/***************************************/
	/* create the inputs and self-controls */
	vrInputCountDataContainers(devinfo);

	/* Allocate memory for specific input types */
	aux->button_inputs = (vr2switch **)vrShmemAlloc((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(vr2switch *));
	aux->button_map_button = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));

	aux->valuator_inputs = (vrValuator **)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(vrValuator *));
	aux->valuator_map_valuator = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_map_sign = (float *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(float));

	aux->sensor6_inputs = (vr6sensor **)vrShmemAlloc((devinfo->num_6sensors) * sizeof(vr6sensor *));
	aux->tracker_map_tracker = (int *)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(int));

	/* Create the inputs within the FreeVR system */
	/* These two functions do the real work of this function, by creating */
	/*   the mappings as they are laid out in the configuration file.     */
	vrInputCreateDataContainers(devinfo, _VruiDDInputs);
	vrInputCreateSelfControlContainers(devinfo, _VruiDDInputs, _VruiDDControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	vrDbgPrintf("_VruiDDCreateFunction(): Done creating VruiDD inputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}


/***************************************************************************/
/* _VruiDDOpenFunction(): this callback opens the communications channel   */
/*    with the physical device or data streaming utility.  This is         */
/*    generally unique for each type of device or stream.                  */
/*                                                                         */
/* There are no return values, but there are two input-device flags that   */
/*    are set to indicate the state of the device:                         */
/*      * open -- indicates when the communications channel is working     */
/*      * operating -- indicates the device is actually ready to send data */
static void _VruiDDOpenFunction(vrInputDevice *devinfo)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;

	vrTrace("_VruiDDOpenFunction", devinfo->name);

	/*******************/
	/* open the device */
	aux->fd_socket = vrSocketCall(aux->dd_host, aux->dd_port);
	if (aux->fd_socket < 0) {
		aux->open = 0;
		devinfo->operating = 0;
		vrErrPrintf("(%s::_VruiDDOpenFunction()::%d) error: "
			RED_TEXT "couldn't open socket '%s:%d' for %s\n" NORM_TEXT,
			__FILE__, __LINE__, aux->dd_host, aux->dd_port, devinfo->name);
		snprintf(aux->version, sizeof(aux->version), "- unconnected VruiDD -");
	} else {
		aux->open = 1;
		if (_VruiDDInitializeDevice(aux) < 0) {
			devinfo->operating = 0;
			vrErrPrintf("_VruiDDOpenFunction(): "
				RED_TEXT "Warning, unable to initialize '%s' VruiDD.\n" NORM_TEXT,
				devinfo->name);
		} else {
			devinfo->operating = 1;
			vrDbgPrintf("_VruiDDOpenFunction(): Done opening VruiDD input device '%s' (operating = %d).\n", devinfo->name, devinfo->operating);
		}
	}

	vrTrace("_VrpnOpenFunction", "returning");

	return;
}


/***************************************************************************/
/* _VruiDDCloseFunction(): this callback closes the communications channel */
/*   with the DEVICE device.                                               */
static void _VruiDDCloseFunction(vrInputDevice *devinfo)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;
	int			count;			/* loop counter */

	vrDbgPrintf("_VruiDDCloseFunction(): About to close VruiDD input device '%s'.\n", devinfo->name);

	_VruiDDCloseDevice(aux);

	/************************/
	/* close the connection */
	if (aux != NULL) {
		if (aux->open)
			vrSocketClose(aux->fd_socket);
		if (aux->incoming_buttons != NULL) vrShmemFree(aux->incoming_buttons);
		if (aux->incoming_buttons_prev != NULL) vrShmemFree(aux->incoming_buttons_prev);
		if (aux->incoming_valuators != NULL) vrShmemFree(aux->incoming_valuators);
		if (aux->incoming_valuators_prev != NULL) vrShmemFree(aux->incoming_valuators_prev);
		for (count = 0; count < 13; count++)
			if (aux->incoming_trackers[count] != NULL) vrShmemFree(aux->incoming_trackers[count]);

#ifdef FREEVR
		/* free the FreeVR specific fields of "aux" */
		/* TODO: ... */
#endif

		vrShmemFree(aux);			/* aka devinfo->aux_data */
	}

	return;
}


/*************************************************************************/
/* _VruiDDResetFunction(): this callback provides the means by which the */
/*    device or input stream can be reset to a good state.               */
static void _VruiDDResetFunction(vrInputDevice *devinfo)
{
	_VruiDDPrivateInfo	*aux = (_VruiDDPrivateInfo *)devinfo->aux_data;

	return;
}


/****************************************************************************/
/* _VruiDDPollFunction(): this callback actually reads the data from the    */
/*    input device or stream using a polling (request) method of operation. */
/*    A check is done to confirm that the device is in "operating" state    */
/*    before attempting to read the data.                                   */
static void _VruiDDPollFunction(vrInputDevice *devinfo)
{
	if (devinfo->operating) {
		_VruiDDGetData(devinfo);
	} else {
		/* TODO: try to open the device again */
	}

	return;
}



	/*******************************/
	/***  FreeVR public routine  ***/
	/*******************************/

/****************************************************************************/
/* vrVruiDDInitInfo(): this function serves as the "public" means by which  */
/*   a device (or input stream) is interfaced with from the FreeVR library. */
/*   This function creates the shared memory for the device version string, */
/*   and then instantiates the new instance with all the callbacks needed   */
/*   for FreeVR to operate the input.                                       */
void vrVruiDDInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("-get from VruiDD device-");
	devinfo->Create = vrCallbackCreateNamed("VruiDD:Create-Def", _VruiDDCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("VruiDD:Open-Def", _VruiDDOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("VruiDD:Close-Def", _VruiDDCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("VruiDD:Reset-Def", _VruiDDResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("VruiDD:PollData-Def", _VruiDDPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("VruiDD:PrintAux-Def", _VruiDDPrintPrivateInfo, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "vrVruiDDInitInfo(): callbacks created.\n");
}
#endif /* } FREEVR */


	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



/****************************************************************************/
/****************************************************************************/
/* The following section is the independent "test" utility that can be used */
/*   to test communications and output from the input device (or stream).   */
#if defined(TEST_APP) /* { */

/*********************************************************************/
/* Ugh, I hate globals, but I don't know a better way to get the     */
/*   "done" value into the interrupt signal function exit_testapp(). */
static	int	done = -1;	/* -1 indicates that communication isn't open yet */

#define	RENDERLINES	0
#define	RENDERCSV	1	/* NOTE: not yet implemented here -- see vr_input.fastrak.c for example */
#define	RENDERSCREEN	2	/* NOTE: not yet implemented here -- see vr_input.{asc_fob,fastrak}.c */


/*******************************************************************/
/* _RenderValuesLine(): Render the Tracking values in a streaming  */
/*   line style format -- with a carriage return, so will overwrite*/
/*   previous lines (unless the screen is insufficiently wide.     */
void _RenderValuesLine(_VruiDDPrivateInfo *aux)
{
	int		devnum;			/* for looping through devices */
	int		button1_value = -1;	/* value of the first actual button in the system (used for chorded-quit) */
	int		button2_value = -1;	/* value of the second actual button in the system (used for chorded-quit) */

	/* for calculating the Euler angles of a position tracker input */
	vrPoint		location_input;		/* The location/translation portion */
	vrQuat		quat_input;		/* The quaternion portion */
	vrEuler		euler_input;		/* The Euler+translation representation */

	/******************************************/
	/* Display tracker (6-sensor) information */
	if (aux->num_dd_trackers > 0) {
		int	ptv_count;		/* for looping over the position-tracker values */
		printf("pt: ");
		for (devnum = 0; devnum < aux->num_dd_trackers; devnum++) {
			for (ptv_count = 0; ptv_count < (devnum == 0 ? 13:3); ptv_count++) {	/* TODO: s/b 13, but doing 3 for now (except for the first) */
				printf("%6.3f ", aux->incoming_trackers[ptv_count][devnum]);
			}
			printf("-- ");

			/* Store the given <input> in FreeVR types (inc. Eulers) */
			vrPointSet3d(&location_input, aux->incoming_trackers[VR_X][devnum], aux->incoming_trackers[VR_Y][devnum], aux->incoming_trackers[VR_Z][devnum]);
			vrQuatSet4d(&quat_input, aux->incoming_trackers[VR_QX][devnum], aux->incoming_trackers[VR_QY][devnum], aux->incoming_trackers[VR_QZ][devnum], aux->incoming_trackers[VR_QW][devnum]);
			vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
		}
	}

	/*****************************************/
	/* Display button (2-switch) information */
	if (aux->num_dd_buttons > 0) {
		printf("          " "buts: ");	/* blank area is to provide space for the timing value */
		for (devnum = 0; devnum < aux->num_dd_buttons; devnum++) {
			printf("%c", (aux->incoming_buttons[devnum] == 0 ? '.' : 'X'));
		}
	}

	/*****************************************/
	/* Display analog (valuator) information */
	if (aux->num_dd_valuators > 0) {
		printf(" vals: ");
		for (devnum = 0; devnum < aux->num_dd_valuators; devnum++) {
			printf("%6.3f ", aux->incoming_valuators[devnum]);
		}
	}


	/* check to see if we should quit */
	if (aux->num_dd_buttons > 2) {
		if (aux->incoming_buttons[0] && aux->incoming_buttons[1])
			done = 1;
	}
	printf("    \r");
	fflush(stdout);
}


/*******************************************************************/
/* _RenderValuesScreen(): Render the Tracking values in a text-gui */
/*   style format.                                                 */
/* NOTE: this routine relies on using the ANSI terminal character  */
/*   codes.  Since this is just a test application, I decided to   */
/*   forgo the requirement of the curses library, and instead      */
/*   limit the terminal on which this version would work.          */
/*   On Linux do "man console_codes" to get the sequence codes     */
/*   for other operations.                                         */
void _RenderValuesScreen(_VruiDDPrivateInfo *aux)
{
	int		devnum;			/* for looping through devices */
	int		start_line = 2;		/* first line on the screen to place data */	/* TODO: for some reason things scroll up two lines -- need to investigate */
	float		percent;		/* used for simple valuator visualization */

	int		num_lines = 40;		/* number of lines for showing chart */
	float		max_trans = 2.0;	/* 2 meters in one direction */
	float		max_rot = 180.0;	/* 180 degrees max in each direction */
	float		max_qual = 255.0;	/* 0-255 is the tracking quality range */	/* TODO: I don't know that Vrui has a quality value */

#if 0
	int		num_columns = 100;
#endif
	int		col_width = 15;		/* the number of characters per column */
	int		col1_offset = 5;	/* the offset before the first column */
	int		butval_offset;		/* the column to output the button & valuators */

	int		header_lines = 6;	/* used to shift the center and the min/max scaling */
	int		center = (num_lines-header_lines) / 2.0 + header_lines + 1;
	int		vertical;		/* the vertical line to render on -- used to prevent underflow */
	float		trans_scale = (num_lines-header_lines) * -0.5 / max_trans;
	float		rot_scale = (num_lines-header_lines) * -0.5 / max_rot;
	float		quat_scale = (num_lines-header_lines) * -0.5 / 1.0; /* (1.0 = max_quat) */

	/* for calculating the Euler angles of a position tracker input */
	vrPoint		location_input;		/* The location/translation portion */
	vrQuat		quat_input;		/* The quaternion portion */
	vrEuler		euler_input;		/* The Euler+translation representation */


	/* move to upper left corner of screen */
	vrPrintf(ANSI_MOVETOUPPERLEFT);

	/* clear to end of screen */
	vrPrintf(ANSI_CLEARTOEOS);

	/* draw a data vs. debug separating line */
	vrPrintf(ANSI_MOVETOROWCOL, num_lines+1, 0);
	vrPrintf("=" TTY_REPEATCHAR, 80);
	vrPrintf(ANSI_MOVETOROWCOL, 1000, 0);

	/* save current cursor position */
	vrPrintf(ANSI_CURSORLOC_SAVE);

	/******************************************/
	/* Display header information */
	vrPrintf(ANSI_MOVETOROWCOL "VruiDD Server reported devices:", start_line + 0, 2);
	vrPrintf(ANSI_MOVE "Trackers : %d", start_line + 1, 10, aux->num_dd_trackers);
	vrPrintf(ANSI_MOVE "Buttons  : %d", start_line + 2, 10, aux->num_dd_buttons);
	vrPrintf(ANSI_MOVE "Valuators: %d", start_line + 3, 10, aux->num_dd_valuators);

	/* Set normal text output to be within the given margin      */
	/*   (This pushes all debug output into the "margin" region, */
	/*   and thus doesn't obscure the screen rendering.)         */
	/*   NOTE: we exaggerate the bottom line to allow the screen */
	/*   to be expanded and still work properly.                 */
	vrPrintf(ANSI_SETMARGIN_TOPBOT, num_lines+2, num_lines+200);

	/******************************************/
	/* Display tracker (6-sensor) information */
	/* start counting at '1' to skip the simulator placeholder */
	for (devnum = 0; devnum < aux->num_dd_trackers; devnum++) {
		int	ptv_count;		/* for looping over the position-tracker values */

		/* NOTE: Ideally we'd have a "vruidd_dev" pointer, but for  */
		/*   some reason, the [devnum] index is on the outside, so  */
		/*   we can't really have a pointer to a single entity from */
		/*   which we can just get the X,Y,Z,etc. values.           */

		/* Store the given <input> in FreeVR types (inc. Eulers) */
		vrPrintf(ANSI_CURSORLOC_RETURN);        /* jump to output margin for possible debugging output */
		vrPointSet3d(&location_input, aux->incoming_trackers[VR_X][devnum], aux->incoming_trackers[VR_Y][devnum], aux->incoming_trackers[VR_Z][devnum]);
		vrQuatSet4d(&quat_input, aux->incoming_trackers[VR_QX][devnum], aux->incoming_trackers[VR_QY][devnum], aux->incoming_trackers[VR_QZ][devnum], aux->incoming_trackers[VR_QW][devnum]);
		vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
		vrPrintf(ANSI_CURSORLOC_SAVE);          /* save cursor for next time */

		/* now print info headers */
		vrPrintf(ANSI_MOVE "Tracker %d", start_line + 4, devnum*col_width + col1_offset, devnum+1);
		vrPrintf(ANSI_MOVE "XYZAERijkW", start_line + 5, devnum*col_width + col1_offset + 1);
#if 1 /* [optional] show the XYZ values under the header */
		vrPrintf(ANSI_MOVE "X=% 9.5f", start_line + 7, devnum*col_width + col1_offset, aux->incoming_trackers[VR_X][devnum]);
		vrPrintf(ANSI_MOVE "Y=% 9.5f", start_line + 8, devnum*col_width + col1_offset, aux->incoming_trackers[VR_Y][devnum]);
		vrPrintf(ANSI_MOVE "Z=% 9.5f", start_line + 9, devnum*col_width + col1_offset, aux->incoming_trackers[VR_Z][devnum]);
#endif


		/* print a *zero* line */
		vrPrintf(ANSI_MOVE RED_TEXT ">----------<" NORM_TEXT, center, devnum*col_width + col1_offset);

#if 1
		vrPrintf(ANSI_MOVE RED_TEXT "|==========|" NORM_TEXT, center + (int)(max_trans*trans_scale)+1, devnum*col_width + col1_offset);	/* top line -- presently overwrites the "XYZAERijkW" header */
#endif
		vrPrintf(ANSI_MOVE RED_TEXT "|==========|" NORM_TEXT, center - (int)(max_trans*trans_scale), devnum*col_width + col1_offset);	/* bottom line */
		vrPrintf(ANSI_MOVE RED_TEXT "|..........|" NORM_TEXT, center + (int)(max_trans*trans_scale*0.5), devnum*col_width + col1_offset);	/* 1/4 down line */
		vrPrintf(ANSI_MOVE RED_TEXT "|..........|" NORM_TEXT, center - (int)(max_trans*trans_scale*0.5), devnum*col_width + col1_offset);	/* 3/4 down line */


#  if 1 /* [optional] switch to reverse mode for moving values */
		vrPrintf(REVERSE_TEXT);	/* switch to reverse */
#  endif

		/* print the X value */
		vertical = (int)(aux->incoming_trackers[VR_X][devnum]*trans_scale)+center;
		if (vertical < 0) {
			/* if we go off the top, stop and make it red */
			vertical = 0;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "X" ANSI_TEXTFORENORM, vertical, devnum*col_width + col1_offset+1);

		/* print the Y value */
		vertical = (int)(aux->incoming_trackers[VR_Y][devnum]*trans_scale)+center;
		if (vertical < 0) {
			/* if we go off the top, stop and make it red */
			vertical = 0;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "Y" ANSI_TEXTFORENORM, vertical, devnum*col_width + col1_offset+1);

		/* print the Z value */
		vertical = (int)(aux->incoming_trackers[VR_X][devnum]*trans_scale)+center;
		if (vertical < 0) {
			/* if we go off the top, stop and make it red */
			vertical = 0;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "X" ANSI_TEXTFORENORM, vertical, devnum*col_width + col1_offset+1);

		/* print the A value (Azimuth) */
		vrPrintf(ANSI_MOVETOROWCOL "A", (int)(euler_input.r[VR_AZIM]*rot_scale)+center, devnum*col_width + col1_offset + 4);

		/* print the E value (Elevation) */
		vrPrintf(ANSI_MOVETOROWCOL "E", (int)(euler_input.r[VR_ELEV]*rot_scale)+center, devnum*col_width + col1_offset + 5);

		/* print the R value (Roll) */
		vrPrintf(ANSI_MOVETOROWCOL "R", (int)(euler_input.r[VR_ROLL]*rot_scale)+center, devnum*col_width + col1_offset + 6);

		/* Output the Quaternion values */
		vrPrintf(ANSI_MOVETOROWCOL "i", (int)(aux->incoming_trackers[VR_QX][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 7);
		vrPrintf(ANSI_MOVETOROWCOL "j", (int)(aux->incoming_trackers[VR_QY][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 8);
		vrPrintf(ANSI_MOVETOROWCOL "k", (int)(aux->incoming_trackers[VR_QZ][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 9);
		vrPrintf(ANSI_MOVETOROWCOL "W", (int)(aux->incoming_trackers[VR_QW][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 10);
		//vrQuatSet4d(&quat_input, aux->incoming_trackers[VR_QX][devnum], aux->incoming_trackers[VR_QY][devnum], aux->incoming_trackers[VR_QZ][devnum], aux->incoming_trackers[VR_QW][devnum]);

		vrPrintf(NORM_TEXT);	/* switch to normal mode (non-reverse) */

#  if 0
		/* print the joystick values */
		if (unitdata->has_button) {
			vrPrintf(ANSI_MOVE "x", (int)((unitdata->joystick[VR_X]+1)*qual_scale)+num_lines-1, unit*col_width + col1_offset+7);
			vrPrintf(ANSI_MOVE "y", (int)((unitdata->joystick[VR_Y]+1)*qual_scale)+num_lines-1, unit*col_width + col1_offset+8);
		}
#  endif
	}

	butval_offset = aux->num_dd_trackers * col_width + col1_offset;

	/*****************************************/
	/* Display button (2-switch) information */
	if (aux->num_dd_buttons > 0) {
		vrPrintf(ANSI_MOVE "buts: ", start_line + 1, butval_offset);
		for (devnum = 0; devnum < aux->num_dd_buttons; devnum++) {
			vrPrintf("%c", (aux->incoming_buttons[devnum] == 0 ? '.' : 'X'));
		}
	}

	/********************************/
	/* Display valuator information */
	if (aux->num_dd_valuators > 0) {
		float	valmin = -1.0;	/* temporary min/max values */
		float	valmax = +1.0;
		//printf(" vals: ");
		for (devnum = 0; devnum < aux->num_dd_valuators; devnum++) {
			vrPrintf(ANSI_MOVE "val %d: %6.3f ", start_line + devnum + 3, butval_offset, devnum, aux->incoming_valuators[devnum]);
#if 0 /* TODO: add min & max fields to the general vrValuator type */
			if ((vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]) < 0.0001) {
				/* When we don't have min/max values yet, then fill with dashes */
				vrPrintf(" %11s", "|---------|");
			} else {
			}
#endif
				percent = (aux->incoming_valuators[devnum]-valmin) / (valmax-valmin);
				vrPrintf(" %11s",
					(percent < 0.11 ? "|-        |" :
					(percent < 0.22 ? "| -       |" :
					(percent < 0.33 ? "|  -      |" :
					(percent < 0.44 ? "|   -     |" :
					(percent < 0.56 ? "|    +    |" :
					(percent < 0.67 ? "|     -   |" :
					(percent < 0.78 ? "|      -  |" :
					(percent < 0.89 ? "|       - |" :
							  "|        -|")))))))));
		}
	}

	/* return to the original cursor position */
	vrPrintf(ANSI_CURSORLOC_RETURN);
	fflush(stdout);
}


/**************************************************************************/
/* exit_testapp(): this function is used as a signal callback, such that  */
/*    when called, it simply sets the "done" flag, so the main while loop */
/*    in main() while know to terminate at the end of the current pass.   */
void exit_testapp()
{
	if (done < 0) {
		/* if this is the case, we haven't even begun the reading loop, so just exit */
		vrPrintf("\nVruiDD: Interrupted prior to entering the read-loop.\n");
		exit(1);
	}

	if (done == 1) {
		/* if we've already been here, this is a sign to force quit */
		printf("\nPersistent interruption... terminating.\n");

		exit(0);
	}

	done = 1;
	printf("\nInterruption -- if program doesn't close cleanly, try interrupting again.\n");
}


/*****************************************************************************/
/* A test program to communicate with a VruiDD device and print the results. */
int main(int argc, char *argv[])
{
static	char			*err_missing_arg = "%s: missing argument to '%s'.\n";
	_VruiDDPrivateInfo	*aux;
	char			*progname;			/* name of the program executable */
	int			count;				/* for looping over things */
	int			read_return;			/* return value from the read function */
	int			rendermode = RENDERLINES;	/* default: how to render the data */
	int			justonce = 0;			/* default: quit after just one output */
#if 0
	int			nodata_flag = 0;		/* default: whether or not to actual show input data */
#endif
	struct timeval		now;				/* to keep track of time since last report */
	vrTime			last_time;			/* time of last reported data */
	vrTime			time_delta = 0.0;		/* current time */

	done = -1;				/* not yet initialized (redundant) */
	signal(SIGINT, exit_testapp);


	/******************************/
	/* setup the device structure */
	aux = (_VruiDDPrivateInfo *)malloc(sizeof(*aux));
	memset(aux, 0, sizeof(*aux));
	_VruiDDInitializeStruct(aux, "default");


	/*********************************************************/
	/* set default parameters based on environment variables */
	aux->dd_host = getenv("VRUIDD_HOST");
	if (aux->dd_host == NULL)
		aux->dd_host = DEFAULT_HOST;		/* default, if no host given */
	aux->dd_port = ((getenv("VRUIDD_PORT")==NULL) ? -1 : atoi(getenv("VRUIDD_PORT")));
	if (aux->dd_port == -1)
		aux->dd_port = DEFAULT_PORT;		/* default, if no port given */

	aux->protocol = ((getenv("VRUIDD_PROTOCOL")==NULL) ? 0 : atoi(getenv("VRUIDD_PROTOCOL")));


	/*****************************************************/
	/* adjust parameters based on command-line-arguments */
	progname = argv[0];
	while ((argc > 1) && (argv[1][0] == '-')) {
		/****************/
		/* Set the port */
		if (!strcmp(argv[1], "-p") || !strcmp(argv[1], "-port")) {
			argv++; argc--;
			if (argc > 1) {
				aux->dd_port = atoi(argv[1]);
				argv++; argc--;
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-screen] [-nodata] [-once] [-p <port>(%d)] [-protocol <prot>(%d)] [<Vrui host> (default = '%s')]\n",
					progname, aux->dd_port, aux->protocol, aux->dd_host);
				exit(1);
			}
			continue;		/* return to the while */
		}

		/********************/
		/* Set the protocol */
		if (!strcmp(argv[1], "-prot") || !strcmp(argv[1], "-protocol")) {
			argv++; argc--;
			if (argc > 1) {
				aux->protocol = atoi(argv[1]);
				argv++; argc--;
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-nodata] [-p <port>(%d)] [-protocol <prot>(%d)] [<Vrui host> (default = '%s')]\n",
					progname, aux->dd_port, aux->protocol, aux->dd_host);
				exit(1);
			}
			continue;		/* return to the while */
		}

		/*****************************************************/
		/* Report information about the device and then quit */
		if (!strcmp(argv[1], "-nodata")) {
			/* NOTE: for other input-test applications we set a separate "nodata_flag" */
			/*   but that isn't necessary here, and the premature exit() was avoiding  */
			/*   the clean-disconnect with the Vrui VRDeviceDaemon.                    */
			done = 1;
			argv++; argc--;
			continue;		/* return to the while */
		}

		/***************************/
		/* set the "justonce" flag */
		if (!strcmp(argv[1], "-1") || !strcmp(argv[1], "-once")) {
			argv++; argc--;
			justonce = 1;
			continue;		/* return to the while */
		}

		/****************************************/
		/* set the "rendermode" flag to screens */
		if (!strcmp(argv[1], "-screen")) {
			argv++; argc--;
			rendermode = RENDERSCREEN;
			continue;		/* return to the while */
		}

		/******************/
		/* Unknown option */
		else {
			/* There are currently no other "-" options, so this is an error */
			/* NOTE: reported defaults might be effected by environment variables */
			fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-nodata] [-p <port>(%d)] [-protocol <prot>(%d)] [<Vrui host> (default = '%s')]\n",
				progname, aux->dd_port, aux->protocol, aux->dd_host);
			exit(1);
		}
	}

	/* if there are any arguments left, use the first as the daemon hostname */
	if (argc > 1) {
		aux->dd_host = strdup(argv[1]);
	}


	/**************************************************/
	/* open the socket port and initialize the device */
	aux->fd_socket = vrSocketCall(aux->dd_host, aux->dd_port);
	if (aux->fd_socket < 0) {
		aux->open = 0;
		fprintf(stderr, RED_TEXT "couldn't open socket '%s:%d'\n" NORM_TEXT, aux->dd_host, aux->dd_port);
		snprintf(aux->version, sizeof(aux->version), "- unconnected VruiDD -");
	} else {
		aux->open = 1;
		done = 0;
		if (_VruiDDInitializeDevice(aux) < 0) {
			vrErrPrintf("main: " RED_TEXT "Warning, unable to initialize VruiDD.\n" NORM_TEXT);
		}
	}

	/**********************/
	/* display the output */

	_VruiDDPrintPrivateInfo(stdout, aux, verbose);

#if 0
	/* quit if flagged to print the device info but not data */
	if (nodata_flag) {
		_VruiDDCloseFunction(&device);
		exit(0);
	}
#endif

#if 0
	/* for screen-rendering, scroll everything up so it won't be erased */
	if (rendermode == RENDERSCREEN && !justonce && aux->open) {
		for (count = 0; count < 60; count++)
			vrPrintf("\n");
		_RenderValuesScreen(aux);
	}
#endif

	/****************************************/
	/* read the data and display the output */
	while(aux->open && !done) {

		/* read any new data */
		read_return = _VruiDDReadInput(aux);

		/* if any new data was received then display it */
		if (read_return > 0) {

			/* save the time of last report */
			gettimeofday(&now, NULL);
			last_time = ((vrTime)now.tv_sec + (vrTime)now.tv_usec / 1000000.0);

			switch(rendermode) {
			default:
			case RENDERLINES:
				if (time_delta > 0.1)
					printf("\n");
				printf("[%7.4lf] ", 0.0);
				_RenderValuesLine(aux);
				break;;
			case RENDERCSV:
#if 0
				_RenderValuesCSV(aux);
#else
				vrPrintf("CSV output rendering mode not yet implemented.\n");
#endif
				break;;
			case RENDERSCREEN:
				_RenderValuesScreen(aux);
				break;;
			}

			/* we're done already if just one printout */
			if (justonce)
				done = 1;
		} else if (read_return == -1) {
			/* there was a communication error during the read, so end */
			done = 1;
		} else {
			/* Just report the amount of time since the last report */
			gettimeofday(&now, NULL);
			time_delta = ((vrTime)now.tv_sec + (vrTime)now.tv_usec / 1000000.0) - last_time;
			vrPrintf("[%7.4lf]\r", time_delta);
		}
	}

	/*****************/
	/* close up shop */

	if (rendermode == RENDERSCREEN && !justonce && aux->open) {
		/* clean up the screen rendering */
		vrPrintf(ANSI_SETMARGIN_TOPBOT_OFF);
		vrPrintf(ANSI_CURSORLOC_RETURN);
	}

	if (aux->open) {
		_VruiDDCloseDevice(aux);

		while (_VruiDDReadInput(aux) >= 0);	/* wait for the -1 response indicating the reception of the STOPSTREAM_REPLY from the daemon */
	}

	if (aux != NULL) {
		if (aux->open)
			vrSocketClose(aux->fd_socket);

		free(aux);
	}

	vrDbgPrintf(BOLD_TEXT "\nVruiDD device closed\n" NORM_TEXT);
}

#endif /* } TEST_APP */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(CAVE) /* { */


	/* ... CAVE stuff here if to also work with CAVElib */


#endif /* } CAVE */

