/* ======================================================================
 *
 *  CCCCC          vr_socket.c
 * CC   CC         Author(s): Bill Sherman, John Stone
 * CC              Created: October 22, 2001  (from code originally written in 1987)
 * CC   CC         Last Modified: September 1, 2014 (with minor comments 11/11/21)
 *  CCCCC
 *
 * Code file for FreeVR socket interfacing.
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/****************************************************************************/
/* Socket API by Bill Sherman                                               */
/*                                                                          */
/* This version is for TCP style sockets only.                              */
/*                                                                          */
/* ROUTINES:                                                                */
/*   int vrSocketCreateListen(port, block)                                  */
/*   int vrSocketCreateListenUDP(port, block)                               */
/*   void vrSocketClose(sock_fd)                                            */
/*   int vrSocketAwaitData(int fd)                                          */
/*   int vrSocketAnswer(sock_fd)                                            */
/*   int vrSocketCall(hostname, port)                                       */
/*   int vrSocketReadMsg(sock_fd, msg, msglen)                              */
/*   int vrSocketReadMsgClean(sock_fd, msg, msglen)                         */
/*   int vrSocketSendMsg(sock_fd, msg)                                      */
/*                                                                          */
/* HISTORY:                                                                 */
/*  1987-1988: original version(s) isockets.c and usockets.c written by     */
/*	Bill Sherman as alternative (IP/TCP vs. Unix) socket routines for   */
/*      a series of programs for card playing over the Internet (way back   */
/*      in 1987).  The Cards programs were written specifically to learn    */
/*      how to write sockets and (n)curses programs.  (But we tested them   */
/*      out by playing Hearts on Friday afternoons.)                        */
/*                                                                          */
/*  Feb 1998 (a decade later): renamed isockets.c to socket_api.c and       */
/*      refined the interface.  This was for my CAVE library tutorial,      */
/*      which eventually became my FreeVR library tutorial.                 */
/*                                                                          */
/*  10/22/2001: incorporated into FreeVR library as vr_socket.c, with       */
/*	functions renamed to fit the FreeVR naming scheme.                  */
/*                                                                          */
/*  11/09/2010: (Bill Sherman)                                              */
/*      Added a new "vrSocketCreateListenUDP()" function that creates a     */
/*      socket specifically to listen on a UDP port -- needed for the       */
/*      DTrack interface.  It's mostly identical from the non-UDP version,  */
/*      And I started my experiments with the old (1987) usockets.c code.   */
/*                                                                          */
/*  11/09/2010: (Bill Sherman)                                              */
/*	Enhanced some of the error reporting to be more informative.        */
/*                                                                          */
/*  06/18/2014: (Bill Sherman)                                              */
/*	Altered the response codes for vrSocketReadMsgClean() to work       */
/*      better with "pvtest".                                               */
/*                                                                          */
/*  09/01/2014: (Bill Sherman)                                              */
/*	Adjusted vrSocketReadMsgClean() to avoid a bug introduced by the    */
/*      change on 06/18/2014.  Now returns specific values for EOF and      */
/*      error results.                                                      */
/*                                                                          */
/* TODO:                                                                    */
/*   - Describe each function and its arguments in comments.                */
/*                                                                          */
/*   - Consider adding a version of SendMsg for non character strings.      */
/*                                                                          */
/*   - Consider whether I should use the recv() and send() system calls     */
/*     instead of read() and write.                                         */
/*                                                                          */
/*   - DONE: Re-investigate which include headers are actually required.    */
/*                                                                          */
/****************************************************************************/
#include <netdb.h>		/* needed for struct hostent */
#include <fcntl.h>		/* needed for fnctl() */
#include <string.h>		/* needed for strlen */
#include <strings.h>		/* needed for bzero */
#include <unistd.h>		/* needed for read() & write() */
#include <stdio.h>		/* needed for printf() & fprintf() */
#include <sys/types.h>	/* not needed on Linux, but good for portability */
#include <sys/socket.h>	/* not needed on Linux, but good for portability */
#ifdef __APPLE__
#  include <sys/file.h>		/* added by Matt Schechtman for Apple? */
#endif

#include "vr_socket.h"		/* primarily to make sure function args match */
#include "vr_debug.h"		/* needed for vrErrPrintf() */

/*************************************************************************/
/* vrSocketCreateListen(): ...    */
/* port is used to assign a port.  If port==0, then an unassigned socket */
/*   is chosen, and the result is passed back via the port variable.     */
/*************************************************************************/
int vrSocketCreateListen(port, block)
	int	*port;		/* port is returned via this CBR variable */
	int	block;		/* flag to indicate whether to block (1) or not (0) */
{
const	int		max_backlog = 5;	/* the max length of pending connections */
struct	sockaddr_in	server_addr;		/* Using the "sockaddr_in" strcture for InterNet protocol */
	int		sock_fd,
			fd_flags, fd_flags_block, fd_flags_noblock;
	socklen_t	sa_length = sizeof(server_addr);/* length of the "server_addr" structure [BS: 02/04/10 -- was unsigned int, changed for Onyx4 gcc] */

	/* Create socket */
	sock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (sock_fd < 0) {
		perror("vrSocketCreateListen(): opening stream socket");
		return -1;
	}

	/* set socket port to blocking or non-blocking as indicated */
	fd_flags = fcntl(sock_fd, F_GETFL, 0);
	fd_flags_block = fd_flags & ~FNDELAY;  /* Turn blocking on */
	fd_flags_noblock = fd_flags | FNDELAY; /* Turn blocking off */
	fcntl(sock_fd, F_SETFL, block ? fd_flags_block : fd_flags_noblock);

	/* Name socket using inet port scheme */
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = INADDR_ANY;
	server_addr.sin_port = htons(*port);
	/* NOTE: bind() takes the polymorphic type "sockaddr", so cast from "sockaddr_in" */
	if (bind(sock_fd, (struct sockaddr *)&server_addr, sizeof(server_addr))) {
		vrErrPrintf("vrSocketCreateListen(): Trouble creating a listening socket: %d\n", *port);
		perror("vrSocketCreateListen():" RED_TEXT " ** binding stream socket **" NORM_TEXT);
		close(sock_fd);
		return -1;
	}

	/* NOTE: getsockname() takes the polymorphic type "sockaddr", so cast from "sockaddr_in" */
	if (getsockname(sock_fd, (struct sockaddr *)&server_addr, &sa_length)) {
		vrErrPrintf("vrSocketCreateListen(): Trouble creating a listening socket: %d\n", *port);
		perror("vrSocketCreateListen():" RED_TEXT " ** getting socket name **" NORM_TEXT);
		return -1;
	}
	*port = ntohs(server_addr.sin_port);
	listen(sock_fd, max_backlog);
	return sock_fd;
}


/*************************************************************************/
/* vrSocketCreateListenUDP(): ...    */
/* port is used to assign a port.  If port==0, then an unassigned socket */
/*   is chosen, and the result is passed back via the port variable.     */
/* UDP sockets are Unix Datagram Protocol sockets -- they send packets   */
/*   of information in complete chunks, but their order is not preserved */
/*   and even their reception is not guaranteed.                         */
/*************************************************************************/
int vrSocketCreateListenUDP(port, block)
	int	*port;		/* port is returned via this CBR variable */
	int	block;		/* flag to indicate whether to block (1) or not (0) */
{
const	int		max_backlog = 5;	/* the max length of pending connections */
struct	sockaddr_in	server_addr;		/* Using the "sockaddr_in" structure for InterNet protocol */
	int		sock_fd,
			fd_flags, fd_flags_block, fd_flags_noblock;
	socklen_t	sa_length = sizeof(server_addr);/* length of the "server_addr" structure [BS: 02/04/10 -- was unsigned int, changed for Onyx4 gcc] */

	/* Create socket */
	sock_fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (sock_fd < 0) {
		perror("vrSocketCreateListenUDP(): opening stream socket");
		return -1;
	}

	/* set socket port to blocking or non-blocking as indicated */
	fd_flags = fcntl(sock_fd, F_GETFL, 0);
	fd_flags_block = fd_flags & ~FNDELAY;  /* Turn blocking on */
	fd_flags_noblock = fd_flags | FNDELAY; /* Turn blocking off */
	fcntl(sock_fd, F_SETFL, block ? fd_flags_block : fd_flags_noblock);

	/* Name socket using inet port scheme */
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = htons(*port);
#if 1 /* BS: 11/27/12 enabled this section -- seems to do better error reporting -- why wasn't this section being used? */ /* 11/11/21: BS: I'm not sure why the 2nd version (#if 0) isn't being used instead */
	/* NOTE: bind() takes the polymorphic type "sockaddr", so cast from "sockaddr_in" */
	if (bind(sock_fd, (struct sockaddr *)&server_addr, sizeof(server_addr))) {
		vrErrPrintf("vrSocketCreateListenUDP(): Trouble creating a listening UDP socket: %d\n", *port);
		perror("vrSocketCreateListenUDP():" RED_TEXT " ** binding stream socket **" NORM_TEXT);
		close(sock_fd);
		return -1;
	}
#else
int return_val;
	/* NOTE: bind() takes the polymorphic type "sockaddr", so cast from "sockaddr_in" */
	return_val = bind(sock_fd, (struct sockaddr *)&server_addr, sizeof(server_addr));
	if (return_val) {
		vrErrPrintf("vrSocketCreateListenUDP(): bind returns %d (UDP port: %d\n", return_val, *port);
		perror("vrSocketCreateListenUDP():" RED_TEXT " ** binding stream socket **" NORM_TEXT);
	}
#endif

	/* NOTE: getsockname() takes the polymorphic type "sockaddr", so cast from "sockaddr_in" */
	if (getsockname(sock_fd, (struct sockaddr *)&server_addr, &sa_length)) {
		perror("vrSocketCreateListenUDP():" RED_TEXT " ** getting socket name **" NORM_TEXT);
		return -1;
	}
	*port = ntohs(server_addr.sin_port);	/* port value to return by reference */
	listen(sock_fd, max_backlog);
	return sock_fd;
}


/******************************************************/
/* vrSocketClose(): ... */
/******************************************************/
void vrSocketClose(sock_fd)
	int	sock_fd;
{
	close(sock_fd);
	shutdown(sock_fd, 2);
}


/*******************************************************************/
/* This is a direct copy of vrSerialAwaitData(). */
/* NOTE: 05/05/21: I'm making this duplicate so the VRPN code can */
/*   make use of it without having to also include the vr_serial.c */
/*   source code.                                                  */
/* NOTE: I tried moving it to vr_utils.c to be more general, but I */
/*   wasn't liking how things were shaping up.                     */
int vrSocketAwaitData(int fd)
{
	struct timeval	awhile;
	fd_set		fds;

	awhile.tv_sec = 0;
	awhile.tv_usec = 5000000/*150000*/;	/* Wait up to .15 second */

	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	return select(fd+1, &fds, NULL, NULL, &awhile);
}


/************************************************************************/
/* vrSocketAnswer(): Gets a socket file-descriptor for an attempt to    */
/*   connect to a listening socket (if there is one).  Returns the      */
/*   file-descriptor number when connected, or "-1" when no connection  */
/*   is made (which is an acceptable situation for non-blocking sockets.*/
/* The two "NULL" arguments to the "accept()" call indicate that we are */
/*    not interested in the address of the connecting socket.           */
int vrSocketAnswer(sock_fd)
	int	sock_fd;
{
	int	asock;

	asock = accept(sock_fd, NULL, NULL);

	return asock;
}


/******************************************************/
/* vrSocketCall(): ... */
/******************************************************/
/* this routine is used by the client end of a connection */
int vrSocketCall(hostname, port /* , block */)
	char	*hostname;
	int	port;
#if 0
	int	block; /* not sure if this will be useful */
#endif
{
struct	sockaddr_in	client_addr;		/* TODO: should this be "sockaddr" type? */
struct	hostent		*hp,
			*gethostbyname();
	int		sock_fd;

	/* Create socket */
	sock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (sock_fd < 0) {
		perror("vrSocketCall(): opening stream socket");
		return -1;
	}

	/* Name socket using file system name */
	if (hostname == NULL) {
		vrErrPrintf("vrSocketCall(): No hostname specified, using 'localhost'.\n");
		hp = gethostbyname("localhost");
	} else {
		hp = gethostbyname(hostname);
	}
	if (hp == NULL) {
		vrErrPrintf("vrSocketCall(): %s: unknown host\n", hostname);
		return -1;
	}
	bcopy(hp->h_addr, &client_addr.sin_addr, hp->h_length);

	client_addr.sin_family = AF_INET;
	client_addr.sin_port = htons(port);

	if (connect(sock_fd, (struct sockaddr *)&client_addr, sizeof(client_addr)) < 0) { /* TODO: why is this "sockaddr", when the type is "sockaddr_in" ?? */
		close(sock_fd);
		vrErrPrintf("vrSocketCall(): Trouble connecting to '%s':%d\n", hostname, port);
		perror("vrSocketCall(): connecting stream socket");
		return -1;
	}
	return sock_fd;
}


/******************************************************/
/* vrSocketReadMsg(): ... */
/******************************************************/
int vrSocketReadMsg(sock_fd, msg, msglen)
	int	sock_fd;
	char	*msg;
	int	msglen;
{
	int	rval;

	bzero(msg, msglen);

	rval = read(sock_fd, msg, msglen);

	return rval;
}


/***************************************************************/
/* vrSocketReadMsgClean(): ...                                 */
/* same as vrSocketReadMsg, but strips tail '\n'               */
/*   06/18/2014 -- fixed two related bugs:                     */
/*     1) subtracting two from rval after erase two characters */
/*     2) not erasing characters if no message received.       */
/* CAUTION: the return values aren't consistent with vrSocketReadMsg() */
/*   Here:                                                             */
/*     * EOF (-1) indicates an EOF -- closed socket                    */
/*     * -2 indicates an error in the read                             */
/*     * 0 (which normally means EOF) indicates a 0-length message     */
/***************************************************************/
int vrSocketReadMsgClean(sock_fd, msg, msglen)
	int	sock_fd;
	char	*msg;
	int	msglen;
{
	int	rval;

	bzero(msg, msglen);
	rval = read(sock_fd, msg, msglen);

	/* return immediately if EOF received */
	if (rval == 0)
		return EOF;

	/* return immediately if a valid message wasn't received */
	if (rval < 0)
		return -2;	/* NOTE: normally -1 is the error signal, but that's the actual value of EOF, so we have to shift. */
	
	/* clean off trailing '\n' */
	if (msg[rval-1] == '\012') {
		rval--;
		msg[rval] = '\0';
	}
	if (msg[rval-1] == '\r') {
		rval--;
		msg[rval] = '\0';
	}

	return rval;
}


/******************************************************/
/* vrSocketSendMsg(): ... */
/******************************************************/
int vrSocketSendMsg(sock_fd, msg)
	int	sock_fd;
	char	*msg;
{
	int	rval;

#if 0 /* this is basically a nop (or should be) but for some reason causes a seg-fault on Linux */
	/* TODO: figure out why this seg faults on Linux */
	msg[strlen(msg)] = '\0';
#endif

#if 0
	rval = write(sock_fd, msg, strlen(msg)+1);		/* 06/15/2006: Why the "+1" -- why send the trailing '\0'?  Today it's causing me problems with the Tcl/TK interface. */
#else
	rval = write(sock_fd, msg, strlen(msg));
#endif
	if (rval < 0)
		perror("vrSocketSendMsg(): unable to send message");

	return rval;
}

