/* ======================================================================
 *
 *  CCCCC          vr_telnet.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: October 22, 2001
 * CC   CC         Last Modified: May 12, 2021 (minor fixes 09/14/21 & 09/18/21 & 10/09/21 & 11/11/21 & 03/30/22)
 *  CCCCC
 *
 * Code file for FreeVR telnet communications process.
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

USAGE:
	Making a connection:
		% telnet <host> <port>
	      [ Enter FreeVR password> <#######> ]
		FreeVR> <command> <arguments>
		...
		FreeVR> ^]
		telnet> quit
		%

	Making a one-off connection:
		% echo 'echo $version' | nc localhost 3000
		% echo "echo \$version" | nc localhost 3000
		% echo 'machine; print ui' | nc localhost 3000
		% echo "set inputdevice[1] nudget2rwT 2 0 1" | nc localhost 3000
		% echo 'term' | nc localhost 3000

		NOTE: most shells require the '$' to be escaped when not in
			single quotes -- ie. the $version in the first example
			 would be first interpreted as a shell variable.

		When there's a password:
		% echo -e '<pw>\0echo $version' | nc localhost 3000

	Connecting via software:
		It is also possible to connect via TCL or Python to create
		a specific GUI interface for manipulating a running FreeVR
		application on the fly.  The included "fvri.tcl" (FreeVR Remote
		Interface) TCL program demonstrates this.


	Possible commands:
		"help" -- print a list of possible commands
		"quit" -- quit the telnet process (further socket
			connections will NOT be possible).
		"close" -- close the socket connection (further socket
			connections to the same port WILL be possible).
	  :-}	"term" -- quit the entire FreeVR application (works by kludge).
	  :-(	"reset" -- reset something (not sure what).
		"nudge" -- listing of the nudge commands (ie. specialized help).
		"history [<n>]" -- show the past <n> (or all) previous commands.
		"!! | !<n> | !<-n> " -- repeat a numbered (or last) command.
		"pause" -- toggle the system pause status
		"setenv <variable name> <value>" -- set an environment variable
		"unsetenv <variable name>" -- unset an environment variable [NOTE: doesn't work under IRIX]
		"echo <string>" -- echo a string (includes variable evaluation)
			possible variables are:
				[NOTE: in man page mention that these aren't specific to the FreeVR
				"telnet" operation, but that this is where users are most likely to
				encounter them.]
			- $version -- short name of the version of the library
			- $fullversion -- longer name of the library version
			- $machine | $hostname -- name of the machine application is running on
			- $arch | $binaryformat -- architecture
			- $compile -- compile string that includes information about when the library was
				compiled and information about the operating system it was compiled on.
			- $name -- name of the FreeVR application
			- $author -- author(s) of the FreeVR application
			- $appstatus -- status of the current FreeVR application
			- $freevrhomedir -- the system location where config and other files can be found
			- $memsize -- size of the shared memory segment (in bytes)
			- $status -- current status of the library (running, ended, etc.)
			- $startup_error -- a bitwise flag of errors that may have occurred on startup (0 is good)
			- $time_immemorial -- the (unix-style) time of when the library was started
			- $sim_time | $time -- the number of seconds that the simulation has been running
				(NOTE: paused time is subtracted out of this)
			- $wall_time -- the current time as reported by the system's operating system
			- $debug_level -- the current value at (and below) which debugging output will be provided
			- $debugthistoo -- a single additional debug level that will be reported
			- $system -- the freevr system object that is active
			- $def_visrenmode -- the default visual rendering mode
			- $win_style -- type of windows used by this library (GLX vs. MSWindows)
			- $gfx_style -- type of graphics library (currently only "OpenGL")
			- $sem_style -- method of semaphores implementation
			- $sock_style -- method of sockets interface implementation (usually "BSD")
			- $mp_style -- the type of multi-processing API used by this FreeVR build (e.g. "FORK")
			- $compile_options -- the "-D" flags set when the FreeVR library was built
			- $<env var name>
		"printstyle" | "style" [<style>] -- set (or print when no argument) the current print style
			NOTE: keep the following commands as "aliases" that still work.
		"verbose" -- set print style to verbose mode.
		"file" -- set print style to config file format mode.
		"brief" -- set print style to brief mode.
		"one_line" -- set print style to one_line mode.
		"machine" -- set print style to machine mode. (including disabling the prompt)
		"diagram" -- set print style to diagram (GraphViz-dot) mode
		"print" | "p" <query> -- print values from running application.
		"printfile" | "pf" [+]<filename> <query> -- print values from running application to a file
			possible print and printfile requests are:
			- "help" -- the list of what can be printed
			- "context" -- the context structure
			- "shmem" -- information about the shared memory system
			- "config" -- the configuration structure
			- "inputinfo" -- the overarching vrInputInfo structure
			- "system" -- the system structure (of the running system)
			- "settings" -- print the system settings values
			- "defaults" -- the system default settings structure (of the running system)
			- "near" -- the near clipping plane value
			- "far" -- the far clipping plane value
			- "procs" -- a summary of the running processes (or threads)
			- "proc[<num>]" -- detailed information about the Nth process (or thread)
			- "ui" -- a summary of all the inputs' user interface functionality (requires application-programmer specification)
			- "inputs" -- a summary of all the system's inputs
			- "input <name>" -- detailed information about a particular (named) input
			- "inputdevices" -- print info about all inputs devices
			- "inputdevice[<num>]" -- print info about the Nth input device
			- "windows" -- print a summary of all the windows
			- "window[<num>]" -- print info about the Nth window
			- "eyelists" -- print info about all eyelists
			- "eyelist[<num>]" -- print info about the eyelist
			- "eyes" -- print info about the eyes of the system
			- "eye[<num>]" -- print info about the Nth eye
			- "users" -- print info about all the active users
			- "user[<num>]" -- print info about the Nth user
			- "object" | "o" <type> list -- list all <type> config objects
			- "object" | "o" <type> <name> -- print info about a config object
				NOTE: <type> can be "address" for which <name> is a memory location
		"set" | "s" <request> -- set some particular setting to the given value.
			possible set-requests are:
			- "input" <variable> <value> -- set a particular input to the given value.
				"<value>" is handled in vr_input.c:
				- "id" (set to the identity matrix)
				- "loc [<x|*> [<y|*> [<z|*>]]]" ('*' means no-change)
				- "move [<x> [<y> [<z> [<azim> [<elev> [<roll]]]]]]" (ungiven values are 0.0)
				- "{ nudger2eA | nra } { id | <x> <y> <z> }" (ungiven values are 0.0)
				- "{ nudger2eT | nrt } { id | <x> <y> <z> }" (ungiven values are 0.0)
			- "inputdevice[<num>] <setting>"
				possible <setting> options:
				- "{ nudget2rwA | nta } { id | <x> <y> <z> }"
				- "{ nudget2rwT | ntt } { id | <x> <y> <z> }"
			- "pause" {0,1} -- pause the entire system.
			- "debuglevel|dl" <value> -- set the global debug level.
			- "debugthistoo|dtt" <value> -- set the global debug-this-too.
			- "debuglevel|dl[<num>]" <value> -- set process <num>'s debug level.
			- "debugthistoo|dtt[<num>]" <value> -- set process <num>'s debug-this-too.
			- "near <value>" -- set the near clipping plane for all (active) windows
			- "far <value>" -- set the far clipping plane for all (active) windows
			- "window[<num>] fps" {0,1} -- turn off/on a window's fps display.
			- "window[<num>] fps_loc" <v1>,<v2> -- relocate a window's fps display.
			- "window[<num>] fps_color" <v1>,<v2>,<v3> -- set color of a window's fps display.
			- "window[<num>] stats" {0,1} -- turn off/on a window's stats display.
			- "window[<num>] frame" {0,1} -- turn off/on a window's frame display in all windows.
			- "window[<num>] world" {0,1} -- turn off/on a window's virtual world display.
			- "window[<num>] frm" <value> -- set the front rendering mode -- currently s/b {0,0x1B00,0x1B01,0x1B02}
			- "window[<num>] brm" <value> -- set the back rendering mode -- currently s/b {0,0x1B00,0x1B01,0x1B02}
			- "window[<num>] nudgex" <value> -- shift the window along the X-axis
			- "window[<num>] nudgey" <value> -- shift the window along the Y-axis
			- "window[<num>] nudgez" <value> -- shift the window along the Z-axis
			- "window[<num>] sim" <cmd> -- move or alter the simulator view of a window
				WARNING: currently there is no check to limit this to simulator windows
				possible move-commands are:
				- "away"
				- "toward"
				- "center"
				- "sethome"
				- "gohome"
				- "posy"
				- "negy"
				- "posx"
				- "negx"
				other (non-movement) simulator commands:
				- "toggle"
				- "togglebit" <bit number>
				- "mask" <maskval>
			- "user[<num>] iod" <v1> -- set the iod of a user.
			- "user[<num>] color" <v1>,<v2>,<v3> -- set the color of a user.
	   :-(		- "user[<num>] nudgeT" <v1>,<v2>,<v3> -- translate the user's travel.
	   :-(		- "user[<num>] nudgeA" <v1>,<v2>,<v3> -- rotate the user's travel.
			- "proc[<num>] end" <v1> -- set the req_proc_term flag of a process
			- "proc[<num>] term" <v1> -- set the proc_term flag of a process -- 11/19/2020: is this wise?
			- "proc[<num>] usec" <v1> -- set the usec_min value of a process
			- "proc[<num>] printcolor" <v1> -- set the print_color flag of a process
			- "proc[<num>] stats_calc" <v1> -- set the flag of whether to calculate stats\n"
			- "proc[<num>] stats_show" <v1> -- set the flag of whether to show these stats\n"
			- "proc[<num>] stats_mask" <v1> -- set the mask of which statistics to show\n"
	   :-|		- "proc[<num>] stats_xloc" <v1> -- set the x location of where to put stats\n" NOTE: the parsing is implemented, but the feature itself isn't.
			- "proc[<num>] stats_yloc" <v1> -- set the y location of where to put stats\n"
	   :-|		- "proc[<num>] stats_width" <v1> -- set the width of the stats display\n" NOTE: the parsing is implemented, but the feature itself isn't.
			- "proc[<num>] stats_top" <v1> -- set the top timeline of the stats display\n"
			- "proc[<num>] stats_interval" <v1> -- set the horizontal time intervals of the stats display\n"
			- "proc[<num>] stats_scale" <v1> -- set the vertical scale of the stats display\n"
			- "proc[<num>] stats_color" <v1>,<v2>,<v3> -- set the background color of the stats display\n"
			- "proc[<num>] stats_opac" <v1> -- set the background opacity of the stats display\n"
		"lock" | "l" <request> -- send a request to a FreeVR lock
			[From the help message]
			help|? -- print this message (a list of possible setting requests)
			num -- print the number of locks in the system
			print|p <lock-address> -- print the state of the lock
			printall|pa -- print the state of all the locks
			trace|t <lock-address> <0|1> -- set the lock's tracing flag
			setwrite|sw <lock-address> -- wait for and set the write state for the lock
			relwrite|rw|uw <lock-address> -- release the write state for the lock
			setread|sr <lock-address> -- wait for and set the read state for the lock
			relread|rr|ur <lock-address> -- release the read state for the lock
			create -- create a new lock
		"barrier" | "b" <request> -- send a request to a FreeVR barrier
			[From the help message]
			help|? -- print this message (a list of possible setting requests)
			num -- print the number of locks in the system
			print|p <barrier-address> -- print the state of the barrier
			printall|pa -- print the state of all the barriers
			sync <barrier-address> -- wait for all members to be at the barrier
			relonce <barrier-address> -- prematurely release the barrier
			create -- create a new barrier
			increment <barrier-address> -- add one member to the barrier
			decrement <barrier-address> -- subtract one member from the barrier
		"relbar" <barrier-address> -- shortcut to release a barrier at the given address


	FreeVR configuration options for telnet-type processes:
		"port" -- TCP port to request
		"portrange" -- number of (consecutive) ports to check.
			If 0 or less, then wait for specified port only.
		"prompt" -- prompt to use (sent to socket clients)
		"password" -- a password that must be typed to access the
			running application.
			NOTE: this is currently stored "in the clear" in the config file
			TODO: make use of PGP to store the public code in the config file

	Application control:
	  :-(	[NYI] a callback function that allows the application first
		dibs at parsing commands that come over the telnet socket.
		Perhaps totally overriding the built-in parsing, or perhaps
		just done first, followed by the built-in work.  If the former,
		then there needs to be a function that can be called to do the
		built-in stuff. [probably the new _TelnetParseCommand() function
		can do that.]


HISTORY:
	22 October 2001 (Bill Sherman)
		Obtained initial code ordering from vr_visren.c.  (Not all of
		  it is needed since the "telnet" type processes don't, yet,
		  have "objects").  Obtained socket parsing code from my
		  example VR application (ex8_socket.c).  Also copied and made
		  minor changes to my socket_api.c file -- called vr_socket.c
		  in the FreeVR library source.

	23 October 2001 (Bill Sherman)
		Added the "print" command, and the corresponding vrInfoGet()
		  function to respond to such queries.  This included the use
		  of fdopen() to convert a standard file descriptor to a stream
		  version that can be used with existing functions that expect
		  a stream as the argument.

	1 November 2001 (Bill Sherman)
		Added the "set" command to set the value of a given variable.
		  Wrote vrInfoSet() function to handle the dirty work.

	5-6 November 2001 (Bill Sherman)
		[... (see diary for now)]

	20 April 2002 (Bill Sherman)
		[... (see diary for now)]

	22 April 2002 (Bill Sherman)
		Added sim move command.

	6 January 2003 (Bill Sherman)
		Added the ability to set the "req_end_term" (nee "end_proc")
		  and "proc_term" (nee "proc_done") fields for a process.
		Also fixed a minor bug in the invalid user error message that
		  gave feedback using the number of windows.

	7 January 2003 (Bill Sherman)
		Implemented the "printobject" command which allows the user
		  to see any of the configuration objects regardless of whether
		  they are used by the current system or not.  A list of all
		  the names of a particular class of object can also be
		  obtained by giving the string "list" as the object name.

	8 January 2003 (Bill Sherman)
		Added "echo" command.

	13 January 2003 (Bill Sherman)
		Added "setenv" command.

	5 March 2003 (Bill Sherman)
		Added "relbar" command.

	7 March 2003 (Bill Sherman)
		Implemented a new "lock" set of commands that can be used
		  to control existing locks in the FreeVR system, or a new
		  test lock can be created.  The use of this is primarily
		  indented for the debugging of locks.  Did pretty much the
		  same for FreeVR "barriers".

	11 March 2003 (Bill Sherman)
		Added setting of a process' "print_color".

	23 April 2003 (Bill Sherman)
		Added code to print the auxiliary data of specific windows.
		Also went with a system wide change of the "opaque" fields
		  to now be called "aux_data".

	13 May 2003 (Bill Sherman)
		Added code to change parameters of the rendering statistics
		  display.  Also to change the "usec_min" time for a process.
		And, because the later also allows this process to have a
		  minimum time per frame, deleted the "loop_delay" field in
		  _TelnetPrivate.

	29 May - 2 June 2003 (Bill Sherman)
		Added commands to set the statistics values associated with
		  each process.  Changed the nature of fps_color to use floats
		  rather than unsigned bytes.
		Added "set pause <n>" and "pause" commands.

	May 2006 (Bill Sherman)
		Added commands to keep up with new settings (including
		  user-interface display settings, and rendermode settings).

	June 13-19 2006 (Bill Sherman)
		Lots of changes, all primarily with the goal of aiding the
		  interface to a "machine" script program to display/alter
		  features of the running application.
		Changes include making a slight alteration to the response
		  to a "windows" & "users" commands when "machine" printstyle
		  has been requested.  Also, improved the command parsing to:
			- allow for multiple commands in a single read
			- use semicolon or newline as command separator
			- ignore beginning whitespace in command

	May 5-6 2013 (Bill Sherman)
		Added the ability to affect the configuration values for
		  window objects that are not in the active configuration
		  (with "object window").  (This is done to turn on the new
		  "showFrame" feature for windows and have them appear in
		  the simulator view.)

	June 12 2013 (Bill Sherman, very early in the  AM)
		Added the ability to nudge window placement in any cardinal
		  direction with the "nudgex", "nudgey" and "nudgez" window
		  commands.

	2 September 2013 (Bill Sherman)
		Adopting a consistent usage of str<...>cmp() functions
		  (using logical negation).
		Also, though not coded here in vr_telnet.c, I fixed the
		  ability to alter the value of a 6-sensor input by fixing
		  the vrAssignGenericInput() function in vr_input.c
		  [reported on 09/03/2013 impl_diary]

	15 September 2013 (Bill Sherman)
		Changed the "0x%p" format to the improved "%#p" format.

	24 August 2014 (Bill Sherman)
		Split the vrTelnetMainLoop() function into separate functions
		  for vrTelnetInitProc(), vrTelnetOneFrame() and then still the
		  MainLoop.  This is to enable merging processes into a single
		  thread.  (Also fixed a couple of bugs along the way.)

		[corresponds to Freevr Version 0.6f]

	19 November 2020 (Bill Sherman)
		Minor change of variable "end_proc" to "req_end_proc" (i.e.
		  *request* to end the process), and "proc_done" to "proc_term".
		Also added two new headers ("vr_parse.h" & "vr_shmem.h").

	9-10 April 2021 (Bill Sherman)
		I added the ability to "nudge" the "t2rw_xform" of an
		  inputdevice to the available commands.  (And on the input
		  object side I enhanced the "r2e_xform" nudging, but the code
		  for that is actually in "vr_input.c".

		I also (finally!) added some code to make it such that I don't
		  have to wait 60 seconds between quitting and reconnecting to
		  the socket again!  To prevent a socket from being held, it's
		  the socket SO_LINGER setting that needs to be changed.

		(Though I also learned that quitting from the telnet side
		  doesn't cause a delay -- ^]quit.)
		[I'm surprised this wasn't on the todo list.]

		I did some cleanup and clarifications of output text.

		I added some additional todo-commands (e.g. nudge user travel)

		I explored the use of netcat/nc and put some basic instructions
		  for that in the USAGE top-comment.  (Should be included with
		  the man-page when it's written.)

	12 April 2021 (Bill Sherman) -- "history & !"
		I added the new "history" command.  I was going to say that I
		  haven't yet actually bothered to implement a history-repeat(!)
		  command, but then I went ahead and did that, which required a
		  small refactoring to remove material from vrTelnetOneFrame()
		  into a new __TelnetParseCommand() function.  For now I only
		  record the first 1024 commands into the history.  If this
		  becomes a problem I can either raise the number or make a
		  rolling array loop -- but that would complicate the code a
		  bit.

		NOTE: this "history" list is persistent -- if one disconnects
		  and then reconnects via telnet (or someone else reconnects),
		  the past history list still exists.  Not sure whether I'd
		  consider this a bug or at least a serious bug, but might be
		  worth mentioning in the man-page once it's written.

	13 April 2021 (Bill Sherman)
		I did a lite more cleanup on the new "history" mechanism.
		Also, I added a new "nudge" command which is really just a
		  specialized help command that shows all the possible nudge
		  operations -- which are really all "set" commands.  Perhaps
		  in the future I'll just have a specialized "nudge" command
		  for all those operations.

	22 April 2021 (Bill Sherman)
		I added a relative history reexecute option (negative number).
		I fixed a small history bug where the "reexecuting" message
		  wasn't being flushed.

		[corresponds to Freevr Version 0.6g]

	4 May 2021 (Bill Sherman)
		I did the "delete me" clean-ups.

	5 May 2021 (Bill Sherman)
		Minor clean-up to remove sprintf() and make sizeof() safer.

		[corresponds to Freevr Version 0.7a]

	7-12 May 2021 (Bill Sherman)
		Wrote the "telnet.7fv" manpage.  As part of that enhanced
		  some of the comments.  Plus added "?" as a synonym for "help"
		  in the *print* and *set* commands.

		[corresponds to Freevr Version 0.7b]

	14 September 2021 (Bill Sherman)
		I made a minor fix to compile on Cygwin -- strchrnul() issue
		  similar to an issue with the Apple (or any non-GNU) compiler.

		[corresponds to Freevr Version 0.7c]

	16-17 September 2021 (Bill Sherman)
		I added the new "diagram" print style.  Also fixed a bug
		  with the "printfile" command -- was crashing.  Also I added
		  help information about the new simulator/window mask setting
		  commands.  Finally, I added the "printstyle|style" command
		  that was in the TODO list.  (I also updated the "telnet.7fv"
		  file.)

		[corresponds to Freevr Version 0.7c]

	18 September 2021 (Bill Sherman)
		I fixed two hidden bugs in the socket reading loop that only
		  appeared when running CQ3A.

		[corresponds to Freevr Version 0.7c]

	09 October 2021 (Bill Sherman)
		I fixed the #if's for the strchrnul() calls which aren't
		  available on Apple MacOS or Cygwin.  I had the conditionals
		  backward for some reason!

	11 November 2021 (Bill Sherman)
		I added calls to close the Telnet sockets when doing a "term"
		  command, which has the effect of immediately closing the
		  connection with telnet.  (Prior to this, you had to hit
		  "return" in the telnet window for it to discover the process
		  was gone.)
		Also, I standardized the formatting of these HISTORY top-comments.

	30 March 2022 (Bill Sherman)
		I fixed "nudge" help command -- it was missing the window
		  rotation nudges
		I fixed the window nudging -- now when I rotate a window via a
		  telnet command, if the frame of that window is shown in the
		  simulator, the frame will correctly readjust.  (Part of the
		  issue is that all the transformations needed to be post-
		  multiplied instead of pre-multplied.)

TODO:
	DONE: Write a "telnet.7fv" manpage.

	DONE: Make "?" a synonym for "help" in the *print* and *set* commands.

	DONE: There probably should be a way to query the current print-style.

	DONE: Add a simulator mask setting capability.

	Print out the actual input map.  (This would be a good place to start
		for just figuring out a good way to permit configuring the
		input map.)

	Figure out whether there might be a way to have the socket reading
		code not block -- ie. do something if there's input, but
		otherwise not wait.  This would be important for the
		single-thread variant.

	Consider allowing names in addition to numbers in the operations that
		operate on a specific object.  E.g.:
			> set inputdevice[vrpnhydra] nta id
		  vs	> set inputdevice[1] nta id

	MOSTLY: Allow all objects in configuration to have values set.  (NOTE:
		I made this work for window objects, so just need to duplicate
		that code-skeleton, and should be easy to make it work for
		any of the object types.)

	Ability to set a value for ALL of the windows/procs/etc. (eg. "stats")
		An example might be:
			FreeVR> s proc[*] stats_calc 0

	Consider an alternative to the "printfile" command to simply add
		"> 'filename'" or ">> 'filename'" to the end of a print
		statement.

	Commands to be implemented:
		- "reset" ?? (I'm not sure what we would reset)
		- open windows for operation (or window processes)
		- start inputs (or input processes)
		- or maybe the above two are just "start-process <process object>"
			which might be a visren process or an input/output process
		- remap existing inputs
		- redirect -- redirect telnet output to file rather than
			back through the socket.  Also need an option to
			end the redirection and close the file.
		- user-travel nudge movements
		- DONE: close operating windows (or window processes)
			(simply set the "end" value of a visren process)
		- DONE: pause -- not sure how much to pause, perhaps on a
			process by process selectivity, with the option to
			pause all.  (This may require some additional code
			in vr_procs.c, and perhaps a new flag in vr_procs.h
			-- or just a way to block a barrier.)

	The other queries commented out in the vrInfoGet() help_msg.
		* Especially "inputbyname" !  [09/03/13]

	Implement an "alias" command -- and make the current implementation
		of "term" a built in alias.  Aliases can be specified in
		the config file, perhaps with an option to read some initial
		telnet commands from another file -- which would often just
		have aliases.  Also, allow them to be set in the .freevrrc
		configuration file.

	MOSTLY: Parse the variables out of the command string for all commands,
		not just the "echo" command -- which will greatly simplify the
		echo command.  Thus a variable can be set for complex
		expressions, and allow the user to refer to the variable
		instead of the expression every time.  This same function
		could also be used to handle the history and aliasing
		mechanisms.

		[I don't know that I'm actually doing this yet, but the
		variable parsing part is done.]

	Figure out how to encode a password in the config file (and also
		therefore, what command can be used by the administrator
		to create the proper string.  (Allow option to require the
		password of the user running the app.)  [Probably a PGP method
		would work -- of course that would probably require an external
		library (unless the code is simple enough to just embed it.]

	When requesting password from socket, disable echo.

	Queries that access the vrFprint<object>() functions should be able
		to pass an optional argument of the level of detail that
		should be printed (i.e. the vrPrintStyle).
		NOTE: Partially solved by having a default print style that
		is passed.
		[and can be set via "brief", "verbose", etc commands.]

	PARTIAL: Fix "term" command.  Current, somewhat kludgey, fix was
		implemented by setting the first 2-way input switch to 1,
		which is the standard way for apps to quit.

	DONE: A command history mechanism (even simple will do).

	DONE: A new "machine" print style that can be set.
		[NOTE: the vr_telnet.c part of this has already been
		implemented, but now it requires all the object print
		functions to handle the new style.]
		[06/19/2006: at least as done as this file is concerned.  As
		new needs arise due to the interface with the Tcl/Tk (or
		other) scripts, they will be implemented in the appropriate
		(other) file.

	PARTIAL: Pass vrContext around through arguments rather than as a global.
		[06/19/2006: only place left to do is vrTelnetMainLoop() -- which
		will leave only the vrTelnetSignalHandler() function, which may
		not be doable.]

*************************************************************************/
#include <stdio.h>	/* standard input and output functions! */
#include <stdlib.h>	/* needed for atof() */
#define __USE_GNU	/* needed for string.h to include strchrnul() */
#include <string.h>	/* needed for strchr() */
#include <signal.h>	/* needed for signal() and associated #defines */
#include <sys/socket.h>	/* needed for "struct linger" data type */

#include "vr_socket.h"
#include "vr_debug.h"
#include "vr_procs.h"	/* needed for proc_info */
#include "vr_config.h"	/* needed for vrContext and callbacks functions */
#include "vr_utils.h"	/* needed for vrSleep() */
#include "vr_objects.h"	/* needed for vrObjectSearch() */
#include "vr_parse.h"	/* needed for vrArgParse{Integer,String}() */
#include "vr_shmem.h"	/* needed for vrFprintShmemInfo() */

#define	TAB	"\t"

/* static (file-scope) globals for the (TODO: each?) telnet process */
static vrProcessInfo	*proc_info = NULL;	/* TODO: is this nece since we also have vrThisProc? */


/************************************************************************/
/** A structure of auxiliary information for a particular *PROCESS*.  **/
typedef struct {
		int		listen_sock;	/* socket to receive commands */
		int		cmd_socket;	/* socket from which to get commands */
		FILE		*cmd_fp;	/* stream file-pointer for command socket */
		int		inet_port;	/* port number to which one should connect */
		int		inet_port_range;/* the number of other (consecutive) ports to try */
		vrPrintStyle	style;		/* the default print style for some queries */
		char		host_machine[256]; /* name of the machine we're running on */
		char		*prompt;	/* command prompt for incoming telnets */
		char		*password;	/* a password that must be given to access socket */

		char		inbuf[1024];	/* used to collect incoming message */
		char		*parse;		/* pointer to where in inbuf to parse from */
		char		*history[1024];	/* array of strings (char *) with history of input (successfully parsed, or not) */
		int		num_history;	/* number of strings in history array */
	} _TelnetPrivate;


/* ===================================================================== */
/* NOTE: the "telnet" type of process (currently, 10/22/01) has no       */
/*   "object", so the object-handling routines that are at the beginning */
/*   of other "process" types such as "visren" and "input" are not       */
/*   present here.                                                       */
/* ===================================================================== */


/******************************************************************/
/* vrTelnetSignalHandler(): on a SIGINT signal, this process will */
/*   close down all things it's responsible for and exit.         */
/******************************************************************/
static void vrTelnetSignalHandler(int which)
{
	switch (which) {
	case SIGUSR2:
		vrMsgPrintf("vrTelnetSignalHandler(): Hey, input process got a USR2 signal\n");
		vrCallbackInvoke(vrContext->callbacks->HandleUSR2);
		break;

	default:
		/* for all other signals, assume death */
		vrMsgPrintf("vrTelnetSignalHandler(): (Telnet proc %d) " RED_TEXT "dying with %s\n" NORM_TEXT, getpid(), vrSigName(which));

		vrMsgPrintf("vrTelnetSignalHandler(): Time of death = %f, number of frames = %d\n",
                        proc_info->frame_wtime, proc_info->frame_count);

		vrMsgPrintf("vrTelnetSignalHandler(): Shared memory usage:%7ld (%ld freed)\n",
			vrShmemUsage(), vrShmemFreed());

		proc_info->req_proc_term = 1;
		vrFprintProcessInfo(stdout, proc_info, verbose);

#ifdef COREDUMP_IS_GOOD
		vrFprintf(stderr, "vrTelnetSignalHandler(): YO aborting in vr_input.c\n");
		abort();
#endif

		/* PAUSE to allow debugger to attach to the process. */
		vrMsgPrintf("vrTelnetSignalHandler(): Process pausing, use 'gdb -p %d' to debug.\n", getpid());
		pause();
	}
}


/***********************************************************/
/* vrInfoGet(): print a response to the given query to the */
/*   given FILE stream.  See the help_msg string for a     */
/*   list of possible queries.  Passing "help" sends this  */
/*   list.                                                 */
/* NOTE: this function basically takes the input string    */
/*   with the "print" or "p" (or whatever) stripped off.   */
/***********************************************************/
/* NOTE: this function (and vrInfoSet(), which follows) could easily */
/*   be moved to another source file, as they don't have much to do  */
/*   with telnetting, other than this is the only process that makes */
/*   use of it at the moment.                                        */
void vrInfoGet(vrContextInfo *context, FILE *file, char *query, vrPrintStyle style)
{
static	char	*help_msg = "Available queries:\n"
			TAB "help|? -- print this list of possible queries\n"
			TAB "context -- print the context structure\n"
			TAB "shmem -- print info about the shared memory system\n"
			TAB "config -- print the configuration structure\n"
			TAB "inputinfo -- the overarching vrInputInfo structure\n"
			TAB "system -- print the system structure\n"
			TAB "settings -- print the system settings values\n"
			TAB "defaults -- print the system default values\n"
			TAB "near -- the near clipping plane value\n"
			TAB "far -- the far clipping plane value\n"
			TAB "procs -- print a summary of all the processes\n"
			TAB "proc[<num>] -- print the structure of the Nth process\n"
			TAB "ui -- prints the inputs' user-interface functionality\n"
			TAB "inputs -- print a list of the available inputs\n"
			TAB "input <name> -- print info about a particular input\n"
		/* :-(	TAB "inputbyname <name> -- print info about a particular input\n"	*/ /* [04/09/21: how is this different than "input <name>" ??] */
		/* :-(	TAB "inputmap -- print the active inputmap\n"			*/
			TAB "inputdevices -- print info about all inputs devices\n"
			TAB "inputdevice[<num>] -- print info about the Nth device\n"
			TAB "windows -- print a summary of all the windows\n"
			TAB "window[<num>] -- print info about the Nth window\n"
			TAB "eyelists -- print info about all eyelists\n"
			TAB "eyelist[<num>] -- print info about the Nth eyelist\n"
			TAB "eyes -- print info about the eyes of the system\n"
			TAB "eye[<num>] -- print info about the Nth eye\n"
			TAB "users -- print info about all the active users\n"
			TAB "user[<num>] -- print info about all the Nth user\n"
		/* :-(	TAB "props -- print info about all the system's props\n"		*/

			/* These print options specify any objects in the configuration */
			/*   whether used in the existing system or not.                */
			TAB "object|o <type> list -- list all <type> config objects\n"
			TAB "object|o <type> <name> -- print info about a config object\n"
			TAB "object|o address <pointer> -- print info about a config object\n"
		"";
static	char		*whitespace = " \t\r\b";
static	char		buffer[4096];	/* a place for putting strings -- currently just for "ui" query */
	vrConfigInfo	*config = context->config;
	int		count;				/* for looping through all objects of the requested type */
	int		ival;				/* integer value -- generally for array index */

	vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Query request received for '%s' [fp = %#p]\n", query, file);


	/******************************************************/
	/*** parse query here -- very simple-minded parsing ***/
	/******************************************************/

	/********/
	/* help */
	if (!strncmp(query, "help", 4) || !strncmp(query, "?", 1)) {
		vrFprintf(file, help_msg);
	} else

	/*********/
	/* shmem */
	if (!strncmp(query, "shmem", 5)) {
		vrFprintShmemInfo(file, NULL, verbose /*s/b 'style', but only verbose implemented */);
	} else

	/***********/
	/* context */
	if (!strncmp(query, "context", 7)) {
		vrFprintContext(file, context, style);
	} else

	/**********/
	/* config */
	if (!strncmp(query, "config", 6)) {
		vrFprintConfig(file, config, style);
	} else

	/*************/
	/* inputinfo */
	if (!strncmp(query, "inputinfo", 9)) {
		vrFprintInput(file, context->input, style);
	} else

	/**********/
	/* system */
	if (!strncmp(query, "system", 6)) {
		vrFprintSystemInfo(file, config->system, style);
	} else

	/************/
	/* settings */
	if (!strncmp(query, "settings", 8)) {
		vrFprintSystemSettings(file, &(config->system->settings), style);
	} else

	/************/
	/* defaults */
	if (!strncmp(query, "defaults", 8)) {
		vrFprintSystemSettings(file, &(config->defaults), style);
	} else

	/********/
	/* near */
	if (!strncmp(query, "near", 4)) {
		if (config->system->settings.near_clip < 0.0)
			vrFprintf(file, "%.3f\n", config->defaults.near_clip);
		else	vrFprintf(file, "%.3f\n", config->system->settings.near_clip);
	} else

	/*******/
	/* far */
	if (!strncmp(query, "far", 4)) {
		if (config->system->settings.far_clip < 0.0)
			vrFprintf(file, "%.3f\n", config->defaults.far_clip);
		else	vrFprintf(file, "%.3f\n", config->system->settings.far_clip);
	} else

	/*********/
	/* procs */
	if (!strncmp(query, "procs", 5)) {
		for (count = 0; count < config->num_procs; count++) {
			vrFprintf(file, "proc[%d]: ", count);
			vrFprintProcessInfo(file, config->procs[count], one_line);
		}
	} else

	/*************/
	/* proc[<n>] */
	if (!strncmp(query, "proc[", 5)) {
		ival = vrAtoI(&query[5]);
		if (ival >= config->num_procs) {
			vrFprintf(file, "Invalid proc request (%d), only %d processes (0..%d).\n", ival, config->num_procs, config->num_procs-1);
		} else {
			if (style == machine) {
				vrFprintf(file, "%d:", ival);
			}
			vrFprintProcessInfo(file, config->procs[ival], style);
		}
	} else

	/***********************/
	/* ui (user interface) */
	if (!strncmp(query, "ui", 2)) {
		vrSprintInputUI(buffer, context->input, sizeof(buffer), style);
		vrFprintf(file, "=================================================================\n");
		vrFprintf(file, buffer);
		if (style == machine) {
			/* For machine parsers (which are probably line-based), we */
			/*   need to send a clear signal that the message is over. */
			vrFprintf(file, "\n==uiend==\n");
		}
	} else

	/**********/
	/* inputs */
	if (!strncmp(query, "inputs", 6)) {
		vrFprintInput(file, context->input, style);
	} else

	/****************/
	/* input <name> */
	if (!strncmp(query, "input ", 6)) {
		vrGenericInput *input;
		input = vrInputGetFromMapname(context, &query[6]);
		if (input != NULL)
			vrFprintInputObject(file, input->my_object, style); /* NOTE: printing the object also prints the value */
		else	vrFprintf(file, "unknown input: '%s'\n", &query[6]);
	} else

	/****************/
	/* inputdevices */
	if (!strncmp(query, "inputdevices", 12)) {
		for (count = 0; count < context->input->num_input_devices; count++) {
			vrFprintf(file, "inputdevice[%d]: ", count);
			vrFprintInputDevice(file, context->input->input_devices[count], one_line);
		}
	} else

	/********************/
	/* inputdevice[<n>] */
	if (!strncmp(query, "inputdevice[", 12)) {
		ival = vrAtoI(&query[12]);
		if (ival >= context->input->num_input_devices) {
			vrFprintf(file, "Invalid input-device request (%d), only %d input devices (0..%d).\n", ival, context->input->num_input_devices, context->input->num_input_devices-1);
		} else {
			vrFprintInputDevice(file, context->input->input_devices[ival], style);
		}
	} else

	/***********/
	/* windows */
	if (!strncmp(query, "windows", 7)) {
		if (style == machine) {
			/* For the "machine" style, print a one line list of the windows */
			vrFprintf(file, "%d", config->num_windows);
			for (count = 0; count < config->num_windows; count++) {
				vrFprintf(file, ":%s", config->windows[count]->name);
			}
			vrFprintf(file, "\n");
		} else {
			for (count = 0; count < config->num_windows; count++) {
				vrFprintf(file, "window[%d]: ", count);
				vrFprintWindowInfo(file, config->windows[count], one_line);
			}
		}
	} else

	/***************/
	/* window[<n>] */
	if (!strncmp(query, "window[", 7)) {
		ival = vrAtoI(&query[7]);
		if (ival >= config->num_windows) {
			vrFprintf(file, "Invalid window request (%d), only %d windows (0..%d).\n", ival, config->num_windows, config->num_windows-1);
		} else {
			vrFprintWindowInfo(file, config->windows[ival], style);
		}
	} else

	/*********/
	/* users */
	if (!strncmp(query, "users", 5)) {
		if (style == machine) {
			/* For the "machine" style, print a one line list of the users */
			vrFprintf(file, "%d", config->num_users);
			for (count = 0; count < config->num_users; count++) {
				vrFprintf(file, ":%s", config->users[count]->name);
			}
			vrFprintf(file, "\n");
		} else {
			for (count = 0; count < config->num_users; count++) {
				vrFprintf(file, "user[%d]: ", count);
				vrFprintUserInfo(file, config->users[count], one_line);
			}
		}
	} else

	/*************/
	/* user[<n>] */
	if (!strncmp(query, "user[", 5)) {
		ival = vrAtoI(&query[5]);
		if (ival >= config->num_users) {
			vrFprintf(file, "Invalid user request (%d), only %d users (0..%d).\n", ival, config->num_users, config->num_users-1);
		} else {
			if (style == machine) {
				vrFprintf(file, "%d:", ival);
			}
			vrFprintUserInfo(file, config->users[ival], style);
		}
	} else

	/*********/
	/* eyelists */
	if (!strncmp(query, "eyelists", 8)) {
		for (count = 0; count < config->num_eyelists; count++) {
			vrFprintf(file, "eyelist[%d]: ", count);
			vrFprintEyelistInfo(file, config->eyelists[count], one_line);
		}
	} else

	/*************/
	/* eyelist[<n>] */
	if (!strncmp(query, "eyelist[", 8)) {
		ival = vrAtoI(&query[8]);
vrFprintf(file, "YO num_eyelists = %d, ival = %d\n", config->num_eyelists, ival);
		if (ival >= config->num_eyelists) {
			vrFprintf(file, "Invalid eyelist request (%d), only %d eyelists (0..%d).\n", ival, config->num_eyelists, config->num_eyelists-1);
		} else {
			vrFprintEyelistInfo(file, config->eyelists[ival], style);
		}
	} else

	/*********/
	/* eyes */
	if (!strncmp(query, "eyes", 4)) {
		for (count = 0; count < config->num_eyes; count++) {
			vrFprintf(file, "eye[%d]: ", count);
			vrFprintEyeInfo(file, config->eyes[count], one_line);
		}
	} else

	/*************/
	/* eye[<n>] */
	if (!strncmp(query, "eye[", 4)) {
		ival = vrAtoI(&query[4]);
		if (ival >= config->num_eyes) {
			vrFprintf(file, "Invalid eye request (%d), only %d eyes (0..%d).\n", ival, config->num_eyes, config->num_eyes-1);
		} else {
			vrFprintEyeInfo(file, config->eyes[ival], style);
		}
	} else

	/*************************/
	/* configuration objects */
	if ((!strncmp(query, "object ", 7)) || (!strncmp(query, "o ", 2))) {
		char	argv[5][128];		/* for parsing argument list */
		char	*str_begin;
		int	str_len;

		if (query[1] == ' ')
			str_begin = &query[1];
		else	str_begin = &query[6];

		/* first argument */
		str_begin += strspn(str_begin, whitespace);	/* skip whitespace */
		str_len = strcspn(str_begin, whitespace);	/* count non-whitespace */
		strncpy(argv[1], str_begin, str_len);
		argv[1][str_len] = '\0';
		str_begin += str_len;

		/* second argument */
		str_begin += strspn(str_begin, whitespace);	/* skip whitespace */
		str_len = strcspn(str_begin, whitespace);	/* count non-whitespace */
		strncpy(argv[2], str_begin, str_len);
		argv[2][str_len] = '\0';
		str_begin += str_len;

		vrFprintObjectTypeInfo(file, context, argv[1], argv[2], style);
	} else

	/************/
	/* default: */ {
		vrFprintf(file, "Unknown query: '%s'.\n", query);
	}
}


/***********************************************************/
/* vrInfoSet(): parse a request and attempt to modify the  */
/*   system as requested.  Errors or other messages *may*  */
/*   be sent to the given FILE stream.  See the help_msg   */
/*   string for a list of possible requests.  Passing      */
/*   "help" prints this list.                              */
/* NOTE: this function basically takes the input string    */
/*   with the "set" or "s" (or whatever) stripped off.     */
/***********************************************************/
/* NOTE: this function (and vrInfoGet(), which precedes) could easily*/
/*   be moved to another source file, as they don't have much to do  */
/*   with telnetting, other than this is the only process that makes */
/*   use of it at the moment.                                        */
void vrInfoSet(vrContextInfo *context, FILE *file, char *request, vrPrintStyle style)
{
static	char	*whitespace = " \t\r\b\n";
static	char	*help_msg = "Available settables:\n"
			TAB "help|? -- print this message (a list of possible setting requests).\n"
			TAB "input <name> <value> -- set the value of an input.\n"
			TAB TAB "input <6-sensor> { id | loc <x> <y> <z> | move <x> <y> <z> <az> <el> <ro> }.\n"
			TAB TAB "input <6-sensor> { nudger2eA | nra } { id | <x> <y> <z> } -- Nudge the r2e Euler angles.\n"
			TAB TAB "input <6-sensor> { nudger2eT | nrt } { id | <x> <y> <z> } -- Nudge the r2e Translation.\n"
			TAB "inputdevice[<num>] <value> -- set the value of an input.\n"
			TAB TAB "inputdevice[<num>] { nudget2rwA | nta } { id | <x> <y> <z> } -- Nudge the t2rw Euler angles.\n"
			TAB TAB "inputdevice[<num>] { nudget2rwT | ntt } { id | <x> <y> <z> } -- Nudge the t2rw Translation.\n"
			TAB "pause {0,1} - pause the entire system.\n"
			TAB "debuglevel|dl <value> -- set the global debug level.\n"
			TAB "debugthistoo|dtt <value> -- set the global debug-this-too.\n"
			TAB "debuglevel|dl[<num>] <value> -- set process <num>'s debug level.\n"
			TAB "debugthistoo|dtt[<num>] <value> -- set process <num>'s debug-this-too.\n"
			TAB "near <value> -- set the global value of the near clipping plane.\n"
			TAB "far <value> -- set the global value of the far clipping plane.\n"
			TAB "window[<num>] ui {0,1} -- turn off/on a window's user-interface display.\n"
			TAB "window[<num>] ui_loc <v1>,<v2> -- relocate a window's user-interface display.\n"
			TAB "window[<num>] ui_color <v1>,<v2>,<v3> -- set color of a window's user-interface display.\n"
			TAB "window[<num>] fps {0,1} -- turn off/on a window's fps display.\n"
			TAB "window[<num>] fps_loc <v1>,<v2> -- relocate a window's fps display.\n"
			TAB "window[<num>] fps_color <v1>,<v2>,<v3> -- set color of a window's fps display.\n"
			TAB "window[<num>] frame {0,1} -- turn off/on a window's frame display in all windows.\n"
			TAB "window[<num>] frm <value> -- set the front rendering mode\n"
			TAB "window[<num>] brm <value> -- set the back rendering mode\n"
			TAB "window[<num>] stats {0,1} -- turn off/on a window's stats display.\n"
			TAB "window[<num>] world {0,1} -- turn off/on a window's virtual world display.\n"
			TAB "window[<num>] nudgex <value> -- shift the window along the X-axis\n"
			TAB "window[<num>] nudgey <value> -- shift the window along the Y-axis\n"
			TAB "window[<num>] nudgez <value> -- shift the window along the Z-axis\n"
			TAB "window[<num>] sim <move command> -- move the view of a simulator window.\n"
			TAB "window[<num>] sim <mask command> -- set the simulator rendering mask.\n"
			TAB "object window[<name>] <any window command> -- affect a command on any window in the configuration.\n"
			TAB "user[<num>] iod <v1> -- set the iod of a user.\n"
			TAB "user[<num>] color <v1>,<v2>,<v3> -- set the color of a user.\n"
	// :-(		TAB "user[<num>] nudgeT <v1>,<v2>,<v3> -- translate the user's travel.\n"
	// :-(		TAB "user[<num>] nudgeA <v1>,<v2>,<v3> -- rotate the user's travel.\n"
			TAB "proc[<num>] end <v1> -- set the req_proc_term flag of a process.\n"
			TAB "proc[<num>] term <v1> -- set the proc_term flag of a process.\n"
			TAB "proc[<num>] usec <v1> -- set the usec_min value of a process.\n"
			TAB "proc[<num>] printcolor <v1> -- set the print_color flag of a process.\n"
			TAB "proc[<num>] stats_calc -- set the flag of whether to calculate stats\n"
			TAB "proc[<num>] stats_show -- set the flag of whether to show these stats\n"
			TAB "proc[<num>] stats_mask -- set the mask of which statistics to show\n"
			TAB "proc[<num>] stats_xloc -- set the x location of where to put stats\n"
			TAB "proc[<num>] stats_yloc -- set the y location of where to put stats\n"
			TAB "proc[<num>] stats_width -- set the width of the stats display\n"
			TAB "proc[<num>] stats_top -- set the top timeline of the stats display\n"
			TAB "proc[<num>] stats_interval -- set the horizontal time intervals of the stats display\n"
			TAB "proc[<num>] stats_scale -- set the vertical scale of the stats display (aka time_scale)\n"
			TAB "proc[<num>] stats_color -- set the background color of the stats display\n"
			TAB "proc[<num>] stats_opac -- set the background opacity of the stats display\n"
		"";
	char		msg_buffer[1024];		/* buffer for sending messages through socket */
	vrConfigInfo	*config = context->config;
	char		*parse;
	int		obj_num;			/* number of the object to modify (ie. proc, user, window) */
	int		value1i, value2i, value3i;	/* some integer values for arguments */
	float		value1f, value2f, value3f;	/* some float values for arguments */

	vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Set request received for '%s' [fp = %#p]\n", request, file);


	/********************************************************/
	/*** parse setting here -- very simple-minded parsing ***/
	/********************************************************/

	/********/
	/* help */
	if (!strncmp(request, "help", 4) || !strncmp(request, "?", 1)) {
		vrFprintf(file, help_msg);
	} else

	/*************************/
	/* input <obj> <setting> */
	if (!strncmp(request, "input ", 6)) {
		vrGenericInput	*input;
		char		*result;

		input = vrInputGetFromMapname(context, &request[6]);
		if (input == NULL) {
			vrFprintf(file, "Unknown input: '%s'\n", &request[6]);
		} else {
			result = vrAssignGenericInput(input, strchr(&request[6], ' '));
			if (result != NULL) {
				vrFprintf(file, "Error: %s\n", result);
			}
			if (style == verbose)
				vrFprintInputValue(file, input, brief);
		}
	} else

	/******************************/
	/* inputdevice[<n>] <setting> */
	if (!strncmp(request, "inputdevice[", 12)) {
		vrVector	settings = { 0.0, 0.0, 0.0 };	/* translation values */
		char		*id_check;			/* for parsing for "id" */

		obj_num = vrAtoI(&request[12]);
		if (obj_num >= config->num_input_devices) {
			vrFprintf(file, "Invalid input-device request (%d), only %d input devices (0..%d).\n", obj_num, config->num_input_devices, config->num_input_devices-1);
		} else if (strchr(&request[12], ']') == NULL) {
			vrFprintf(file, "Invalid input-device request -- no closing ']'.\n");
		} else {
			parse = strchr(&request[12], ']')+1;
			parse += strspn(parse, whitespace);	/* skip white */

			/*******************************************************************/
			/*** <input device>  "nra|nudget2rwa" { id | x-rot y-rot z-rot } ***/
			if (!strncasecmp(parse, "nudget2rwa", 10) || !strncasecmp(parse, "nta", 3)) {
				/* Nudge the t2rw matrix rotation */
				parse = strchr(&parse[2], ' ');				/* skip to a space */

				/* check whether to reset to identity translation */
				/* NOTE: I need to leave the whitespace for the value checks below, */
				/*   so I'm using a second, temporary, variable for the "id" check. */
				/* [I'm not 100% sure that's correct, so maybe test w/o the next line */
				id_check = parse + strspn(parse, whitespace);		/* skip white for "id" check */
				if (!strncmp(id_check, "id ", 2)) {
					/* set t2rw_xform to the identity */
					vrMatrixSetIdentityRotationOnly(config->input_devices[obj_num]->t2rw_xform);
				} else {
					/* expect up to three translation values */
					if (parse != NULL && strlen(parse) > 0) {	/* done if no more chars */
						parse += strspn(parse, whitespace);	/* skip white */
						settings.v[VR_X] = atof(parse);
						parse = strchr(parse, ' ');		/* skip to next space */
					}
					if (parse != NULL && strlen(parse) > 0) {	/* done if no more chars */
						parse += strspn(parse, whitespace);	/* skip white */
						settings.v[VR_Y] = atof(parse);
						parse = strchr(parse, ' ');		/* skip to next space */
					}
					if (parse != NULL && strlen(parse) > 0) {	/* done if no more chars */
						parse += strspn(parse, whitespace);	/* skip white */
						settings.v[VR_Z] = atof(parse);
						parse = strchr(parse, ' ');		/* skip to next space */
					}

					/* Now do the rotation in-place */
					vrMatrixPostMult(config->input_devices[obj_num]->t2rw_xform, vrMatrixSetRotationId(vrMatrixCreate(), VR_X, settings.v[VR_X]));  /* TODO: a small memory leak here -- the created matrices aren't freed */
					vrMatrixPostMult(config->input_devices[obj_num]->t2rw_xform, vrMatrixSetRotationId(vrMatrixCreate(), VR_Y, settings.v[VR_Y]));
					vrMatrixPostMult(config->input_devices[obj_num]->t2rw_xform, vrMatrixSetRotationId(vrMatrixCreate(), VR_Z, settings.v[VR_Z]));
				}
				if (style == verbose)
					vrFprintInputDevice(file, config->input_devices[obj_num], brief);
			} else

			if (!strncasecmp(parse, "nudget2rwt", 10) || !strncasecmp(parse, "ntt", 3)) {
				/* Nudge the t2rw matrix translation */
				parse = strchr(&parse[2], ' ');				/* skip to a space */

				/* check whether to reset to identity translation */
				/* NOTE: I need to leave the whitespace for the value checks below, */
				/*   so I'm using a second, temporary, variable for the "id" check. */
				/* [I'm not 100% sure that's correct, so maybe test w/o the next line */
				id_check = parse + strspn(parse, whitespace);		/* skip white for "id" check */
				if (!strncmp(id_check, "id ", 2)) {
					/* set t2rw_xform to the identity */
					vrMatrixSetTranslationOnlyAd(config->input_devices[obj_num]->t2rw_xform, settings.v);
				} else {
					/* expect up to three translation values */
					if (parse != NULL && strlen(parse) > 0) {	/* done if no more chars */
						parse += strspn(parse, whitespace);	/* skip white */
						settings.v[VR_X] = atof(parse);
						parse = strchr(parse, ' ');		/* skip to next space */
					}
					if (parse != NULL && strlen(parse) > 0) {	/* done if no more chars */
						parse += strspn(parse, whitespace);	/* skip white */
						settings.v[VR_Y] = atof(parse);
						parse = strchr(parse, ' ');		/* skip to next space */
					}
					if (parse != NULL && strlen(parse) > 0) {	/* done if no more chars */
						parse += strspn(parse, whitespace);	/* skip white */
						settings.v[VR_Z] = atof(parse);
						parse = strchr(parse, ' ');		/* skip to next space */
					}

					vrMatrixPostTranslateAd(config->input_devices[obj_num]->t2rw_xform, settings.v);
				}
				if (style == verbose)
					vrFprintInputDevice(file, config->input_devices[obj_num], brief);
			} else {
				vrFprintf(file, "Invalid input-device request -- unknown setting.\n");
			}
		}
	} else

	/***********************/
	/* pause <value> */
	if (!strncmp(request, "pause ", 6)) {
		context->paused = vrAtoI(&request[6]);
		if (style == verbose)
			vrFprintf(file, "Set System Pause value to %d\n", context->paused);
	} else

	/***********************/
	/* debug_level[[<proc_num>]] <value> */
	if (!strncmp(request, "debuglevel ", 11)) {
		config->defaults.debug_level = vrAtoI(&request[11]);
		if (style == verbose)
			vrFprintf(file, "Set Global DebugLevel to %d\n", config->defaults.debug_level);
	} else
	if (!strncmp(request, "dl ", 3)) {
		config->defaults.debug_level = vrAtoI(&request[3]);
		if (style == verbose)
			vrFprintf(file, "Set Global DebugLevel to %d\n", config->defaults.debug_level);
	} else
	if (!strncmp(request, "debuglevel[", 11)) {
		obj_num = vrAtoI(&request[11]);
		if (obj_num >= config->num_procs) {
			vrFprintf(file, "Invalid proc request (%d), only %d processes (0..%d).\n", obj_num, config->num_procs, config->num_procs-1);
		} else if (strchr(&request[11], ']') == NULL) {
			vrFprintf(file, "Invalid proc request -- no closing ']'.\n");
		} else {
			config->procs[obj_num]->settings.debug_level = vrAtoI(strchr(&request[11], ']')+1);
			if (style == verbose)
				vrFprintf(file, "Set Process[%d] DebugLevel to %d\n", obj_num, config->procs[obj_num]->settings.debug_level);
		}
	} else
	if (!strncmp(request, "dl[", 3)) {
		obj_num = vrAtoI(&request[3]);
		if (obj_num >= config->num_procs) {
			vrFprintf(file, "Invalid proc request (%d), only %d processes (0..%d).\n", obj_num, config->num_procs, config->num_procs-1);
		} else if (strchr(&request[3], ']') == NULL) {
			vrFprintf(file, "Invalid proc request -- no closing ']'.\n");
		} else {
			config->procs[obj_num]->settings.debug_level = vrAtoI(strchr(&request[3], ']')+1);
			if (style == verbose) {
				vrFprintf(file, "Set Process[%d] DebugLevel to %d\n", obj_num, config->procs[obj_num]->settings.debug_level);
			}
		}
	} else

	/***********************/
	/* debug_exact[[<proc_num>]] <value> */
	if (!strncmp(request, "debugthistoo ", 13)) {
		config->defaults.debug_exact = vrAtoI(&request[13]);
		if (style == verbose)
			vrFprintf(file, "Set Global DebugThisToo to %d\n", config->defaults.debug_exact);
	} else
	if (!strncmp(request, "dtt ", 4)) {
		config->defaults.debug_exact = vrAtoI(&request[4]);
		if (style == verbose)
			vrFprintf(file, "Set Global DebugThisToo to %d\n", config->defaults.debug_exact);
	} else
	if (!strncmp(request, "debugthistoo[", 13)) {
		obj_num = vrAtoI(&request[13]);
		if (obj_num >= config->num_procs) {
			vrFprintf(file, "Invalid proc request (%d), only %d processes (0..%d).\n", obj_num, config->num_procs, config->num_procs-1);
		} else if (strchr(&request[13], ']') == NULL) {
			vrFprintf(file, "Invalid proc request -- no closing ']'.\n");
		} else {
			config->procs[obj_num]->settings.debug_exact = vrAtoI(strchr(&request[13], ']')+1);
			if (style == verbose)
				vrFprintf(file, "Set Process[%d] DebugThisToo to %d\n", obj_num, config->procs[obj_num]->settings.debug_exact);
		}
	} else
	if (!strncmp(request, "dtt[", 4)) {
		obj_num = vrAtoI(&request[4]);
		if (obj_num >= config->num_procs) {
			vrFprintf(file, "Invalid proc request (%d), only %d processes (0..%d).\n", obj_num, config->num_procs, config->num_procs-1);
		} else if (strchr(&request[4], ']') == NULL) {
			vrFprintf(file, "Invalid proc request -- no closing ']'.\n");
		} else {
			config->procs[obj_num]->settings.debug_exact = vrAtoI(strchr(&request[4], ']')+1);
			if (style == verbose)
				vrFprintf(file, "Set Process[%d] DebugThisToo to %d\n", obj_num, config->procs[obj_num]->settings.debug_exact);
		}
	} else

	/****************/
	/* near <value> */
	if (!strncmp(request, "near ", 5)) {
		value1f = atof(&request[5]);
		/* NOTE: "-1.0" represents the "default", and will allow inheritance */
		if ((value1f == -1.0) || (value1f > 0.0)) {
			vrSetNear(context, value1f);
			if (style == verbose)
				vrFprintf(file, "Set Global near clipping plane to %f\n", value1f);
		} else {
			vrFprintf(file, "Unable to set near clipping plane to non-positive value: %f\n", value1f);
		}
	} else

	/***************/
	/* far <value> */
	if (!strncmp(request, "far ", 4)) {
		value1f = atof(&request[4]);
		/* NOTE: "-1.0" represents the "default", and will allow inheritance */
		if ((value1f == -1.0) || (value1f > 0.0)) {
			vrSetFar(context, value1f);
			if (style == verbose)
				vrFprintf(file, "Set Global far clipping plane to %f\n", value1f);
		} else {
			vrFprintf(file, "Unable to set far clipping plane to non-positive value: %f\n", value1f);
		}
	} else

	/******************************************************/
	/* window[<n>] {ui,ui_loc,ui_color,fps,fps_loc,fps_color,nudge[xyz],stats,sim} <value(s)> */
	/* or object window[<name>] {ui,ui_loc,ui_color,fps,fps_loc,fps_color,nudge[xyz],stats,sim} <value(s)> */
	/* TODO: I'd like to also have the ability to give an object window number as another option. */
	if ((!strncmp(request, "window[", 7)) || (!strncmp(request, "object window[", 14))) {
		vrWindowInfo	*window = NULL;

		/* determine the window requested (from the active config or from the complete list */
		if (strchr(&request[7], ']') == NULL) {
			vrFprintf(file, "Invalid window request -- no closing ']'.\n");
		} else if (request[0] == 'w') {
			/* the standard request is for an enumerated active window */
			obj_num = vrAtoI(&request[7]);
			if (obj_num >= config->num_windows) {
				vrFprintf(file, "Invalid window request (%d), only %d windows (0..%d).\n", obj_num, config->num_windows, config->num_windows-1);
			} else {
				parse = strchr(&request[7], ']')+1;	/* begin parsing the command after the object string */
				window = config->windows[obj_num];
			}
		} else if (request[0] == 'o') {
			/* one may also request a named window from the entire configuration */
			parse = strchr(&request[14], ']')+1;	/* begin parsing the command after the object string */
			parse[-1] = '\0';			/* end the object portion of the string at the closing square-bracket */
			window = (vrWindowInfo *)vrObjectSearch(context, VROBJECT_WINDOW, &request[14]);
			parse[-1] = ']';			/* Restore the string for future skipping of command */
			/* TODO: if window == NULL here, check whether a number was given and use that */
			if (window == NULL) {
				vrFprintf(file, "Invalid window request: no window named '%s'.\n", &request[14]);
			}
		}

		/* Now that we have a pointer to a window parse the rest of the request */
		if (window != NULL) {
			parse += strspn(parse, whitespace);	/* skip white */

			if (!strncmp(parse, "ui ", 3)) {
				parse += 3;					/* skip "ui " */
				value1i = vrAtoI(parse);
				window->ui_show = value1i;
				if (style == verbose)
					vrFprintf(file, "set window[%d] ui to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "ui_loc ", 7)) {
				parse += 7;					/* skip "ui_loc " */
				value1f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value2f = atof(parse);
				window->ui_loc[0] = value1f;
				window->ui_loc[1] = value2f;
				if (style == verbose)
					vrFprintf(file, "set window[%d] ui_loc to %.2f,%.2f.\n", obj_num, value1f, value2f);
			} else if (!strncmp(parse, "ui_color ", 9)) {
				parse += 9;					/* skip "ui_color " */
				value1f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value2f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value3f = atof(parse);
				window->ui_color[0] = value1f;
				window->ui_color[1] = value2f;
				window->ui_color[2] = value3f;
				if (style == verbose)
					vrFprintf(file, "set window[%d] ui_color to %.3f,%.3f,%.3f.\n", obj_num, value1f, value2f, value3f);
			} else if (!strncmp(parse, "fps ", 4)) {
				parse += 4;					/* skip "fps " */
				value1i = vrAtoI(parse);
				window->fps_show = value1i;
				if (style == verbose)
					vrFprintf(file, "set window[%d] fps to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "fps_loc ", 8)) {
				parse += 8;
				value1f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value2f = atof(parse);
				window->fps_loc[0] = value1f;
				window->fps_loc[1] = value2f;
				if (style == verbose)
					vrFprintf(file, "set window[%d] fps_loc to %.2f,%.2f.\n", obj_num, value1f, value2f);
			} else if (!strncmp(parse, "fps_color ", 10)) {
				parse += 10;
				value1f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value2f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value3f = atof(parse);
				window->fps_color[0] = value1f;
				window->fps_color[1] = value2f;
				window->fps_color[2] = value3f;
				if (style == verbose)
					vrFprintf(file, "set window[%d] fps_color to %.3f,%.3f,%.3f.\n", obj_num, value1f, value2f, value3f);
			} else if (!strncmp(parse, "frame ", 6)) {
				parse += 6;				/* skip "frame " */
				value1i = vrAtoI(parse);
				window->show_in_simulator = value1i;
				if (style == verbose)
					vrFprintf(file, "set window[%s] frame to %d.\n", window->name, value1i);
			} else if (!strncmp(parse, "stats ", 6)) {
				parse += 6;				/* skip "stats " */
				value1i = vrAtoI(parse);
				window->stats_show = value1i;
				if (style == verbose)
					vrFprintf(file, "set window[%d] stats to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "world ", 6)) {
				parse += 6;				/* skip "world " */
				value1i = vrAtoI(parse);
				window->world_show = value1i;
				if (style == verbose)
					vrFprintf(file, "set window[%d] world to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "sim ", 4)) {
				parse += 4;				/* skip "sim " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				vrSimulatorMoveString(window, parse);
			} else if (!strncmp(parse, "frm ", 4)) {
				parse += 4;				/* skip "frm " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				window->frontrendermode = value1f;
			} else if (!strncmp(parse, "brm ", 4)) {
				parse += 4;				/* skip "brm " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				window->backrendermode = value1f;
			} else if (!strncmp(parse, "nudgex ", 7)) {
				parse += 7;				/* skip "nudgex " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPostTranslate3d(window->rw2w_xform, value1f, 0.0, 0.0);
				window->coords_ll[VR_X] += value1f;
				window->coords_lr[VR_X] += value1f;
				window->coords_ul[VR_X] += value1f;
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
				//vrSocketSendMsg(file, msg_buffer);
#endif
			} else if (!strncmp(parse, "nudgey ", 7)) {
				parse += 7;				/* skip "nudgey " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPostTranslate3d(window->rw2w_xform, 0.0, value1f, 0.0);
				window->coords_ll[VR_Y] += value1f;
				window->coords_lr[VR_Y] += value1f;
				window->coords_ul[VR_Y] += value1f;
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
			} else if (!strncmp(parse, "nudgez ", 7)) {
				parse += 7;				/* skip "nudgez " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPostTranslate3d(window->rw2w_xform, 0.0, 0.0, value1f);
				window->coords_ll[VR_Z] += value1f;
				window->coords_lr[VR_Z] += value1f;
				window->coords_ul[VR_Z] += value1f;
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
			} else if (!strncmp(parse, "nudgezp ", 7)) {
				parse += 7;				/* skip "nudgez " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPreTranslate3d(window->rw2w_xform, 0.0, 0.0, value1f);
				window->coords_ll[VR_Z] += value1f;
				window->coords_lr[VR_Z] += value1f;
				window->coords_ul[VR_Z] += value1f;
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
			} else if (!strncmp(parse, "nudgerx ", 8)) {
				parse += 8;				/* skip "nudgerx " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPostRotateAboutOriginId(window->rw2w_xform, VR_X, value1f);
				//vrMatrixPostRotateId(window->rw2w_xform, VR_X, value1f);
				/* TODO: the coords values need corrected for the showFrame visualization to be correct */
				//window->coords_ul[VR_Y] += value1f;
{
	vrMatrix	tmpmat;
	vrMatrixSetIdentity(&tmpmat);
	vrMatrixPostTranslate3d(&tmpmat, 0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), 0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), 0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrMatrixPostRotateId(&tmpmat, VR_X, -value1f);
	vrMatrixPostTranslate3d(&tmpmat, -0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), -0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), -0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrPointTransformByMatrix((vrPoint*)(window->coords_ll), (vrPoint*)(window->coords_ll), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_ul), (vrPoint*)(window->coords_ul), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_lr), (vrPoint*)(window->coords_lr), &tmpmat);
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
}
				
			} else if (!strncmp(parse, "nudgery ", 8)) {
				parse += 8;				/* skip "nudgery " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPostRotateAboutOriginId(window->rw2w_xform, VR_Y, value1f);
				//vrMatrixPostRotateId(window->rw2w_xform, VR_Y, value1f);
				/* TODO: the coords values need corrected for the showFrame visualization to be correct */
				//window->coords_ul[VR_Y] += value1f;
{
	vrMatrix	tmpmat;
	vrMatrixSetIdentity(&tmpmat);
	vrMatrixPostTranslate3d(&tmpmat, 0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), 0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), 0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrMatrixPostRotateId(&tmpmat, VR_Y, -value1f);
	vrMatrixPostTranslate3d(&tmpmat, -0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), -0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), -0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrPointTransformByMatrix((vrPoint*)(window->coords_ll), (vrPoint*)(window->coords_ll), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_ul), (vrPoint*)(window->coords_ul), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_lr), (vrPoint*)(window->coords_lr), &tmpmat);
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
}
			} else if (!strncmp(parse, "nudgeryp ", 8)) {
				parse += 8;				/* skip "nudgery " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPreRotateAboutOriginId(window->rw2w_xform, VR_Y, value1f);
				//vrMatrixPreRotateId(window->rw2w_xform, VR_Y, value1f);
				/* TODO: the coords values need corrected for the showFrame visualization to be correct */
				//window->coords_ul[VR_Y] += value1f;
{
	vrMatrix	tmpmat;
	vrMatrixSetIdentity(&tmpmat);
	vrMatrixPostTranslate3d(&tmpmat, 0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), 0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), 0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrMatrixPostRotateId(&tmpmat, VR_Y, -value1f);
	vrMatrixPostTranslate3d(&tmpmat, -0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), -0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), -0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrPointTransformByMatrix((vrPoint*)(window->coords_ll), (vrPoint*)(window->coords_ll), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_ul), (vrPoint*)(window->coords_ul), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_lr), (vrPoint*)(window->coords_lr), &tmpmat);
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
}
			} else if (!strncmp(parse, "nudgerz ", 8)) {
				parse += 8;				/* skip "nudgerz " */
				parse += strspn(parse, whitespace);	/* skip extra whitespace */
				value1f = atof(parse);
				vrMatrixPostRotateAboutOriginId(window->rw2w_xform, VR_Z, value1f);
				//vrMatrixPostRotateId(window->rw2w_xform, VR_Z, value1f);
				/* TODO: the coords values need corrected for the showFrame visualization to be correct */
				//window->coords_ul[VR_Y] += value1f;
{
	vrMatrix	tmpmat;
	vrMatrixSetIdentity(&tmpmat);
	vrMatrixPostTranslate3d(&tmpmat, 0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), 0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), 0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrMatrixPostRotateId(&tmpmat, VR_Z, -value1f);
	vrMatrixPostTranslate3d(&tmpmat, -0.5*(window->coords_lr[VR_X]+window->coords_ul[VR_X]), -0.5*(window->coords_lr[VR_Y]+window->coords_ul[VR_Y]), -0.5*(window->coords_lr[VR_Z]+window->coords_ul[VR_Z]));
	vrPointTransformByMatrix((vrPoint*)(window->coords_ll), (vrPoint*)(window->coords_ll), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_ul), (vrPoint*)(window->coords_ul), &tmpmat);
	vrPointTransformByMatrix((vrPoint*)(window->coords_lr), (vrPoint*)(window->coords_lr), &tmpmat);
	vrCalcPerspIntermediaries(window);
#if 1 /* output new coords (for debugging, probably delete me) */
				vrPrintf("Coords: (%.2f %.2f %.2f) (%.2f %.2f %.2f) (%.2f %.2f %.2f)\n",
					window->coords_ll[VR_X], window->coords_ll[VR_Y], window->coords_ll[VR_Z],
					window->coords_lr[VR_X], window->coords_lr[VR_Y], window->coords_lr[VR_Z],
					window->coords_ul[VR_X], window->coords_ul[VR_Y], window->coords_ul[VR_Z]);
#endif
}
			} else {
				vrFprintf(file, "Unknown window request: '%s'.\n", request);
			}
		}
	} else

	/**************************************************/
	/* user[<n>] {iod,color} <value(s)> */
	if (!strncmp(request, "user[", 5)) {
		obj_num = vrAtoI(&request[5]);
		if (obj_num >= config->num_users) {
			vrFprintf(file, "Invalid user request (%d), only %d users (0..%d).\n", obj_num, config->num_users, config->num_users-1);
		} else if (strchr(&request[5], ']') == NULL) {
			vrFprintf(file, "Invalid user request -- no closing ']'.\n");
		} else {
			parse = strchr(&request[5], ']')+1;
			parse += strspn(parse, whitespace);	/* skip white */

			if (!strncmp(parse, "iod ", 4)) {
				parse += 4;
				value1f = atof(parse);
				config->users[obj_num]->iod = value1f;
				if (style == verbose)
					vrFprintf(file, "set user[%d] iod to %.3f.\n", obj_num, value1f);
			} else if (!strncmp(parse, "color ", 6)) {
				parse += 6;
				value1i = vrAtoI(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value2i = vrAtoI(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value3i = vrAtoI(parse);
				config->users[obj_num]->color[0] = value1i;
				config->users[obj_num]->color[1] = value2i;
				config->users[obj_num]->color[2] = value3i;
				if (style == verbose)
					vrFprintf(file, "set user[%d] color to %d,%d,%d.\n", obj_num, value1i, value2i, value3i);
			} else {
				vrFprintf(file, "Unknown user request: '%s'.\n", request);
			}
		}
	} else

	/**************************************************/
	/* proc[<n>] {end,done,printcolor,usec,stats_calc,stats_show,stats_mask,stats_xloc,stats_yloc,stats_width,stats_top,stats_interval,stats_scale,stats_color,stats_opac} <value(s)> */
	if (!strncmp(request, "proc[", 5)) {
		obj_num = vrAtoI(&request[5]);
		if (obj_num >= config->num_procs) {
			vrFprintf(file, "Invalid process request (%d), only %d processes (0..%d).\n", obj_num, config->num_procs, config->num_procs-1);
		} else if (strchr(&request[5], ']') == NULL) {
			vrFprintf(file, "Invalid process request -- no closing ']'.\n");
		} else {
			parse = strchr(&request[5], ']')+1;
			parse += strspn(parse, whitespace);	/* skip white */

			if (!strncmp(parse, "end ", 4)) {
				parse += 4;
				value1i = vrAtoI(parse);
				config->procs[obj_num]->req_proc_term = value1i;
				if (style == verbose)
					vrFprintf(file, "set proc[%d] req_proc_term to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "term ", 5)) {
				/* 11/19/2020: Hmmmm, do we really want to be able to set "proc_term" vs. "req_proc_term"? */
				parse += 5;
				value1i = vrAtoI(parse);
				config->procs[obj_num]->proc_term = value1i;
				if (style == verbose)
					vrFprintf(file, "set proc[%d] proc_term to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "printcolor ", 11)) {
				parse += 11;
				value1i = vrAtoI(parse);
				config->procs[obj_num]->print_color = value1i;
				vrDbgPrintfN(AALWAYS_DBGLVL, "New print color for proc[%d] is %d\n", obj_num, value1i);
				if (style == verbose)
					vrFprintf(file, "set proc[%d] print_color to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "usec ", 5)) {
				parse += 5;
				value1i = vrAtoI(parse);
				config->procs[obj_num]->usec_min = value1i;
				vrDbgPrintfN(AALWAYS_DBGLVL, "New minimal usec delay for proc[%d] is %d\n", obj_num, value1i);
				if (style == verbose)
					vrFprintf(file, "set proc[%d] usec_min to %d.\n", obj_num, value1i);
			} else if (!strncmp(parse, "stats_calc ", 11)) {
				parse += 11;
				value1i = vrAtoI(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->calc_flag = value1i;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->calc_flag to %d.\n", obj_num, value1i);
				}
			} else if (!strncmp(parse, "stats_show ", 11)) {
				parse += 11;
				value1i = vrAtoI(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->show_flag = value1i;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->show_flag to %d.\n", obj_num, value1i);
				}
			} else if (!strncmp(parse, "stats_mask ", 11)) {
				parse += 11;
				value1i = vrAtoI(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->show_mask = value1i;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->show_mask to %d.\n", obj_num, value1i);
				}
			} else if (!strncmp(parse, "stats_xloc ", 11)) {
				parse += 11;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->xloc = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->xloc to %f.\n", obj_num, value1f);
				}
			} else if (!strncmp(parse, "stats_yloc ", 11)) {
				parse += 11;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->yloc = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->yloc to %f.\n", obj_num, value1f);
				}
			} else if (!strncmp(parse, "stats_width ", 12)) {
				parse += 12;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->width = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->width to %f.\n", obj_num, value1f);
				}
			} else if (!strncmp(parse, "stats_top ", 10)) {
				parse += 10;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->top_time = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->top_time to %f.\n", obj_num, value1f);
				}
			} else if (!strncmp(parse, "stats_interval ", 15)) {
				parse += 15;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->hline_interval = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->hline_interval to %f.\n", obj_num, value1f);
				}
			} else if (!strncmp(parse, "stats_scale ", 12)) {
				parse += 12;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->time_scale = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->time_scale to %f.\n", obj_num, value1f);
				}
			} else if (!strncmp(parse, "stats_color ", 12)) {
				parse += 12;
				value1f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value2f = atof(parse);
				parse = (strchr(parse, ',') == NULL ?		/* skip to past next comma */
					parse + strcspn(parse, whitespace) :	/*   or next whitespace    */
					strchr(parse, ',')+1);
				parse += strspn(parse, whitespace);		/* skip white */
				value3f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->back_color[0] = value1f;
					config->procs[obj_num]->stats->back_color[1] = value2f;
					config->procs[obj_num]->stats->back_color[2] = value3f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->back_color to %.3f,%.3f,%.3f.\n", obj_num, value1f, value2f, value3f);
				}
			} else if (!strncmp(parse, "stats_opac ", 11)) {
				parse += 11;
				value1f = atof(parse);
				if (config->procs[obj_num]->stats != NULL) {
					config->procs[obj_num]->stats->back_color[3] = value1f;
					if (style == verbose)
						vrFprintf(file, "set proc[%d] stats->back_color[3] to %f.\n", obj_num, value1f);
				}
			} else {
				vrFprintf(file, "Unknown proc request: '%s'.\n", request);
			}
		}
	} else

	/************/
	/* default: */ {
		vrFprintf(file, "Unknown request: '%s'.\n", request);
	}
}


/************************************************************/
/* vrLockCmd(): parse a request to alter the state of a     */
/*   FreeVR lock, and process it.  This is basically meant  */
/*   for debugging locks -- since they are frequently       */
/*   implemented in a different manner on differing systems.*/
/* NOTE: this function basically takes the input string     */
/*   with the "lock" or "l" (or whatever) stripped off.     */
/************************************************************/
/* NOTE: this function (and vrInfoGet(), which precedes) could easily*/
/*   be moved to another source file, as they don't have much to do  */
/*   with telnetting, other than this is the only process that makes */
/*   use of it at the moment.                                        */
void vrLockCmd(vrContextInfo *context, FILE *file, char *command, vrPrintStyle style)
{
static	char	*whitespace = " \t\r\b\n";
static	char	*help_msg = "Available settables:\n"
			TAB "help|? -- print this message (a list of possible setting requests)\n"
			TAB "num -- print the number of locks in the system\n"
			TAB "print|p <lock> -- print the state of the lock\n"
			TAB "printall|pa -- print the state of all the locks\n"
#if 0
			TAB "trace|t <lock> <name> -- put the lock into tracing mode\n"
			TAB "trace|t <lock> stop -- remove the lock from tracing mode\n"
#else
			TAB "trace|t <lock> <0|1> -- set the lock's tracing flag\n"
#endif
			TAB "setwrite|sw <lock> -- wait for and set the write state for the lock\n"
			TAB "relwrite|rw|uw <lock> -- release the write state for the lock\n"
			TAB "setread|sr <lock> -- wait for and set the write state for the lock\n"
			TAB "relread|rr|ur <lock> -- release the write state for the lock\n"
			TAB "create -- create a new lock\n"
		"";
	vrLock		lock;			/* lock upon which we are to operate */
#if 0
	char		*name;			/* name argument for some (trace) commands */
#endif
	char		argv[5][128];		/* for parsing argument lists */
	char		*str_begin;		/* for parsing next argument */
	int		str_len;		/* length of current argument */


	vrTrace("vrLockCmd", "top of vrLockCmd");
	vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Lock command received for '%s' [fp = %#p]\n", command, file);


	/********************************************************/
	/*** parse command here -- very simple-minded parsing ***/
	/********************************************************/

	/********/
	/* help */
	if ((!strncmp(command, "help", 4)) || (!strncmp(command, "?", 1))) {
		vrFprintf(file, help_msg);
	} else

	/**********/
	/* num */
	if (!strncmp(command, "num", 3)) {
		vrFprintf(file, "%d\n", vrCountLockList(context->head_lock));
	} else

	/**********/
	/* create */
	if (!strncmp(command, "create", 6)) {
		lock = vrLockCreateName(context, "telnet created");
		vrFprintf(file, "Created new lock: %#p\n", lock);
	} else

	/****************/
	/* print <lock> */
	if ((!strncmp(command, "print ", 6)) || (!strncmp(command, "p ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[6];

		lock = (vrLock)vrAtoP(str_begin);

		if (vrLockCheck(lock)) {
			vrFprintLock(file, lock, style);
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
	} else

	/*******************/
	/* printall <lock> */
	if ((!strncmp(command, "printall", 8)) || (!strncmp(command, "pa", 2))) {
		lock = context->head_lock;

		if (vrLockCheck(lock)) {
			vrFprintLockList(file, lock, style);
		} else {
			vrFprintf(file, "No locks!\n", lock);
		}
	} else

#if 0 /* test me { */
	/************************************/
	/* trace <lock> { <name> | "stop" } */
	if ((!strncmp(command, "trace ", 6)) || (!strncmp(command, "t ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[6];

		/* first argument */
		str_begin += strspn(str_begin, whitespace);	/* skip whitespace */
		str_len = strcspn(str_begin, whitespace);	/* count non-whitespace */
		strncpy(argv[1], str_begin, str_len);
		argv[1][str_len] = '\0';
		str_begin += str_len;

		lock = (vrLock)vrAtoP(argv[1]);

		if (vrLockCheck(lock)) {
			/* second argument */
			str_begin += strspn(str_begin, whitespace);	/* skip whitespace */
			str_len = strcspn(str_begin, whitespace);	/* count non-whitespace */
			strncpy(argv[2], str_begin, str_len);
			argv[2][str_len] = '\0';
			str_begin += str_len;

			if (!strncmp(argv[2], "stop", 4))
				name = NULL;
			else	name = argv[2];

			vrLockTrace(lock, name);
			vrFprintf(file, "Tracing lock: %#p as '%s'\n", lock, name);
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
#else /* } { */
	/**********************/
	/* trace <lock> <0|1> */
	if ((!strncmp(command, "trace ", 6)) || (!strncmp(command, "t ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[6];

		/* first argument */
		str_begin += strspn(str_begin, whitespace);	/* skip whitespace */
		str_len = strcspn(str_begin, whitespace);	/* count non-whitespace */
		strncpy(argv[1], str_begin, str_len);
		argv[1][str_len] = '\0';
		str_begin += str_len;

		lock = (vrLock)vrAtoP(argv[1]);

		if (vrLockCheck(lock)) {
			/* second argument */
			str_begin += strspn(str_begin, whitespace);	/* skip whitespace */
			str_len = strcspn(str_begin, whitespace);	/* count non-whitespace */
			strncpy(argv[2], str_begin, str_len);
			argv[2][str_len] = '\0';
			str_begin += str_len;

			if (vrAtoI(argv[2])) {
				vrLockTrace(lock, 1);
				vrFprintf(file, "Tracing lock: %#p\n", lock);
			} else {
				vrLockTrace(lock, 0);
				vrFprintf(file, "Not Tracing lock: %#p\n", lock);
			}
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
#endif /* } */
	} else

	/*******************/
	/* setwrite <lock> */
	if ((!strncmp(command, "setwrite ", 9)) || (!strncmp(command, "sw ", 3))) {
		if (command[2] == ' ')
			str_begin = &command[3];
		else	str_begin = &command[9];

		lock = (vrLock)vrAtoP(str_begin);

		if (vrLockCheck(lock)) {
			vrFprintf(file, "Setting lock %#p for writing -- may have to wait.\n", lock);
			fflush(file);				/* force the string to be sent */
			vrLockWriteSet(lock);
			vrFprintf(file, "Lock %#p now set for writing.\n", lock);
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
	} else

	/*******************/
	/* relwrite <lock> */
	if ((!strncmp(command, "relwrite ", 9)) || (!strncmp(command, "rw ", 3)) || (!strncmp(command, "uw ", 3))) {
		if (command[2] == ' ')
			str_begin = &command[3];
		else	str_begin = &command[9];

		lock = (vrLock)vrAtoP(str_begin);

		if (vrLockCheck(lock)) {
			vrFprintf(file, "Releasing lock %#p from writing.\n", lock);
			fflush(file);				/* force the string to be sent */
			vrLockWriteRelease(lock);
			vrFprintf(file, "Lock %#p now released from writing.\n", lock);
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
	} else

	/******************/
	/* setread <lock> */
	if ((!strncmp(command, "setread ", 8)) || (!strncmp(command, "sr ", 3))) {
		if (command[2] == ' ')
			str_begin = &command[3];
		else	str_begin = &command[8];

		lock = (vrLock)vrAtoP(str_begin);

		if (vrLockCheck(lock)) {
			vrFprintf(file, "Setting lock %#p for reading -- may have to wait.\n", lock);
			fflush(file);				/* force the string to be sent */
			vrLockReadSet(lock);
			vrFprintf(file, "Lock %#p now set for reading.\n", lock);
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
	} else

	/******************/
	/* relread <lock> */
	if ((!strncmp(command, "relread ", 8)) || (!strncmp(command, "rr ", 3)) || (!strncmp(command, "ur ", 3))) {
		if (command[2] == ' ')
			str_begin = &command[3];
		else	str_begin = &command[8];

		lock = (vrLock)vrAtoP(str_begin);

		if (vrLockCheck(lock)) {
			vrFprintf(file, "Releasing lock %#p from reading.\n", lock);
			fflush(file);				/* force the string to be sent */
			vrLockReadRelease(lock);
			vrFprintf(file, "Lock %#p now released from reading.\n", lock);
		} else {
			vrFprintf(file, "Invalid Lock: %#p\n", lock);
		}
	} else

	/************/
	/* default: */ {
		vrFprintf(file, "Unknown lock command: '%s'.\n", command);
	}
}


/************************************************************/
/* vrBarrierCmd(): parse a request to alter the state of a  */
/*   FreeVR barrier, and process it.  This is basically for */
/*   debugging barriers -- since they are built on locks,   */
/*   which are frequently implemented in a different manner */
/*   on differing systems.                                  */
/* NOTE: this function basically takes the input string     */
/*   with the "barrier" or "b" (or whatever) stripped off.  */
/************************************************************/
/* NOTE: this function (and vrInfoGet(), which precedes) could easily*/
/*   be moved to another source file, as they don't have much to do  */
/*   with telnetting, other than this is the only process that makes */
/*   use of it at the moment.                                        */
void vrBarrierCmd(vrContextInfo *context, FILE *file, char *command, vrPrintStyle style)
{
static	char	*help_msg = "Available settables:\n"
			TAB "help|? -- print this message (a list of possible setting requests)\n"
			TAB "num -- print the number of locks in the system\n"
			TAB "print|p <barrier> -- print the state of the barrier\n"
			TAB "printall|pa -- print the state of all the barriers\n"
			TAB "sync <barrier> -- wait for all members to be at the barrier\n"
			TAB "relonce <barrier> -- prematurely release the barrier\n"
			TAB "create -- create a new barrier\n"
			TAB "increment <barrier> -- add one member to the barrier\n"
			TAB "decrement <barrier> -- subtract one member from the barrier\n"
		"";
	vrBarrier	*barrier;		/* barrier upon which we are to operate */
	char		*str_begin;		/* for parsing next argument */

	vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Barrier command received for '%s' [fp = %#p]\n", command, file);

	/********************************************************/
	/*** parse command here -- very simple-minded parsing ***/
	/********************************************************/

	/********/
	/* help */
	if ((!strncmp(command, "help", 4)) || (!strncmp(command, "?", 1))) {
		vrFprintf(file, help_msg);
	} else

	/**********/
	/* num */
	if (!strncmp(command, "num", 3)) {
		vrFprintf(file, "%d\n", vrCountBarrierList(context->head_barrier));
	} else

	/**********/
	/* create */
	if (!strncmp(command, "create", 6)) {
		barrier = vrBarrierCreate(context, "telnet created", 1);
		vrFprintf(file, "Created new barrier: %#p\n", barrier);
	} else

	/****************/
	/* print <barrier> */
	if ((!strncmp(command, "print ", 6)) || (!strncmp(command, "p ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[6];

		barrier = (vrBarrier *)vrAtoP(str_begin);

		if (vrBarrierCheck(barrier)) {
			vrFprintBarrier(file, barrier, style);
		} else {
			vrFprintf(file, "Invalid Barrier: %#p\n", barrier);
		}
	} else

	/*******************/
	/* printall <barrier> */
	if ((!strncmp(command, "printall", 8)) || (!strncmp(command, "pa", 2))) {
		barrier = context->head_barrier;

		if (vrBarrierCheck(barrier)) {
			vrFprintBarrierList(file, barrier, style);
		} else {
			vrFprintf(file, "No barriers.\n", barrier);
		}
	} else

	/***************/
	/* sync <barrier> */
	if ((!strncmp(command, "sync ", 5)) || (!strncmp(command, "s ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[5];

		barrier = (vrBarrier *)vrAtoP(str_begin);

		if (vrBarrierCheck(barrier)) {
			vrFprintf(file, "Barrier %#p status: %d waiting, first? %d, last? %d.\n",
				barrier,
				barrier->num_waiting,
				vrBarrierFirstToSync(barrier),
				vrBarrierLastToSync(barrier));
			vrFprintf(file, "Synching on barrier %#p -- may have to wait.\n", barrier);
			fflush(file);				/* force the string to be sent */
			vrBarrierSync(barrier);
			vrFprintf(file, "Barrier %#p now synced.\n", barrier);
		} else {
			vrFprintf(file, "Invalid Barrier: %#p\n", barrier);
		}
	} else

	/*******************/
	/* relonce <lock> */
	if ((!strncmp(command, "relonce ", 8)) || (!strncmp(command, "ro ", 3)) || (!strncmp(command, "r ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else if (command[2] == ' ')
			str_begin = &command[3];
		else	str_begin = &command[8];

		barrier = (vrBarrier *)vrAtoP(str_begin);

		if (vrBarrierCheck(barrier)) {
			vrFprintf(file, "Barrier %#p status: %d waiting, first? %d, last? %d.\n",
				barrier,
				barrier->num_waiting,
				vrBarrierFirstToSync(barrier),
				vrBarrierLastToSync(barrier));
			vrFprintf(file, "Releasing Barrier %#p prematurely.\n", barrier);
			fflush(file);				/* force the string to be sent */
			vrBarrierReleaseOnce(barrier);
			vrFprintf(file, "Barrier %#p now released from writing.\n", barrier);
		} else {
			vrFprintf(file, "Invalid Barrier: %#p\n", barrier);
		}
	} else

	/********************/
	/* increment <lock> */
	if ((!strncmp(command, "increment ", 10)) || (!strncmp(command, "i ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[10];

//vrPrintf("str_begin = '%s'\n", str_begin); /* delete me */
		barrier = (vrBarrier *)vrAtoP(str_begin);
//vrPrintf("barrier = %p\n", barrier); /* delete me */
//vrPrintf("vrBarrierCheck(barrier) = %d\n", vrBarrierCheck(barrier)); /* delete me */


		if (vrBarrierCheck(barrier)) {
			vrBarrierIncrement(barrier, 1);
			vrFprintf(file, "Incremented members of barrier %#p to %d.\n", barrier, barrier->num_clients);
		} else {
			vrFprintf(file, "Invalid Barrier: %#p\n", barrier);
		}
	} else

	/********************/
	/* decrement <lock> */
	if ((!strncmp(command, "decrement ", 10)) || (!strncmp(command, "d ", 2))) {
		if (command[1] == ' ')
			str_begin = &command[2];
		else	str_begin = &command[10];

		barrier = (vrBarrier *)vrAtoP(str_begin);

		if (vrBarrierCheck(barrier)) {
			vrBarrierDecrement(barrier, 1);
			vrFprintf(file, "Decremented members of barrier %#p to %d.\n", barrier, barrier->num_clients);
		} else {
			vrFprintf(file, "Invalid Barrier: %#p\n", barrier);
		}
	} else

	/************/
	/* default: */ {
		vrFprintf(file, "Unknown barrier command: '%s'.\n", command);
	}
}


/************************************************************/
/* vrEcho(): echo a message back to the given FILE stream.  */
/*   This primary use for this is to evaluate any variables */
/*   found in the string to be echoed.                      */
/* NOTE: this function basically takes the input string    */
/*   with the "echo" or "e" (or whatever) stripped off.    */
/* NOTE: the echoing will end when reaching a semicolon or */
/*   a newline.                                            */
/***********************************************************/
/* NOTE: this function also could easily be moved to another     */
/*   source file, as they don't have much to do with telnetting, */
/*   other than this is the only process that makes use of it at */
/*   the moment.                                                 */
void vrEcho(vrContextInfo *context, FILE *file, char *string)
{
static	char		*whitespace = " \t\r\b";
static	char		*strnumchars = "_:abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.0123456789";
	int		count;
	int		string_length;
	char		*var_begin;		/* where the variable name begins in string */
	int		var_len;		/* length of variable name (less one) */
	char		variable_name[256];	/* copy of the full variable name */
	char		*variable_value;	/* returned value of the variable */

	string_length = strlen(string);

	for (count = 0; count < string_length; count++) {
		if (string[count] == ';' || string[count] == '\n') {
			/* the semicolon and newline indicate the end of this command, so break out of the loop */
			break;
		} else if (string[count] != '$') {
			/* for all other characters (except the '$' which indicates a variable), just echo it */
			vrFprintf(file, "%c", string[count]);
		} else {
			var_begin = &string[count+1];	/* variable name begins right after '$' */
			if (strspn(var_begin, whitespace) > 0) {
				/* if whitespace follows the '$' then just print the '$' and move on */
				vrFprintf(file, "%c", string[count]);
			} else {
				/* otherwise treat the next character and all alphanums following as variable name */
				/* NOTE: this will pass on non-alphanumerics in the first position */
				/*   for special operations such as the $?<var> construct.         */
				var_len = strspn(var_begin+1, strnumchars);
				strncpy(variable_name, var_begin, var_len+1);
				variable_name[var_len+1] = '\0';
				count += var_len+1;

				variable_value = vrEvaluateVariable(context, variable_name);
				if (variable_value != NULL)
					vrFprintf(file, "%s", variable_value);
				else	vrFprintf(file, "<null>");
			}

		}
	}
	vrFprintf(file, "\n");
}


/*****************************************************************************/
/* vrEchoTest(): uses the new vrSubstituteVariablesInString() function, and  */
/*   for now is how I'm testing the robustness of that function.  But in the */
/*   future, this may be how all telnet commands are processed.              */
void vrEchoTest(vrContextInfo *context, FILE *file, char *string)
{
	vrFprintf(file, ">>>%s\n", vrSubstituteVariablesInString(context, string));
}


/**********************************************************/
/* vrSetenv(): Set the value of an environment variable.  */
/* NOTE: this function basically takes the input string   */
/*   with the "setenv" (or whatever) stripped off.        */
/*****************************************************************/
/* NOTE: this function also could easily be moved to another     */
/*   source file, as they don't have much to do with telnetting, */
/*   other than this is the only process that makes use of it at */
/*   the moment.                                                 */
void vrSetenv(FILE *file, char *string)
{
static	char		*whitespace = " \t\r\b";
static	char		variable_name[256];	/* copy of the full variable name */
static	char		assignment_string[1024];/* string to give to putenv to make assignment */
	int		var_len;		/* length of variable name (less one) */
	char		*variable_value;	/* returned value of the variable */

	string += strspn(string, whitespace);	/* skip white */
	var_len = strcspn(string, whitespace);	/* measure non-whitespace -- this is the variable name */
	strncpy(variable_name, string, var_len);
	variable_value = string + var_len + strspn(string+var_len, whitespace);

	snprintf(assignment_string, sizeof(assignment_string), "%s=%s", variable_name, variable_value);
//vrPrintf("making assignment '%s'\n", assignment_string); /* delete me */

	if (putenv(assignment_string)) {
		vrErrPrintf(RED_TEXT "FreeVR telnet: Unable to set the variable '%s'\n" NORM_TEXT, variable_name);
		vrFprintf(file, "Unable to make assign value to environment variable\n");
	}
}


/***********************************************************/
/* vrUnsetenv(): Set the value of an environment variable. */
/* NOTE: this function basically takes the input string    */
/*   with the "unsetenv" (or whatever) stripped off.       */
/*****************************************************************/
/* NOTE: this function also could easily be moved to another     */
/*   source file, as they don't have much to do with telnetting, */
/*   other than this is the only process that makes use of it at */
/*   the moment.                                                 */
/* NOTE: this doesn't work under IRIX, and perhaps other Unixii. */
void vrUnsetenv(FILE *file, char *string)
{
static	char		*whitespace = " \t\r\b";
static	char		variable_name[256];	/* copy of the full variable name */
static	char		assignment_string[1024];/* string to give to putenv to make assignment */
	int		var_len;		/* length of variable name (less one) */

	string += strspn(string, whitespace);	/* skip white */
	var_len = strcspn(string, whitespace);	/* measure non-whitespace -- this is the variable name */
	strncpy(variable_name, string, var_len);

	/* NOTE: this doesn't work under IRIX, and perhaps other Unixii */
	snprintf(assignment_string, sizeof(assignment_string), "%s", variable_name);
//vrPrintf("making assignment '%s'\n", assignment_string); /* delete me */

	if (putenv(assignment_string)) {
		vrErrPrintf(RED_TEXT "FreeVR telnet: Unable to unset the variable '%s'\n" NORM_TEXT, variable_name);
		vrFprintf(file, "Unable to unset the environment variable\n");
	}
}


/******************************************************************/
/* vrTelnetInitProc(): Initialize the socket communication with   */
/*   this process.                                                */
/******************************************************************/
void vrTelnetInitProc(vrProcessInfo *myproc_info)
{
	_TelnetPrivate	*aux;			/* pointer to auxiliary data of the telnet process */
	int		result;			/* used to parse socket port */
	int		inet_port_max;		/* highest number of port to attempt opening */
struct	linger		linger_opt = { 1, 0 };	/* values for the SO_LINGER control options */


	vrTrace("vrTelnetInitProc", "beginning");
	proc_info = myproc_info;

#if 0 /* TODO: readd this -- not doing this allows for seg faults, so we can debug easier */
	vrSetSignalHandler(vrTelnetSignalHandler);
#endif

	/***************************************************************/
	/* create your own little memory space for proc-specific stuff */
	myproc_info->aux_data = (_TelnetPrivate *)vrShmemAlloc(sizeof(*aux));
	aux = myproc_info->aux_data;
//vrPrintf("vrTelnetInitProc(): allocated %d bytes for telnet process aux_data -- at %p\n", sizeof(*aux), aux);


	/*******************************************************/
	/* set some initial values for the auxiliary structure */
	aux->inet_port = 3000;		/* the current default socket port */
	aux->inet_port_range = 2;	/* the number of other (consecutive) ports to try */
	aux->listen_sock = -1;		/* -1 indicates ... */
	aux->cmd_socket = -1;		/* -1 indicates no one is connected */
	aux->cmd_fp = NULL;
	aux->style = verbose;		/* default to verbose query outputs */
	aux->prompt = vrShmemStrDup(BOLD_TEXT "FreeVR>" NORM_TEXT);
	aux->password = vrShmemStrDup("");
	aux->num_history = 0;		/* no items in history list yet */


	/****************************************************************/
	/* adjust auxiliary structure values based on process arguments */
	if (vrArgParseInteger(myproc_info->args, "port", &result)) {
		aux->inet_port = result;
	}

	if (vrArgParseInteger(myproc_info->args, "portrange", &result)) {
		aux->inet_port_range = result;
	}

	vrArgParseString(myproc_info->args, "prompt", &(aux->prompt));

	vrArgParseString(myproc_info->args, "password", &(aux->password));


	/********************************************/
	/* output debugging info about this process */
	if (vrDbgDo(DEFAULT_DBGLVL+1)) {
		vrMsgPrintf("FreeVR: Telnet proc is: ");
		vrFprintProcessInfo(stdout, myproc_info, verbose);
	}


	/******************************/
	/* initialize the socket port */
	vrTrace("vrTelnetInitProc", "initializing socket port");
	gethostname(aux->host_machine, 128);

	/* if range is 0 or less, wait for specific port */
	if (aux->inet_port_range > 0) {
		/* loop through N possible Internet ports to hopefully find one that works */
		vrDbgPrintf("FreeVR (telnet): Attempting to use ");
		for (inet_port_max = aux->inet_port + aux->inet_port_range; aux->listen_sock < 0 && aux->inet_port <= inet_port_max; aux->inet_port++) {
			vrDbgPrintf("port %d ... ", aux->inet_port);
			vrDbgFlush();
			aux->listen_sock = vrSocketCreateListen(&aux->inet_port, 0);
		}
		aux->inet_port--;
		vrDbgPrintf("\n");
	} else {
		/* loop over time until the requested socket port is available */
		aux->listen_sock = vrSocketCreateListen(&aux->inet_port, 0);
		while (aux->listen_sock < 0) {
			vrMsgPrintf("FreeVR (telnet): " RED_TEXT "Port %d -- unable to open, will continue to try.\n" NORM_TEXT, aux->inet_port);
			vrSleep(5000000);		/* 5 seconds */
			aux->listen_sock = vrSocketCreateListen(&aux->inet_port, 0);

			/* break out of the waiting for socket loop if this process has been set for termination */
			if (myproc_info->req_proc_term)
				break;
		}
	}

	if (aux->listen_sock < 0) {
		vrTrace("vrTelnetInitProc", "unable to find an open socket");
		vrMsgPrintf("FreeVR (telnet): " RED_TEXT "WARNING: Unable to find an open socket port.\n" NORM_TEXT);
	} else {
		vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s %s:%d\n", __FILE__, __LINE__, "vrTelnetInitProc", "Listening on socket", aux->host_machine, aux->inet_port);
		vrMsgPrintf("\rFreeVR (telnet): Connect to '%s:%d' to send commands to FreeVR\n", aux->host_machine, aux->inet_port);
	}

	/* Turn off the post-disconnect linger */
	result = setsockopt(aux->listen_sock, SOL_SOCKET, SO_LINGER, (void *)&linger_opt, sizeof(linger_opt));
	if (result < 0) {
		vrTrace("vrTelnetInitProc", "unable set linger socket option.");
		vrMsgPrintf("FreeVR (telnet): " RED_TEXT "WARNING: Unable to set linger socket option.\n" NORM_TEXT);
	}

	/* prepare the parsing buffer */
	aux->parse = &(aux->inbuf[0]);
	aux->parse[0] = '\0';
//vrPrintf("End of INIT: parse = '%s'(%d), '%c'\n", aux->parse, strlen(aux->parse), aux->parse[0]);
}


/*************************************************************/
/* vrTelnetTermProc(): Close the socket port associated with */
/*   this process.                                           */
/*************************************************************/
void vrTelnetTermProc(vrProcessInfo *myproc_info)
{
	_TelnetPrivate	*aux;			/* pointer to auxiliary data of the telnet process */

	aux = myproc_info->aux_data;
	if (aux->listen_sock > 0)
		vrSocketClose(aux->listen_sock);
}


/********************************************************************/
/* _TelnetParseCommand(): Parse one command from user (or history). */
/********************************************************************/
/* TODO: decide whether I prefer one over the other: vrSocketSendMsg() vs. vrFprintf() */
int _TelnetParseCommand(vrProcessInfo *myproc_info, char *parse)
{
static	char	*help_msg =
			"\nNOTE: full documentation is availble in the \"telnet\" manpage.\n\n"
			"Socket commands:\n"
			TAB "help -- this message\n"
			TAB "history [<n>] -- show past commands\n"
			TAB "{!! | !<n> | !<-n> } -- reexecute past commands\n"
			TAB "close -- close the connection\n"
			TAB "quit -- quit the telnet process\n"
		  	TAB "term -- terminate FreeVR application\n"
		  /*	TAB "reset -- nothing (yet)\n"						*/
			TAB "pause -- toggle the system pause status\n"
			TAB "echo <string> -- echo the string back to the output, with variable substitutions\n"
			TAB "printstyle|style [<style>] -- set print style to given mode\n"
			TAB "verbose -- set print style to verbose mode\n"
			TAB "brief -- set print style to brief mode\n"
			TAB "one_line -- set print style to one_line mode\n"
			TAB "machine -- set print style to machine mode (and disable the prompt)\n"
			TAB "diagram -- set print style to diagram (GraphViz-dot) mode\n"
			TAB "fileformat -- set the print style to file_format mode\n"
			TAB "print|p <query> -- print some info about the FreeVR app|system\n"
			TAB "printfile|pf [+]<filename> <query> -- print some info about the FreeVR app to a file\n"
			TAB "set|s <request> -- set the value of part of the system\n"
			TAB "setenv <envvar> <value> -- set the value of an environment variable\n"
			TAB "unsetenv <envvar> -- unset the environment variable\n"
			TAB "lock <command> -- send a command to a FreeVR lock\n"
			TAB "barrier <command> -- send a command to a FreeVR barrier\n"
			TAB "relbar <address> -- release the barrier at that address\n"
		/* :-(	TAB "procdown <num> -- stop the given process\n"			*/
		/* :-(	TAB "procup <name> -- start a process of the given name\n"	*/
		"";
static	char	*help_msg_nudge = "Nudge commands:\n"
			TAB "set input <6-sensor> { nudger2eA | nra } { id | <x> <y> <z> } -- shift 6-sensor's r2e angles\n"
			TAB "set input <6-sensor> { nudger2eT | nrt } { id | <x> <y> <z> } -- shift 6-sensor's r2e location\n"
			TAB "set inputdevice[<num>] { nudget2rwA | nta } { id | <x> <y> <z> } -- shift the t2rw Euler angles\n"
			TAB "set inputdevice[<num>] { nudget2rwT | ntt } { id | <x> <y> <z> } -- shift the t2rw Translation \n"
			TAB "set window[<num>] nudgex <value> -- shift the window along the X-axis\n"
			TAB "set window[<num>] nudgey <value> -- shift the window along the Y-axis\n"
			TAB "set window[<num>] nudgez <value> -- shift the window along the Z-axis\n"
			TAB "set window[<num>] nudgerx <value> -- rotatedthe window around the X-axis\n"
			TAB "set window[<num>] nudgery <value> -- rotatedthe window around the Y-axis\n"
			TAB "set window[<num>] nudgerz <value> -- rotatedthe window around the Z-axis\n"
	// :-(		TAB "set user[<num>] nudgeT <v1>,<v2>,<v3> -- translate the user's travel.\n"
	// :-(		TAB "set user[<num>] nudgeA <v1>,<v2>,<v3> -- rotate the user's travel.\n"
		"";
static	char		*whitespace = " \t\r\b";
static	char		*digits = "0123456789";
	vrContextInfo	*context;		/* The overall FreeVR context -- obtained from myproc_info */
	_TelnetPrivate	*aux;			/* pointer to auxiliary data of the telnet process */
	int		ival;			/* integer value -- currently for history length */
	int		count;			/* loop counter */
	char		msg_buffer[1024];	/* buffer for sending messages through socket */
	char		*parse_begin;		/* pointer to original first character */
	int		history_relative;	/* whether a history re-parse is relative or absolute */


	context = myproc_info->context;
	aux = myproc_info->aux_data;
	parse_begin = parse;

	/* add current input to "history" list */
	if (aux->num_history < (sizeof(aux->history)/sizeof(aux->history[0]))) {
		/* Duplicate the string about to be parsed into history array */
		/*   NOTE: strchrnul() points to either the '\n' or '\0',     */
		/*   whichever comes first.  In the case where there is a '\n'*/
		/*   then overwrite that with the nul ('\0') to terminate the */
		/*   string without keeping the newline.                      */
#if defined(__APPLE__) || defined(__CYGWIN__)
		/* Apple & Cygwin don't have strchrnul(), so need to use a conditional */
		if (strchr(parse, '\n') != NULL)
			aux->history[aux->num_history] = vrShmemMemDup(parse, (strchr(parse, '\n') - parse)+1);
		else	aux->history[aux->num_history] = vrShmemMemDup(parse, strlen(parse)+1);
#else
		aux->history[aux->num_history] = vrShmemMemDup(parse, (strchrnul(parse, '\n') - parse)+1);
#endif
#if defined(__APPLE__) || defined(__CYGWIN__)
		/* Apple & Cygwin don't have strchrnul(), so need to use a conditional */
		if (strchr(aux->history[aux->num_history], '\n') != NULL)
			*strchr(aux->history[aux->num_history], '\n') = '\0';
#else
		(strchrnul(aux->history[aux->num_history], '\n'))[0] = '\0';	/* odd syntax is correct, but should be unnecessary, other than to make gcc happy. */
#endif
		aux->num_history++;
		vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): adding '%s' to history list: %d\n", aux->history[aux->num_history-1], aux->num_history);
	}

	/********/
	/* quit */
	if (!strncmp(parse, "quit", 4)) {
		/* set the process for termination */
		myproc_info->req_proc_term = 1;
	} else

	/********/
	/* term */
	/* TODO: this should eventually be handled as an "alias" */
	if (!strncmp(parse, "term", 4)) {
		/* terminate FreeVR */
#if 0
		vrExit();
		exit(0);
#else /* kludge method is to set the first 2-way input to 1 (and then close sockets) */
		vrInfoSet(context, aux->cmd_fp, "input 2-way[0] 1", aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */

		vrSleep(200);		/* .0002 seconds */
		if (aux->listen_sock > 0)
			vrSocketClose(aux->listen_sock);
		close(aux->cmd_socket);
		aux->cmd_socket = -1;
#endif
	} else

	/********/
	/* help */
	if (!strncmp(parse, "help", 4) || !strncmp(parse, "?", 1)) {
		/* give usage information */
		vrSocketSendMsg(aux->cmd_socket, help_msg);
	} else

	/**************/
	/* nudge help */
	if (!strncmp(parse, "nud", 3)) {
		/* give nudge usage information */
		vrSocketSendMsg(aux->cmd_socket, help_msg_nudge);
	} else

	/*****************/
	/* history [<n>] */
	if (!strncmp(parse, "history", 7)) {
		/* List past command history */
		parse += 7;
		ival = aux->num_history;	/* show all history by default */

		/* if there's whitespace after the command, look for a number */
		if (strspn(parse, whitespace) > 0) {
			parse += strspn(parse, whitespace);	/* skip whitespace */
			if (strspn(parse, digits) > 0) {		/* look for positive number of digits */
				ival = vrAtoI(parse);
				if (ival > aux->num_history)
					ival = aux->num_history;
			} else if (strcspn(parse, digits) > 0) {
				vrSocketSendMsg(aux->cmd_socket, RED_TEXT "Error: Unknown argument to \"history\" command.\n" NORM_TEXT);
			}
		}

		/* Output the history list */
		for (count = aux->num_history-ival; count < aux->num_history; count++) {
			snprintf(msg_buffer, sizeof(msg_buffer), BOLD_TEXT "%3d:" NORM_TEXT " %s\n", count, aux->history[count]);
			vrSocketSendMsg(aux->cmd_socket, msg_buffer);
		}
	} else

	/*************************************************/
	/* historyrepeat: ![{ ! | <n> | <-n> | <chars>}] */
	/* TODO: finish this section -- the <chars> option is not yet implemented */
	if (!strncmp(parse, "!!", 2)) {
		/* TODO: simply repeat the last command */
		aux->num_history--;				/* remove this line from the history */
		if (aux->num_history > 0) {
			vrFprintf(aux->cmd_fp, BOLD_TEXT "reexecuting: '%s'\n" NORM_TEXT, aux->history[aux->num_history-1]);
			fflush(aux->cmd_fp);	/* socket fp's don't flush often */
			_TelnetParseCommand(myproc_info, aux->history[aux->num_history-1]);
		} else {
			/* There is no history yet -- this was the first thing! */
			vrSocketSendMsg(aux->cmd_socket, RED_TEXT "Error: There is no history yet.\n" NORM_TEXT);
		}
	} else
	if (!strncmp(parse, "!", 1)) {
		aux->num_history--;				/* remove this line from the history */

		parse++;					/* skip the "!" */
		parse += strspn(parse, whitespace);		/* skip whitespace */

		/* check for a relative reference */
		history_relative = 0;
		if (parse[0] == '-') {
			parse++;
			history_relative = 1;
		}

		/* get the numeric input */
		if (strspn(parse, digits) > 0) {	/* look for positive number of digits */
			ival = vrAtoI(parse);
			if (history_relative)
				ival = aux->num_history - ival;
			if (ival < aux->num_history && ival >= 0) {
				/* repeat this history number */
				vrFprintf(aux->cmd_fp, BOLD_TEXT "reexecuting: '%s'\n" NORM_TEXT, aux->history[ival]);
				fflush(aux->cmd_fp);	/* socket fp's don't flush often */
				_TelnetParseCommand(myproc_info, aux->history[ival]);
			} else {
				/* no number in the history */
				vrSocketSendMsg(aux->cmd_socket, RED_TEXT "Error: There no history of that yet.\n" NORM_TEXT);
			}
		} else {
			/* TODO: search back for a command that starts with the same given string */
			vrSocketSendMsg(aux->cmd_socket, RED_TEXT "Error: non-numeric history hasn't been implemented yet.\n" NORM_TEXT);
		}

	} else

	/*********/
	/* close */
	if (!strncmp(parse, "close", 5)) {
		/* close socket connection */
		vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Client ending communication.\n");
		close(aux->cmd_socket);
		aux->cmd_socket = -1;
	} else

	/*********/
	/* reset */
	if (!strncmp(parse, "reset", 5)) {
		/* make the system/world as it was in the beginning */

		/* TODO: do stuff */

	} else

	/*********************************************/
	/* verbose|brief|one_line|machine|fileformat */
	if ((!strncmp(parse, "printstyle ", 10)) || (!strncmp(parse, "style", 5))) {
		/* skip the operation name */
		parse += (parse[0] == 'p' ? 10 : 5);

		/* if there's whitespace after the command, look for a style */
		if (strspn(parse, whitespace) > 0) {
			parse += strspn(parse, whitespace);	/* skip whitespace */

			/* Pass the remaining string as a style option back */
			/*   to the parser.                                 */
			aux->style = vrPrintStyleValue(parse);
		}
		vrFprintf(aux->cmd_fp, BOLD_TEXT "printstyle is: '%s'\n" NORM_TEXT, vrPrintStyleName(aux->style));
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else
	if (!strncmp(parse, "verbose", 7)) {
		aux->style = verbose;
	} else
	if (!strncmp(parse, "brief", 5)) {
		aux->style = brief;
	} else
	if (!strncmp(parse, "one_line", 8)) {
		aux->style = one_line;
	} else
	if (!strncmp(parse, "machine", 7)) {
		aux->style = machine;
		vrSocketSendMsg(aux->cmd_socket, "\n");
		fflush(aux->cmd_fp);
	} else
	if (!strncmp(parse, "diagram", 7)) {
		aux->style = diagram;
	} else
	if (!strncmp(parse, "file", 4)) {
		aux->style = file_format;
	} else
	if (!strncmp(parse, "fileformat", 10)) {
		aux->style = file_format;
	} else

	/*********/
	/* print */
	if (!strncmp(parse, "print ", 6)) {
		/* print some information about the system */
		vrInfoGet(context, aux->cmd_fp, &parse[6], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else
	if (!strncmp(parse, "p ", 2)) {
		/* (shortcut) print some information about the system */
		vrInfoGet(context, aux->cmd_fp, &parse[2], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/*************/
	/* printfile */
	if ((!strncmp(parse, "printfile ", 10)) || (!strncmp(parse, "pf ", 3))) {
		/* print some information about the system to a file */
		FILE	*file;		/* the file to output to */
		char	*filename;	/* the name of the file to output to */
		char	*opentype = "w";/* how to write to the file (from beginning[default], or append) */

		/* skip the operation name */
		parse += (parse[1] == 'r' ? 10 : 3);
		if (parse[2] == ' ')
			parse = parse + 3;

		filename = parse;

		parse += strcspn(parse, whitespace);		/* skip non-white */
		parse[0] = '\0';
		parse++;
		parse += strspn(parse, whitespace);		/* skip non-white */

		/* open the file */
		/* NOTE: if filename begins with a '+' do an append */
		if (filename[0] == '+') {
			opentype = "a";
			filename++;
		}

		file = fopen(filename, opentype);

		/* output the information */
		if (file != NULL) {
			vrInfoGet(context, file, parse, aux->style);
			fclose(file);
		} else {
			vrSocketSendMsg(aux->cmd_socket, "Unable to open file for output.\n");
		}
	} else

	/*******/
	/* set */
	if (!strncmp(parse, "set ", 4)) {
		/* change some information about the system */
		vrInfoSet(context, aux->cmd_fp, &parse[4], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else
	if (!strncmp(parse, "s ", 2)) {
		/* change some information about the system */
		vrInfoSet(context, aux->cmd_fp, &parse[2], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/********/
	/* lock */
	if (!strncmp(parse, "lock ", 5)) {
		/* change some information about the lock */
		vrLockCmd(context, aux->cmd_fp, &parse[5], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else
	if (!strncmp(parse, "l ", 2)) {
		/* change some information about the lock */
		vrLockCmd(context, aux->cmd_fp, &parse[2], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/***********/
	/* barrier */
	if (!strncmp(parse, "barrier ", 8)) {
		/* change some information about the barrier */
		vrBarrierCmd(context, aux->cmd_fp, &parse[8], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else
	if (!strncmp(parse, "b ", 2)) {
		/* change some information about the barrier */
		vrBarrierCmd(context, aux->cmd_fp, &parse[2], aux->style);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/**********/
	/* relbar */
	if (!strncmp(parse, "relbar ", 7)) {
		/* release a barrier just once */
		vrBarrierReleaseOnce((void *)vrAtoP(&parse[7]));
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/*********/
	/* pause */
	if (!strncmp(parse, "pause", 5)) {
		/* toggle the system pause value */
		context->paused ^= 1;
		if (aux->style == verbose)
			vrFprintf(aux->cmd_fp, "Set System Pause value to %d\n", context->paused);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/**********/
	/* setenv */
	if (!strncmp(parse, "setenv ", 7)) {
		/* set a machine environment variable */
		vrSetenv(aux->cmd_fp, &parse[7]);
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/********/
	/* echo */
	if (!strncmp(parse, "echo ", 5)) {
		/* echo command -- repeat string to output w/ variable substitutions */
#if 1 /* for testing new vrSubstituteVariablesInString() function */
		vrEchoTest(context, aux->cmd_fp, &parse[5]);
#else
		vrEcho(context, aux->cmd_fp, &parse[5]);
#endif
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else

	/*******/
	/* nop */
	if (!strcmp(parse, "")) {
		/* empty command -- do nothing */
		fflush(aux->cmd_fp);	/* socket fp's don't flush often */
	} else


	/************/
	/* default: */ {
//vrPrintf("parse = '%s'(%d), '%c'\n", parse, strlen(parse), parse[0]);
		/* NOTE: the "missing argument" is because a trailing space is */
		/*   part of the command match, so type e.g. "barrier" will    */
		/*   result in this event, providing a confusing response.     */
		vrSocketSendMsg(aux->cmd_socket, RED_TEXT "Error: Unknown command (or missing argument).\n" NORM_TEXT);
		vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Unknown command warning sent.\n", parse);
	}


	/* shift the parse pointer to be after the first semicolon or newline */
	/*   (ie. in case there is an additional command in the buffer)       */
	parse += strcspn(parse, ";\n");			/* skip to ;|NL */
	parse++;					/* skip over ;|NL */


	return (parse - parse_begin);	/* return number of characters parsed */
}


/******************************************************************/
/* vrTelnetOneFrame(): Handle one interface with the user.        */
/******************************************************************/
void vrTelnetOneFrame(vrProcessInfo *myproc_info)
{
static	char		*whitespace = " \t\r\b";
	_TelnetPrivate	*aux;			/* pointer to auxiliary data of the telnet process */
	int		result;			/* used to parse socket port */
	int		chars_parsed;		/* number of characters parsed by _TelnetParseCommand() */

static int timeshere = 0;
	timeshere++;

	aux = myproc_info->aux_data;
//vrPrintf(">>>>>: timeshere = %d, strlen = %d, '%c' aux is at %p\n", timeshere, strlen(aux->parse), aux->parse[0], aux);

	/* calculate frame rates and set the process time values */
	myproc_info->frame_count++;
	vrProcessCalcFrameRate(myproc_info);

	/* measure: update time measurement array index */
	vrProcessStatsNextFrame(myproc_info->stats);

	/*******************************/
	/* Handle Commands from socket */

	/* if no connection (and a listen socket is open), see if someone is calling */
	if (aux->cmd_socket < 0 && aux->listen_sock >= 0) {
		/* aux->cmd_socket will be -1 when no one is connected */

		aux->cmd_socket = vrSocketAnswer(aux->listen_sock);
		if (aux->cmd_socket > 0) {
			vrTrace("vrTelnetOneFrame", "socket connection established");
			vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): New client connected.\n");
//vrPrintf("!!!!!: timeshere = %d, strlen = %d, '%c' aux is at %p\n", timeshere, strlen(aux->parse), aux->parse[0], aux);
			aux->cmd_fp = fdopen(aux->cmd_socket, "r+");

			/* Verify password from socket */
			if (strlen(aux->password) > 0) {
/*** TODO: turn off echoing!! ***/
				vrSocketSendMsg(aux->cmd_socket, "Enter FreeVR password [NOTE: currently visible (TODO: make invisible)]> ");

				/* wait for the password input */
				while (vrSocketReadMsgClean(aux->cmd_socket, aux->inbuf, sizeof(aux->inbuf)) <= 0)
					vrSleep(myproc_info->usec_min);

				/* compare the given password with config password */
				/* TODO: use an encoded password [PGP perhaps] */
				if (!strcmp(aux->password, aux->inbuf)) {
					vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Correct password given.\n");
				} else {
					vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): " RED_TEXT "Incorrect password given.\n" NORM_TEXT);
					vrSocketSendMsg(aux->cmd_socket, "Incorrect password.\n");
					close(aux->cmd_socket);
					aux->cmd_socket = -1;
				}

/*** TODO: turn on echoing!! ***/

				/* if correct password, then prevent interpretation as input command */
				aux->parse = strchr(aux->inbuf, '\0');
				aux->parse++;
			}

			/* if correct password, then send a prompt */
			if ((aux->cmd_socket > 0) && (aux->style != machine)) {
				vrSocketSendMsg(aux->cmd_socket, aux->prompt);
			}
		}
	}

	/* if a connection, see whether client sent a command.  If so, act on it */
	if (aux->cmd_socket > 0) {
//vrPrintf("$$$$$: strlen = %d, '%c'\n", strlen(aux->parse), aux->parse[0]);

		/* If no leftover material then refill input buffer */
		if (aux->parse[0] == '\0') {
			result = vrSocketReadMsgClean(aux->cmd_socket, aux->inbuf, sizeof(aux->inbuf));
			aux->parse = aux->inbuf;
//vrPrintf("result = %d, parse = '%s'(%d), '%c'\n", aux->parse, strlen(aux->parse), aux->parse[0]);

			switch (result) {
			case  EOF:
				vrTrace("vrTelnetOneFrame", "socket connection closed");
				vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): Client stopped communicating.\n");
				close(aux->cmd_socket);	/* NOTE: vrSocketClose() is not used because it is for the listen socket. */
				fclose(aux->cmd_fp);
				aux->cmd_socket = -1;
				return;//break;

			/* NOTE: previously "0" indicated EOF, but now it simply indicates a 0-length message. */
			/*   So now what used to be the error number (-1) is used by EOF, so we use -2 to      */
			/*   indicate an actual error.                                                         */
			case -2:
				vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): socket read error\n");
				return;//break;
			default:
				vrDbgPrintfN(PARSE_DBGLVL, "FreeVR (telnet): socket cmd = '%s'\n", aux->parse);
			}
		}

	/* TODO: allow a programmer-defined callback to be inserted here to */
	/*   have first dibs at the incoming commands.  OR, call either the */
	/*   callback, or a function containing the following commands,     */
	/*   which itself might be called from the programmer code.  Also,  */
	/*   we might want to have the option of a DSO-based function being */
	/*   used to do the parsing, etc.  With the DSO function defined in */
	/*   the process-config arguments.                                  */
	/* NOTE: this parsing by the application would be in another thread,*/
	/*   so make sure to test the implications of that.                 */

		/* first, skipover whitespace */
		aux->parse += strspn(aux->parse, whitespace);			/* skip white */
//vrPrintf("2: parse = '%s'(%d), '%c'\n", aux->parse, strlen(aux->parse), aux->parse[0]);

		/* if nothing to parse, don't parse it */
		if (strlen(aux->parse) > 0) {
			/* parse command here -- very simple-minded parsing */
			vrTrace("vrTelnetOneFrame", BOLD_TEXT "parsing input" NORM_TEXT);

			chars_parsed = _TelnetParseCommand(myproc_info, aux->parse);
			aux->parse += chars_parsed;
		}

		/***************************************************************************/
		/* prompt for another command (if socket still open & no command in inbuf) */
		if ((aux->cmd_socket > 0) && (aux->style != machine) && (aux->parse[0] == '\0')) {
			vrSocketSendMsg(aux->cmd_socket, aux->prompt);
		}
	}
//else vrPrintf("No socket command\n");
}


/********************************************************/
/* the main loop for each telnet communications process */
/********************************************************/
void vrTelnetMainLoop(vrProcessInfo *myproc_info)
{
	_TelnetPrivate	*aux;			/* pointer to auxiliary data of the telnet process */
	vrTime		loop_wtime;		/* time of the beginning of each loop */

	aux = myproc_info->aux_data;

	/*************************/
	/*** Main process loop ***/
	/*************************/

	vrTrace("vrTelnetMainLoop", BOLD_TEXT "beginning process loop" NORM_TEXT);
	loop_wtime = vrCurrentWallTime();
	while (!myproc_info->req_proc_term) {
//vrPrintf("XX: parse = '%s'(%d), '%c'\n", aux->parse, strlen(aux->parse), aux->parse[0]);

		/* NOTE: I'm not sure why the frame delay and barrier stuff are at the top */
		/*   of this loop -- probably due to how things work for the visren loop.  */

		/**************************/
		/* do minimal frame delay */
		vrSleep(myproc_info->usec_min - (long)((vrCurrentWallTime() - loop_wtime) * 1000000.0));

		/*************************************************************/
		/* Sync with other processes in group & calculate frame info */
		vrProcessSync(myproc_info, 0, 0);		/* NOTE: the last two arguments are for stats info that we don't worry about for telnet */

		/* start measuring for minimal loop time */
		loop_wtime = vrCurrentWallTime();

//vrPrintf("Calling vrTelnetOneFrame()\n");
		vrTelnetOneFrame(myproc_info);
	}

	vrSocketClose(aux->listen_sock);

	vrTrace("vrTelnetMainLoop", "ending");
}

